$TITLE ('COS DCE Operate-mode Main Program, version 01.032')
$PAGELENGTH(42)
;*****************************************************************************
; File -- opermain.a51
; Purpose -- COS DCE 8051 operate code.
;
; Author -- William I. Clement, Ph.D., President
;           Clement Engineering, Inc.
;           1261 Dogwood Road
;           Arnold, MD  21012-2220
;           (410) 518-6591
;
; Change History --
; 23Oct00   First Operate-mode version based on Boot-mode verion 01.05
; 10Nov00   WIC, delivered operate version 01.000
; 13Nov00   WIC, version 01.001 changes:
;           (1) LFDDNLOD command now fills unused portion of download buffer
;               with character 'DD'.
;           (2) HST error parameter for Diagnostic 02 is now the computed
;               CRC for the upload region (LFMXFER), vice the expected CRC.
;           (3) Upload/Download computed CRC (LFMXFER) now in housekeeping.
; 14Nov00   WIC, version 01.002 changes:
;           (1) Populate the status words used in operate but not boot (an
;               oversight in the previous deliveries).
;           (2) Redefine CRC regions so as not to cover any patchable
;               constants or scratchpad areas.  These were causing numerous
;               CRC change diagnostics.  The regions are now defined by labels
;               found in this file and translated into the CRC regions
;               structures found in 'patchcst.a51'.
; 05Dec00   WIC, version 01.003 changes:
;           (1) Correct comments in unused ISRs.
;           (2) Add minor efficiency to HV_RAMP activity.
;           (3) Add comments where 'equ' values hide the name of the parameter
;               in the .lst file.
;           (4) Add timing information to routine 'read_counters'.
;           (5) Change name of 'TLC' counter to the more-conventional name
;               'SDC'.
; 11Dec00   WIC, version 01.004 changes:
;           (1) Change some initial values for various tasks (from test vals
;               to more-operational values).
;           (2) Ensure 4th byte of FEC/DEC/SDC is zero in housekeeping.
; 12Dec00   WIC, version 01.005 changes:
;           (1) Calls to enable/disable Timer1 ISR were replaced with calls
;               to enable/disable Timer0 ISR (Timer1 is not used).
; 12Dec00   WIC, version 01.006 changes:
;           (1) Activity_ILIM failed to read ADC if HV was off; now fixed.
; 13Dec00   WIC, version 01.007 changes:  nothing significant.
; 13Dec00   WIC, version 01.008 changes:
;           (1) 'ticks_per_sec' = 50, vice 250
;           (2) VILIM persistence = 3 for both A and B segments
;           (3) ILIM persistence = 30 for AUX current
; 13Dec00   WIC, version 01.009 changes:
;           (1) Primary and Secondary command channels are now set to HIGH
;               interrupt priority -- thus allowing the Timer0 to be
;               interrupted.  Also, Primary and Secondary command ISRs will
;               now use register bank 3.
; 13Dec00   WIC, version 01.010 changes:
;           (1) 'ticks_per_sec' = 250 (back to original value)
;           (2) VILIM persistence = 15 for both A and B segments
;           (3) ILIM persistence = 150 for AUX current
; 14Dec00   WIC, version 01.011 changes:
;           (1) 'ACTIVITY_ILIM' no longer checks for the special case of
;               current reading = 255.
;           (2) 'ACTIVITY_ILIM' now shuts off grid power and clears HV settings
;               (VMAX, VSET, plus my XDATA params) when an over-current
;               shutoff occurs.
;           (3) VILIM persistence = 20 msec for both A and B segments
;           (4) ILIM persistence = 248 msec for AUX current
;           (5) Make diagnostics 2A, 2B, and 2E into HST errors by adding a
;               non-null parameter to the list (patchcst.a51).
;           (6) DCE_LFHVPWR command does not clear VMAX upon power ON but
;               still clears it on power OFF.
;           (7) 'ACTIVITY_ILIM' now uses ramping task (when a current-limit
;               overcount occurs) to set the HV to the lesser of VLOW and
;               VDAC (current DAC setting).
;           (8) 'ACTIVITY_CRP' ramps BOTH segments to VLOW when EITHER has
;               a CRP violation.
;           (9) Fix reading of PH data in housekeeping.
;           (10) Various initial values in the patchble constants changed.
; 14Dec00   WIC, version 01.012 changes:
;           (1) CRP intervals in housekeeping are 1-byte quantities with the
;               second byte being left as zero.
;           (2) 'safe_controls' now zeros VSET and VMAX for both Segments.
;           (3) Digitizer commands changed -- documentation wrong.  The upper
;               nibble of each digitizer command had to be increased by 1.
; 14Dec00   WIC, version 01.013 changes:
;           (1) Fixed 'DCE_LFGEWK' to use the 'DAC_ADDR' data value (like all
;               other digitizer commands).
; 14Dec00   WIC, version 01.014 changes:
;           (1) Removed call to 'INIT_HARDWARE' which initialized the
;               digitizer settings.  This routine will go into the next
;               version of BOOT.
;           (2) Write LED pattern to show OPERATE mode.
; 14Dec00   WIC, version 01.015 changes:
;           (1) Add ACTEL reset command, DCE_LFDRSTA, opcode = FA
; 15Dec00   WIC, version 01.016 changes:
;           (1) Added call to 'INIT_CRC' in 'DCE_LFDUPLOD' command to prevent
;               background CRC errors while uploading (mostly to conform
;               with Boot version 01.12 and later).
; 02Jan01   WIC, version 01.017 changes:
;           (1) LFHVPWR command clears circular buffers and histograms used
;               by the ILIM and CRP tasks when commanded to turn HV power ON.
; 08Jan01   WIC, version 01.018 changes:
;           (1) Timer0 ISR now calls DOOR_STOP when the 3-minute door timer
;               times out (formerly done in the door task -- only when AUX
;               power was enabled).
;           (2) In subroutine 'DCE_LFRACTRS', the logic level for indicating
;               that the door is actually unlatched was inverted.  Now, '0'
;               means latched and '1' is unlatched.
;           (3) Change comment in subroutine 'DCE_LFRMDIR' concerning
;               parameter 1, as follows:
;               PARAM1 = Door Direction  0..2 (0=Safe, 1=Open, 2=Close)
; 08Jan01   WIC, version 01.019 changes:
;           (1) Add diagnostic 10 if attempting to turn on either actuator
;               1 or 2 without first having enabled the actuators.
; 09Jan01   WIC, version 01.020 changes:
;           (1) Change polarity of door direction in the 'DCE_LFRACTRS' in
;               accordance with the change in version 1018.
; 09Jan01   WIC, version 01.021 changes:
;           (1) Commands 'DCE_LFRACT1' and 'DCE_LFRACT2' set the 3-minute door
;               timer to ensure a 'DOOR_STOP' is issued within the specified
;               countdown time.  'DOOR_STOP' also turns off actuators.
; 09Jan01   WIC, version 01.022 changes:
;           (1) Make parameter 'DOOR_TIMEOUT' a patchable constant.
; 09Jan01   WIC, version 01.023 changes:
;           (1) Update LFHSTATE variable in both 'safe_controls' and
;               'DCE_LFHVPWR' subroutines.
; 09Jan01   WIC, version 01.024 changes:
;           (1) In 'safe_controls', make the software voltage settings
;               match the hardware (i.e., LFHVTGTA, LFHVTGTB, LFHVDACA,
;               and LFHVDACB).
; 09Jan01   WIC, version 01.025 changes:
;           (1) Add function 'DCE_LFSAFE' which calls 'DOOR_STOP' and
;               'SAFE_CONTROLS'.  Opcode = 'DD'.
; 09Jan01   WIC, version 01.026 changes:
;           (1) Clear bit variable LFHVENA in routine 'SAFE_CONTROLS'.  This
;               disallows voltage commands until re-enabled.
;           (2) Clear bit variable LFHVENA in 'ACTIVITY_ILIM' when a shutdown
;               occurs.
; 10Jan01   WIC, version 01.027 changes:
;           (1) Status Bits 4 changed be CONTROLS2 (lockable register) and
;               Status Bits 5 changed to CONTROLS1 (not lockable).  See
;               subroutine 'SETUP_TO_BUILD_HKP'.
; 10Jan01   WIC, version 01.028 changes:
;           (1) All CRC regions in Upper 32KB have a delta CRC which forces
;               the CRC for the region to be 'C0DE'.
; 11Jan01   WIC, version 01.029 changes:
;           (1) Remove HST error parameters for diagnostics which do not exist
;               in Operate mode (leftovers from Boot).
;           (2) Diagnostic 22 is now an HST error.  Its error parameter is the
;               starting address of the CRP circular buffer for the segment
;               which caused the shutdown.
;           (3) Diagnostic 26 is now an HST error.  Its error parameter is the
;               Auxillary current reading at time of shutdown.
; 12Jan01   WIC, version 01.030 changes:
;           (1) Change the criteria by which calls to 'ACTIVITY_DOOR' are
;               made (see Timer0 ISR).
; 17Jan01   WIC, version 01.031 changes:
;           (1) Housekeeping now reports the full 16-bit value of the
;               LFHRAMPT parameter (vice only the lower 8 bits).
; 25Jan01   WIC, version 01.032 changes:
;           (1) Because a call to 'CLEAR_RAM' with length 0000 will clear
;               all 65,536 bytes of XDATA, the LFPCRP command fails if the
;               interval is set to '0' because it tries to clear 3*0 = 0 bytes
;               and actually clears all of XDATA.  Change LFPCRP command to
;               NOT clear RAM when LFPINT = 0 is commanded.
;*****************************************************************************

$SET (OPERATE_MAIN=1)
$INCLUDE(common.inc)

; Comment-out the following line for the flight build ...
;$SET (INCLUDE_TEST_CODE=1)

;*****************************************************************************
;*****************************************************************************
;                         Allocate Internal 8051 RAM
;*****************************************************************************
;*****************************************************************************
   DSEG  at  30h   ; begin immediately after the bit-addressable registers
CTR_TICK:          DS   2   ; counts the Timer0 rollovers
CMDBUFF:           DS   4   ; used for command packet handling
HKPBUFF:           DS   4   ; used while sending packets to CS computer
tempDPTR:          DS   2
maxSP:             DS   1   ; maximum Stack Pointer value
; CRC Cyclic Redundancy Check calculation variables ...
CRCH:              DS   1   ; CRC high byte
CRCL:              DS   1   ; CRC low byte
DXL:               DS   1   ; CRC register
DXH:               DS   1   ; CRC register
TEMP:              DS   1   ; Temporary register
TEMP1:             DS   1   ; Temporary register
; CRC Cyclic Redundancy Check calculation variables (LFDCRC and background)...
MEMPTR:            DS   2   ; pointer to next address for CRC calculation
CRC_LEN:           DS   2   ; number of bytes remaining in background CRC
CRCHa:             DS   1   ; CRCH stored here while process interrupted
CRCLa:             DS   1   ; CRCL stored here while process interrupted
REGION_CRC:        DS   1   ; current CRC region number (for background task)
; Housekeeping Packet background task
SCRIPT_PTR:        DS   2   ; pointer to next housekeeping script entry
HKP_PTR:           DS   2   ; pointer to next housekeeping location to write
HKP_PARAMS:        DS   4   ; holds parameters for multi-item readings
; Operate mode additional data items ...
DAC_ADDR:          DS   1   ; D/A converter address on either TDC
DOOR_TIMER:        DS   1   ; Three-minute door activity timeout
ILIM_STATE:        DS   1   ; ILIM task state variable
ILIM_COUNTDOWN:    DS   1   ; countdown after HV turned off
HVRAMP_COUNTDOWN:  DS   1   ; schedules HV ramp task
TIMER_COUNTDOWN:   DS   1   ; schedules the sub-second counter

   BSEG  at  00h   ; start at base of bit-addressable registers
; These bits are the event flags which control which tasks may be run
; within the main executive loop and which tasks can interrupt a background
; task ...
RXD_FLAG:          DBIT  1  ; 1 = command opcode received
SEND_DUMP_FLAG:    DBIT  1  ; 1 = dump packet queued to send
SEND_HKP_FLAG:     DBIT  1  ; 1 = housekeeping packet queued to send
HKP_BUILD_FLAG:    DBIT  1  ; 1 = building housekeeping packet
CRC_FLAG:          DBIT  1  ; 1 = background CRC task enabled
; These bits control the operation of various functions ...
SRC_FLAG:          DBIT  1  ; memory source for copy (0=XDATA, 1=CODE)
REGION_FLAG:       DBIT  1  ; memory source for backgnd CRC (0=XDATA, 1=CODE)
CALC_CRC:          DBIT  1  ; 1 = calculate CRC for upload packet
NOP_FLAG:          DBIT  1  ; 1 = LFDNOOP command received
CMD_BEGUN:         DBIT  1  ; 1 = Operate command was started
EXE_FLAG:          DBIT  1  ; 1 = Operate command fully executed
WDOG_ENABLED:      DBIT  1  ; 1 = Watchdog enabled
; Operate mode additional data items ...
LFHVENA:           DBIT  1  ; 1 = HV enabled, 0 = HV disabled
LFRMENA:           DBIT  1  ; 1 = Motor enabled, 0 = disabled
LFHRMPA:           DBIT  1  ; 1 = HV ramping on Segment A, 0 = not ramping A
LFHRMPB:           DBIT  1  ; 1 = HV ramping on Segment B, 0 = not ramping B
DOOR_LATCH:        DBIT  1  ; previous value of door latch status
USE_SEGMENT:       DBIT  1  ; 0 = Segment A, 1 = Segment B
;-----------------------------------------------------------------------------


;*****************************************************************************
;*****************************************************************************
;                             Boot-to-Operate ICD
;*****************************************************************************
;*****************************************************************************
   ; BIT variable "IVEC" determines whether interrupts use the BOOT space
   ; or OPERATE space.
   ; If in BOOT (IVEC=0), ISRs jump to the local routines (in PROM).
   ; If in OPERATE (IVEC=1), ISRs jump to offsets 8003h, 800Bh, etc.
   ; to look there for a jump to the actual OPERATE mode ISR.  This
   ; facilitates the parallel development of BOOT and OPERATE with minimal
   ; interface constraints.
   BSEG  at  07Fh
IVEC:              DBIT  1  ; 0=Boot mode ISRs, 1=Operate mode ISRs

$if INCLUDE_TEST_CODE
   ; The below code is enough of the "boot" code to get to operate mode ...
   ; Code start ...
   CSEG  at  0000h
   ljmp  8000h
   ; ISR jumps ...
   CSEG  at  0003h
   ljmp  8003h
   CSEG  at  000Bh
   ljmp  800Bh
   CSEG  at  0013h
   ljmp  8013h
   CSEG  at  001Bh
   ljmp  801Bh
   CSEG  at  0023h
   ljmp  8023h
   CSEG  at  002Bh
   ljmp  802Bh
   CSEG  at  0033h
   ljmp  8033h
$endif
;-----------------------------------------------------------------------------


;*****************************************************************************
;*****************************************************************************
;         Operate Reset Vector and Interrupt Service Routine Vectors
;*****************************************************************************
;*****************************************************************************
   CSEG  at  8000h              ; Reset code must start at 8000h for Operate
reg3      equ   $

   ; Jump to the operate entry point ...
   ljmp   OPERATE_MAIN

   ; Primary Command-Received ISR ...
   CSEG  at  8003h
   ljmp   ISR_Pri_Cmd_Received  ; Operate mode ISR 03

   ; Timer0 ISR -- executes at a regular interval for main loop timing ...
   CSEG  at  800Bh
   ljmp   ISR_Timer0            ; Operate mode ISR 0B

   ; Secondary Command-Received ISR ...
   CSEG  at  8013h
   ljmp   ISR_Sec_Cmd_Received  ; Operate mode ISR 13

   ; Timer1 ISR -- Not Used
   CSEG  at  801Bh
   ljmp   ISR_Timer1            ; Operate mode ISR 1B

   ; Serial Port ISR -- Not Used
   CSEG  at  8023h
   ljmp   ISR_Serial_Port       ; Operate mode ISR 23

   ; Timer2 ISR -- Not Used
   CSEG  at  802Bh
   ljmp   ISR_Timer2            ; Operate mode ISR 2B

   ; PCA ISR -- Not Used
   CSEG  at  8033h
   ljmp   ISR_PCA               ; Operate mode ISR 33
;-----------------------------------------------------------------------------

   CSEG  at  8040h

;*****************************************************************************
;*****************************************************************************
;               Operate Mode Interrupt Service Routines (ISRs)
;                           (use Register Bank 1)
;*****************************************************************************
;*****************************************************************************
; Subroutine -- ISR_Pri_Cmd_Received
; Purpose -- Handles External interrupt 0 which is invoked when a 32-bit
;            command word is present in its buffer and ready for reading by
;            the 8051.  This is the PRIMARY command channel interrupt.
; Inputs -- none
; Outputs -- command data value written to memory
; Timing -- 88 machine cycles (0.132 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ISR_Pri_Cmd_Received:
   ISR_SAVE_STATE
   mov   PSW, #(3 SHL 3)    ; use register bank 3

   ; Setup DPTR and fetch command MSB from registers ...
   mov   DPTR, #RXD_HARDWARE + 300h
   lcall RXD_READ           ; move data to command buffer

   ISR_RESTORE_STATE
   reti
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ISR_Sec_Cmd_Received
; Purpose -- Handles External interrupt 1 which is invoked when a 32-bit
;            command word is present in its buffer and ready for reading by
;            the 8051.  This is the SECONDARY command channel interrupt.
; Inputs -- none
; Outputs -- command data value written to memory
; Timing -- 88 machine cycles (0.132 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ISR_Sec_Cmd_Received:
   ISR_SAVE_STATE
   mov   PSW, #(3 SHL 3)    ; use register bank 3

   ; Setup DPTR and fetch command MSB from registers ...
   mov   DPTR, #RXD_HARDWARE + 700h
   lcall RXD_READ           ; move data to command buffer

   ISR_RESTORE_STATE
   reti
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ISR_Timer0
; Purpose -- Timer0 interrupts occur at a rate of "ticks_per_sec" times per
;            second.  This ISR decrements a counter which, upon reaching zero,
;            indicates that one second has elapsed.  At that time, a 4-byte
;            seconds counter is incremented.  Timer0 is the base timer for the
;            PCA's watchdog timer (allowing it the full 10-second timeout).
;            Timer0 is setup in the 16-bit counter mode and is reloaded each
;            call to a period of '1/ticks_per_sec' seconds.
; Inputs -- none
; Outputs -- mLFCTIME (4-bytes) incremented each second.
; Timing -- 
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ISR_Timer0:
hvramp_per_sec  equ   10    ; HV ramp call rate (calls per second)
sub_second      equ   50    ; timer sub-second rate (ticks per second)
ticks_per_sec   equ   250   ; ILIM task call rate (calls per second)
clock_freq      equ   8000000
latency         equ   18h   ; number of machine cycles before timer update
count_value     equ   (65536 - clock_freq/12/ticks_per_sec + latency)
   ISR_SAVE_STATE
   mov   PSW, #(1 SHL 3)    ; use register bank 1

   mov   TL0, #0            ; prevent rollover while loading timer
   ; load '#HIGH(count_value)' ...
   mov   TH0, #HIGH(count_value)
   ; load '#LOW(count_value)' ...
   mov   TL0, #LOW(count_value)

   ; Monitor stack pointer growth (SP > maxSP) ...
   clr   C                  ; prepare for SUBB operation
   mov   A, maxSP           ; look at stack pointer for growth
   subb  A, SP              ; ACC = maxSP - SP
   jnc   ISR_Timer0_10      ; jump if SP <= maxSP
   mov   maxSP, SP          ; store new (higher) stack pointer value

   ; Call all the tasks whose timing intervals have expired ...

   ;--------------------------------------------------------------------------
   ; Current limit protection task called 'ticks_per_sec' times per second ...
ISR_Timer0_10:
   ; If HV is ON (or post-shutdown counter not expired), call
   ; current limit protection task ...
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR
   jb    ACC.0, ISR_Timer0_101  ; jump to ILIM activity if HV ON
   mov   A, ILIM_COUNTDOWN  ; post-shutdown counter
   jz    ISR_Timer0_102     ; jump if no longer collecting data
   dec   ILIM_COUNTDOWN
ISR_Timer0_101:
   lcall ACTIVITY_ILIM      ; call the ILIM activity
ISR_Timer0_102:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
   ; HV ramping task called 'hvramp_per_sec' times per second ...
ISR_Timer0_20:
   djnz  HVRAMP_COUNTDOWN, ISR_Timer0_202
   ; load '#ticks_per_sec/hvramp_per_sec' ...
   mov   HVRAMP_COUNTDOWN, #ticks_per_sec/hvramp_per_sec
   jnb   LFHRMPA, ISR_Timer0_201  ; jump if NOT ramping Segment A
   clr   USE_SEGMENT              ; 0 = Segment A, 1 = Segment B
   lcall ACTIVITY_HV_RAMP
ISR_Timer0_201:
   jnb   LFHRMPB, ISR_Timer0_202  ; jump if NOT ramping Segment B
   setb  USE_SEGMENT              ; 0 = Segment A, 1 = Segment B
   lcall ACTIVITY_HV_RAMP
ISR_Timer0_202:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
   ; sub-second timer increments 'sub_second' times per second ...
ISR_Timer0_30:
   djnz  TIMER_COUNTDOWN, ISR_Timer0_301
   ; load '#ticks_per_sec/sub_second' ...
   mov   TIMER_COUNTDOWN, #ticks_per_sec/sub_second
   ; increment 'mTICKS' sub-second counter ...
   mov   DPTR, #mTICKS
   movx  A, @DPTR
   inc   A
   movx  @DPTR, A
ISR_Timer0_301:
   ;--------------------------------------------------------------------------

   ; All activities below occur only once per second ...

   ; Check for expiration of the one-second interval ...
   djnz  CTR_TICK+0, ISR_Timer0_99  ; decrement counter LSB
   djnz  CTR_TICK+1, ISR_Timer0_99  ; decrement counter MSB

   ; one-second interval expired --> reset count-down register ...
   ; load '#LOW(ticks_per_sec)' ...
   mov   CTR_TICK+0, #LOW(ticks_per_sec)
$if  LOW(ticks_per_sec)
   ; load '#HIGH(ticks_per_sec)+1' ...
   mov   CTR_TICK+1, #HIGH(ticks_per_sec)+1
$else
   ; load '#HIGH(ticks_per_sec)' ...
   mov   CTR_TICK+1, #HIGH(ticks_per_sec)
$endif

   ; Increment seconds counter ...
   mov   DPTR, #mLFCTIME
   lcall INCREMENT_LWORD

   ; Display timer LSB on LEDs ...
   mov   DPTR, #mLFCTIME
   movx  A, @DPTR
   mov   DPTR, #LED
   movx  @DPTR, A

   ; Clear the sub-second timer on the one-second interval ...
   mov   DPTR, #mTICKS
   clr   A
   movx  @DPTR, A

   ; Read Front-end counters HERE ONLY, since reading them clears them and
   ; multiple code portions use the values ...
   lcall read_counters

   ; decrement door activity 3-minute timer, if non-zero ...
ISR_Timer0_40:
   mov   A, DOOR_TIMER
   jz    ISR_Timer0_401
   djnz  DOOR_TIMER, ISR_Timer0_401
   lcall DOOR_STOP          ; Perform 'Door Stop' on reaching zero
ISR_Timer0_401:

   ;--------------------------------------------------------------------------
   ; Count Rate Protection task ...
ISR_Timer0_50:
   ; Must have HV ON for this task (CONTROLS1.0 = '1') ...
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR
   jnb   ACC.0, ISR_Timer0_501  ; jump if HV OFF
   clr   USE_SEGMENT        ; 0 = Segment A, 1 = Segment B
   lcall ACTIVITY_CRP
   setb  USE_SEGMENT        ; 0 = Segment A, 1 = Segment B
   lcall ACTIVITY_CRP
ISR_Timer0_501:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
   ; Door Protection task ...
ISR_Timer0_60:
   ; Call 'ACTIVITY_DOOR' if ANY of the following conditions is true ...
   ; (1) Motor commands enabled (LFRMENA = '1') is true ...
   jb    LFRMENA, ISR_Timer0_601
   ; (2) Motor power is on (CONTROLS1.5 = '1') ...
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR
   jb    ACC.5, ISR_Timer0_601
   ; (3) Actuators Enabled (CONTROLS2.4 = '1') ...
   mov   DPTR, #CONTROLS2
   clr   C2_LOCK            ; enable reading of CONTROLS2
   movx  A, @DPTR
   setb  C2_LOCK
   jb    ACC.4, ISR_Timer0_601
   sjmp  ISR_Timer0_602     ; skip 'ACTIVITY_DOOR'
ISR_Timer0_601:
   lcall ACTIVITY_DOOR
ISR_Timer0_602:
   ;--------------------------------------------------------------------------

$if INCLUDE_TEST_CODE
   ; If we are testing software, this call will inject a command ...
   lcall test_commands
$endif

ISR_Timer0_99:
   ISR_RESTORE_STATE
   reti
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ISR_Timer1
; Purpose -- Unused
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ISR_Timer1:
   NOP6                     ; padding for future patch jumps
   mov   DPTR, #mBADISR
   mov   A, #1Bh            ; Address of ISR is 801Bh
   movx  @DPTR, A
   inc   DPTR
   mov   A, #80h
   movx  @DPTR, A
   mov   A, #LOW(DIAG0031)  ; Diagnostic:  Attempt to execute nonexistent ISR
   lcall report_diagnostic
   reti
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ISR_Serial_Port
; Purpose -- Unused
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ISR_Serial_Port:
   NOP6                     ; padding for future patch jumps
   mov   DPTR, #mBADISR
   mov   A, #23h            ; Address of ISR is 8023h
   movx  @DPTR, A
   inc   DPTR
   mov   A, #80h
   movx  @DPTR, A
   mov   A, #LOW(DIAG0031)  ; Diagnostic:  Attempt to execute nonexistent ISR
   lcall report_diagnostic
   reti
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ISR_Timer2
; Purpose -- Unused
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ISR_Timer2:
   NOP6                     ; padding for future patch jumps
   mov   DPTR, #mBADISR
   mov   A, #2Bh            ; Address of ISR is 802Bh
   movx  @DPTR, A
   inc   DPTR
   mov   A, #80h
   movx  @DPTR, A
   mov   A, #LOW(DIAG0031)  ; Diagnostic:  Attempt to execute nonexistent ISR
   lcall report_diagnostic
   reti
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ISR_PCA
; Note -- The watchdog timer is implemented using the Programmable Counter
;         Array (PCA) module.  This function does not require an interrupt
;         to function.
; Purpose -- Unused
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ISR_PCA:
   NOP6                     ; padding for future patch jumps
   mov   DPTR, #mBADISR
   mov   A, #33h            ; Address of ISR is 8033h
   movx  @DPTR, A
   inc   DPTR
   mov   A, #80h
   movx  @DPTR, A
   mov   A, #LOW(DIAG0031)  ; Diagnostic:  Attempt to execute nonexistent ISR
   lcall report_diagnostic
   reti
;-----------------------------------------------------------------------------


;*****************************************************************************
;*****************************************************************************
;                      Operate-Mode ISR support routines
;*****************************************************************************
;*****************************************************************************
; Subroutine -- RXD_READ
; Purpose -- Read command from I/O registers.  If command is 80000000h,
;            force a Power-on reset.  Otherwise, if command address is in
;            the range 0040h-045Ah, store the 2 command data bytes in the
;            given address.
; Note -- This routine is only called by "ISR_Pri_Cmd_Received" and
;         "ISR_Sec_Cmd_Received".
; Inputs -- DPTR set to
;             "RXD_HARDWARE+300h" for Primary command receive channel
;                  (external interrupt 0) or
;             "RXD_HARDWARE+700h" for Secondary command channel receive
;                  (external interrupt 1).
;           I.E., DPTR points to MSB of incoming command for the appropriate
;           command channel.
; Outputs -- System reset if command is 80000000h.  Command data stored in
;            address provided if within proper range.  Flag 'RXD_FLAG' set to
;            '1' if a process command was received (i.e. command with address
;            0440h)
; Timing -- See timing for the ISRs which call this routine.
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
RXD_READ:
   movx  A,@DPTR            ; read command MSB (bits 31..24)
   mov   CMDBUFF+3,A        ; store in internal RAM
   dec   DPH                ; point to next command byte (bits 23..16)
   movx  A,@DPTR            ; read command byte
   mov   CMDBUFF+2,A        ; store bits 23..16
   dec   DPH
   movx  A,@DPTR
   mov   CMDBUFF+1,A        ; store bits 15..8
   dec   DPH
   movx  A,@DPTR            ; read LSB (this also clears hardware BUSY line)
   mov   CMDBUFF+0,A        ; store bits 7..0

   ; If command MS-bit is '1' and any of the other bits 31 bits are
   ; non-zero, command word is invalid ...
   ; If command MS-bit is '1' and ALL other bits are zero, force a power-on
   ; reset ...
   ; Otherwise (MS-bit is '0'), process the command word ...
RXD_READ_00:
   mov   A, CMDBUFF+3          ; command MSB
   jnb   ACC.7, RXD_READ_10    ; MS-bit is '0' --> OK
   anl   A, #7Fh               ; mask off MS-bit
   jnz   RXD_READ_90           ; other bits non-zero --> invalid
   mov   A, CMDBUFF+2          ; check next byte
   jnz   RXD_READ_90           ; non-zero bits --> invalid
   mov   A, CMDBUFF+1          ; check next byte
   jnz   RXD_READ_90           ; non-zero bits --> invalid
   mov   A, CMDBUFF+0          ; check next byte
   jnz   RXD_READ_90           ; non-zero bits --> invalid
   ljmp  DCE_LFDRSTP           ; Cmd = 80000000h --> power-on reset.

RXD_READ_10:    ; Valid dword.  Perform address range checking ...

   ; Valid range for writing to memory is 0040h-045Ah (memory upload and
   ; command packet).
   ; ACC holds MSB of command.  Mask off most-sig 3 bits (since they are
   ; "unused") and check range.
   ; Subtract 0040h from CMDBUFF+3:CMDBUFF+2 to ensure that address is
   ; greater than or equal to 0040h ...
   clr   C                  ; prepare for SUBB operation
   mov   A, CMDBUFF+2       ; lower address byte
   subb  A, #LOW(0040h)
   mov   A, CMDBUFF+3       ; upper address byte
   anl   A, #1Fh            ; mask off most-sig 3 bits (unused bits)
   mov   CMDBUFF+3, A       ; store value to preclude need to mask later
   subb  A, #HIGH(0040h)
   jc    RXD_READ_90        ; address less than 0040h -- too low
   
   ; Subtract 045Bh from CMDBUFF+3:CMDBUFF+2 to ensure that address is
   ; less than or equal to 045Ah.
   ; Note:  Carry flag is already clear.
   mov   A, CMDBUFF+2       ; address LSB
   subb  A, #LOW(045Bh)
   mov   A, CMDBUFF+3       ; address MSB
   subb  A, #HIGH(045Bh)
   jnc   RXD_READ_90        ; address greater than 045Ah -- too high

   ; address is in the range 0040h - 045Ah, so store the least-sig 2 bytes
   ; of the command (i.e., the data bytes) in memory ...
   
   mov   A, CMDBUFF+3       ; Put address in DPTR
   mov   DPH, A             ; MSB
   mov   A, CMDBUFF+2
   mov   DPL, A             ; LSB
   ; now store the 2 data bytes ...
   mov   A, CMDBUFF+0       ; data LSB
   movx  @DPTR, A
   inc   DPTR
   mov   A, CMDBUFF+1       ; data MSB
   movx  @DPTR, A

   ; See if a command opcode was received (address = RS17)
   setb  RXD_FLAG           ; ASSUME an command opcode received
   mov   DPTR, #RS17        ; address of command opcode LSB
   movx  A, @DPTR
   jnz   RXD_READ_99        ; opcode was received
   clr   RXD_FLAG           ; opcode NOT received
   sjmp  RXD_READ_99        ; done
   
RXD_READ_90:
   ; Increment the "command word received but not stored" counter ...
   mov   DPTR, #mCWNS
   lcall INCREMENT_WORD
   ; Store the "unstored" command word ...
   mov   DPTR, #mWNS
   mov   A, CMDBUFF+0       ; data LSB
   movx  @DPTR, A
   inc   DPTR
   mov   A, CMDBUFF+1       ; data MSB
   movx  @DPTR, A
   inc   DPTR
   mov   A, CMDBUFF+2       ; address LSB
   movx  @DPTR, A
   inc   DPTR
   mov   A, CMDBUFF+3       ; address MSB
   movx  @DPTR, A

; clean up and exit
RXD_READ_99:
   ret                      ; RXD_READ
;-----------------------------------------------------------------------------

FILL_TO_ADDRESS_CRC    8500h, 0FFh, 00543h
reg4      equ   $
reg3len   equ   $-reg3

;*****************************************************************************
;*****************************************************************************
;                         Operate-Mode Main Executive
;*****************************************************************************
;*****************************************************************************
; Subroutine -- OPERATE_MAIN
; Purpose -- This is the entry for operate code.  Upon any type of reset, the
;            first instruction is to jump to this routine.  Here various
;            internal and external RAM locations are initialized, timers and
;            interrupts are initialized, and finally, a JUMP (no return) is
;            made to the main executive loop.
; Inputs -- none
; Outputs -- none
; Timing -- ~58,000 machine cycles (87.0 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
OPERATE_MAIN:
   ; Initialize the processor ...
   mov   IE, #0             ; disable all interrupts
   mov   IP, #0             ; all interrupts at same proiority level
   mov   TCON, #0           ; stop timers
   mov   SP, #_STACK        ; Reset stack to base level

$if INCLUDE_TEST_CODE
   ; Copy CODE space to XDATA to simulate actual system in which both use
   ; a shared space ...
   ; Source address in R3:R2 ...
   mov   R3, #80h
   mov   R2, #00h
   ; Destination address in R5:R4 ...
   mov   R5, #80h
   mov   R4, #00h
   ; Number of bytes to copy in R0:R7 ...
   mov   R0, #40h
   mov   R7, #00h
   ; Source Memory space is CODE ...
   setb  SRC_FLAG
   ; Now perform the copy ...
   lcall MEM_COPY
$endif

   ; Disable the Programmable Counter Array (PCA) watchdog timer ...
   lcall DISABLE_DOG

   ; Safe the hardware ...
   lcall SAFE_CONTROLS

   ; Clear Internal and External RAM ...
   lcall INIT_RAM

   ;--------------------------------------------------------------------------
   ; Setup interrupts ...
   setb  IVEC               ; ensure Operate mode ISRs are used
   setb  EA                 ; global interrupt enable

   ; Setup standard 8051 Timers ...
   mov   TMOD, #11h         ; Timer0 in 16-bit timer mode
                            ; Timer1 in 16-bit timer mode
   ; Pre-set Timer0 count-down register ...
   ; load '#LOW(ticks_per_sec)' ...
   mov   CTR_TICK+0, #LOW(ticks_per_sec)
$if  LOW(ticks_per_sec)
   ; load '#HIGH(ticks_per_sec)+1' ...
   mov   CTR_TICK+1, #HIGH(ticks_per_sec)+1
$else
   ; load '#HIGH(ticks_per_sec)' ...
   mov   CTR_TICK+1, #HIGH(ticks_per_sec)
$endif
   setb  TR0                ; start Timer0
   setb  ET0                ; enable Timer0 interrupt
   setb  TF0                ; force immediate call to Timer0

   ; Setup Primary command interface ...
   setb  IP.0               ; interrupt priority HIGH
   clr   IT0                ; interrupt is low-level triggered
   setb  EX0                ; enable interrupt
   
   ; Setup Secondary command interface ...
   setb  IP.2               ; interrupt priority HIGH
   clr   IT1                ; interrupt is low-level triggered
   setb  EX1                ; enable interrupt
   ;--------------------------------------------------------------------------

   ; Start background CRC calculations ...
   mov   REGION_CRC, #1     ; Start background CRC checking with PROM
   lcall INIT_CRC           ; initialize internal variables for this region
   lcall CLEAR_OLD_CRCS     ; clear all 'previous CRC' values to prevent CRC
                            ; comparison errors (initially)

   ; Setup for housekeeping packet construction ...
   lcall SETUP_TO_BUILD_HKP

   ljmp  OPERATE_MAIN_EXECUTIVE ; start main executive loop
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- OPERATE_MAIN_EXECUTIVE
; Purpose -- Main Executive loop cyclically calls all task subroutines.
; Inputs -- none
; Outputs -- none
; Timing -- forever (until reset)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
OPERATE_MAIN_EXECUTIVE:

   mov   A, SP
   xrl   A, #_STACK         ; see if SP = _STACK
   jz    OPERATE_MAIN_EXECUTIVE_10 ; jump if OK
   mov   A, #LOW(DIAG0017)  ; Diagnostic:  Stack Growth
   lcall report_diagnostic
   mov   SP, #_STACK        ; reset the stack pointer

OPERATE_MAIN_EXECUTIVE_10:

   ; Increment the performance monitor ...
   mov   DPTR, #mLFDPERF
   lcall INCREMENT_WORD

   ;--------------------------------------------------------------------------
   ; Process a command packet ...
   ; Timing -- 1040 machine cycles (1.56 msec @ 8-MHz) plus DCE command time.
TASK_PROCESS_COMMAND:
   ; Reasons to skip this task for this loop ...
   jnb   RXD_FLAG, TASK_PROCESS_COMMAND_99 ; no opcode received
   ; No reason to skip, so enter the task ...
   clr   RXD_FLAG            ; clear the flag
   lcall PROCESS_COMMAND     ; process the command
TASK_PROCESS_COMMAND_99:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
   ; If dump packet queued, send it.
   ; Timing -- 22,642 machine cycles (34.0 msec @ 8-MHz).
TASK_SEND_DUMP_PKT:
   ; Reasons to skip this task for this loop ...
   jnb   SEND_DUMP_FLAG, TASK_SEND_DUMP_PKT_99 ; dump packet not queued
   jb    RXD_FLAG,       TASK_SEND_DUMP_PKT_99 ; opcode received
   ; No reason to skip, so enter the task ...
   clr   SEND_DUMP_FLAG      ; clear the flag
   lcall SEND_DUMP_PKT       ; send the packet
TASK_SEND_DUMP_PKT_99:
   ;--------------------------------------------------------------------------
   
   ;--------------------------------------------------------------------------
   ; If housekeeping packet queued, send it.
   ; Timing -- 22,642 machine cycles (34.0 msec @ 8-MHz).
TASK_SEND_HKP_PKT:
   ; Reasons to skip this task for this loop ...
   jnb   SEND_HKP_FLAG,  TASK_SEND_HKP_PKT_99 ; housekeeping packet not queued
   jb    RXD_FLAG,       TASK_SEND_HKP_PKT_99 ; opcode received
   jb    SEND_DUMP_FLAG, TASK_SEND_HKP_PKT_99 ; dump packet queued
   ; No reason to skip, so enter the task ...
   clr   SEND_HKP_FLAG       ; clear the flag
   lcall SEND_HKP_PKT        ; send the packet
TASK_SEND_HKP_PKT_99:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
   ; Build Housekeeping packet ...
   ; Timing -- ~46,900 machine cycles (70.4 msec @ 8-MHz) minimum.
TASK_BUILD_HKP:
   ; Reasons to skip this task for this loop ...
   jnb   HKP_BUILD_FLAG, TASK_BUILD_HKP_99 ; not flagged to build HKP
   jb    RXD_FLAG,       TASK_BUILD_HKP_99 ; opcode received
   jb    SEND_DUMP_FLAG, TASK_BUILD_HKP_99 ; dump packet queued
   jb    SEND_HKP_FLAG,  TASK_BUILD_HKP_99 ; houskeeping packet queued !!
   ; No reason to skip, so enter the task ...
   ;    NOTE:  This is a BACKGROUND task (other events interrupt it), so it
   ;    will clear appropriate flags when it is complete.
   lcall BUILD_HKP          ; Build housekeeping from the script

   ; When finished building housekeeping, clear command buffer to prepare
   ; for next command to arrive ...
   jb    HKP_BUILD_FLAG, TASK_BUILD_HKP_99 ; STILL flagged to build HKP
   lcall CLEAR_CMD_PKT      ; clear 14-word command buffer
   lcall CLEAR_HKP_PKT      ; clear PARTS of the housekeeping work area
TASK_BUILD_HKP_99:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
   ; Compute background CRC ...
   ; Timing -- 52,336 machine cycles for 1024 bytes (78.5 msec @ 8-MHz).
   ;   ISR interruptions add enter/exit overhead.  Worst case w/o interrupts
   ;   is 5.02 seconds for 65,536 bytes @ 8-MHz.
TASK_BG_CRC:
   ; Reasons to skip this task for this loop ...
   jnb   CRC_FLAG,       TASK_BG_CRC_99 ; CRC background task NOT flagged
   jb    RXD_FLAG,       TASK_BG_CRC_99 ; opcode received
   jb    SEND_DUMP_FLAG, TASK_BG_CRC_99 ; dump packet queued
   jb    SEND_HKP_FLAG,  TASK_BG_CRC_99 ; housekeeping packet queued
   ; No reason to skip, so enter the task ...
   ;    NOTE:  This is a BACKGROUND task (other events interrupt it), so it
   ;    will clear appropriate flags when it is complete.
   lcall BG_CRC              ; background CRC task
TASK_BG_CRC_99:
   ;--------------------------------------------------------------------------

   ; Done with tasks for this loop.  Back to the top.
OPERATE_MAIN_EXECUTIVE_99:
   ljmp  OPERATE_MAIN_EXECUTIVE ; back to top of the loop
;-----------------------------------------------------------------------------


;*****************************************************************************
;*****************************************************************************
;                    Main-Executive-Loop Support Routines
;*****************************************************************************
;*****************************************************************************
; Subroutine -- PROCESS_COMMAND
; Purpose -- Process a command packet.  Called when RXD_FLAG=1 in main exec.
; Inputs -- Command packet in RAM (buffer address 'RS17')
; Outputs -- Command checked for correctness, executed if correct.  Counters
;            incremented, memory monitors updated, housekeeping packet script
;            initialized for building housekeeping packet.
; Timing -- 1040 machine cycles (1.56 msec @ 8-MHz), includes all subroutine
;           calls from this routine, EXCLUDING the time required by the
;           particular Operate-mode function being processed.
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
PROCESS_COMMAND:
   ; Verify command packet conforms to proper format ...
   lcall CHECK_CMD_PACKET   ; return value:  ACC=0 if command packet OK
   jnz   PROCESS_COMMAND_30 ; bad command packet -- send housekeeping
   
   ; Good command packet received ...
   ; If watchdog enabled, feed the dog ...
   jnb   WDOG_ENABLED, PROCESS_COMMAND_10  ; jump if watchdog NOT enabled
   lcall ENABLE_DOG         ; enable AND feed watchdog

PROCESS_COMMAND_10:
   ; Execute the command ...
   lcall COMMAND_VECTOR

   ; Increment packet counter ...
   mov   DPTR, #mLFCPKTL
   lcall INCREMENT_LWORD

   ; Check for LFDNOOP command ...
   jbc   NOP_FLAG, PROCESS_COMMAND_30  ; jump if command is LFDNOOP

   ; Increment commands-received counter ...
   mov   DPTR, #mLFDCMDR
   lcall INCREMENT_WORD

   ; Check if command was executed ...
   jnb   EXE_FLAG, PROCESS_COMMAND_20  ; jump if command NOT executed
   clr   EXE_FLAG
   clr   CMD_BEGUN                     ; must clear "command begun" as well

   ; Increment commands-executed counter   
   mov   DPTR, #mLFDCMDX
   lcall INCREMENT_WORD
   
   ; record the last command executed ...
   mov   DPTR, #RS17        ; command opcode within command buffer
   movx  A, @DPTR
   mov   DPTR, #mLFDLSTC    ; store in 'mLFDLSTC'
   movx  @DPTR, A

   sjmp  PROCESS_COMMAND_30

PROCESS_COMMAND_20:
   jnb   CMD_BEGUN, PROCESS_COMMAND_30  ; jump if cmd never begun or illegal
   clr   CMD_BEGUN

   ; Report that the command did not complete ...
   mov   A, #LOW(DIAG0013)  ; Diagnostic:  Command not completed
   lcall report_diagnostic

PROCESS_COMMAND_30:
   clr   CMD_BEGUN          ; clear "command begun"

   ; Setup for housekeeping packet construction ...
   lcall SETUP_TO_BUILD_HKP

PROCESS_COMMAND_99:
   ret                      ; PROCESS_COMMAND
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ADC_READ
; Purpose -- Perform A/D conversion, returning value in ACC
; Inputs -- MUX already points to proper channel for ADC to read, delay
;           multiples in 'mMUX_DELAY' and 'mADC_DELAY'
; Outputs -- ADC reading in ACC
; Registers -- DPTR, ACC, R0, R1
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ADC_READ:
   ; Delay 'mMUX_DELAY' times 0.01 msec ...
   mov   DPTR, #mMUX_DELAY
   movx  A, @DPTR
   jz    ADC_READ_20        ; jump if delay = 0
   dec   A
   jz    ADC_READ_20        ; jump if delay = 1
   mov   R1, A
ADC_READ_10:
   mov   R0, #2
   djnz  R0, $
   djnz  R1, ADC_READ_10

ADC_READ_20:
   ; Start the ADC by writing ANYTHING to address 'ADC_SPACE' ...
   mov   DPTR, #ADC_SPACE
   movx  @DPTR, A

   ; Delay 'mADC_DELAY' times 0.01 msec ...
   mov   DPTR, #mADC_DELAY
   movx  A, @DPTR
   jz    ADC_READ_40        ; jump if delay = 0
   dec   A
   jz    ADC_READ_40        ; jump if delay = 1
   mov   R1, A
ADC_READ_30:
   mov   R0, #2
   djnz  R0, $
   djnz  R1, ADC_READ_30

ADC_READ_40:
   mov   DPTR, #ADC_SPACE
   movx  A, @DPTR           ; read the analog value

ADC_READ_99:
   ret                      ; ADC_READ
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- CLEAR_HKP_PKT
; Purpose -- Clear parts of the housekeeping packet (after packet is sent).
;            Note that the housekeeping packet is not really cleared, but,
;            rather, those working variables which are copied into the
;            housekeeping packet by the routine 'BUILD_HKP'.
; Inputs -- none
; Outputs -- Specific parts of housekeeping packet cleared (see below)
; Timing -- 80 machine cycles (0.120 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CLEAR_HKP_PKT:

   ; Clear array of 8 HST errors ...
   mov   DPTR, #mLFDERR     ; first XDATA location to clear
   mov   R6, #HIGH(8)       ; number of bytes to clear in R6:R7
   mov   R7, #LOW(8)
   lcall CLEAR_RAM          ; clear the memory

   ; Clear Performance monitor ...
   mov   DPTR, #mLFDPERF
   clr   A
   movx  @DPTR, A
   inc   DPTR
   movx  @DPTR, A

   ; Clear the number of diagnostics stored since last housekeeping sent
   mov   DPTR, #mCURR_NUM
   clr   A
   movx  @DPTR, A

   ; Clear the PH Historgram data in hardware ...
   setb  PHA_DIS            ; Enable writing of PH_A (by disabling ACTEL)
   mov   DPTR, #PH_A
   mov   R6, #HIGH(512)     ; 512 bytes handles both PH 'A' and 'B'
   mov   R7, #LOW(512)
   lcall CLEAR_RAM
   clr   PHA_DIS            ; Disable writing of PH_A (enabling ACTEL)

CLEAR_HKP_PKT_99:
   ret                      ; CLEAR_HKP_PKT
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- CLEAR_CMD_PKT
; Purpose -- Clear the 14 command words in order that commanding errors may
;            be detected and flagged.
; Inputs -- none
; Outputs -- 14 command packet words cleared
; Timing -- 186 machine cycles (0.279 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CLEAR_CMD_PKT:
   mov   DPTR, #RS17        ; first XDATA location to clear
   mov   R6, #HIGH(28)      ; number of bytes to clear in R6:R7
   mov   R7, #LOW(28)
   lcall CLEAR_RAM          ; clear the memory
CLEAR_CMD_PKT_99:
   ret                      ; CLEAR_HKP_PKT
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- read_counters
; Purpose -- Read FEC_A, FEC,B, DEC_A, DEC_B, SDC_1, and SDC_2 from I/O and
;            store as 24-bit values in XDATA.  Since multiple locations
;            within the code use the FEC readings and since reading the
;            hardware clears it, we read it once per second using this
;            routine.  The other counters are read here as well so that they
;            correlate in time with the FEC readings.  This routine does not
;            use the MEM_COPY function because of the need for speed (this
;            brute force method is 7 times faster than using MEM_COPY).
; Inputs -- none
; Outputs -- XDATA memory locations mFEC_A, mFEC_B, mDEC_A, mDEC_B, mSDC_1,
;            and mSDC_2 are written (little-endian).
; Timing -- 104 machine cycles to read counters    (0.156 msec @ 8-MHz)
;           YYY machine cycles total, read + write (0. msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
read_counters:

   ; Read the counters as fast as possible, push them onto stack, and only
   ; later put them into XDATA ...
   ; Read FEC_A ...
   mov   DPTR, #FEC_A
   movx  A, @DPTR           ; read LSB
   push  ACC
   inc   DPTR
   movx  A, @DPTR
   push  ACC
   inc   DPTR
   movx  A, @DPTR           ; read MSB (and clear register)
   push  ACC
   ; Read DEC_A ...
   mov   DPTR, #DEC_A
   movx  A, @DPTR           ; read LSB
   push  ACC
   inc   DPTR
   movx  A, @DPTR
   push  ACC
   inc   DPTR
   movx  A, @DPTR
   push  ACC
   ; Read SDC_1 ...
   mov   DPTR, #SDC_1
   movx  A, @DPTR           ; read LSB
   push  ACC
   inc   DPTR
   movx  A, @DPTR
   push  ACC
   inc   DPTR
   movx  A, @DPTR
   push  ACC

   ; Read FEC_B ...
   mov   DPTR, #FEC_B
   movx  A, @DPTR           ; read LSB
   push  ACC
   inc   DPTR
   movx  A, @DPTR
   push  ACC
   inc   DPTR
   movx  A, @DPTR
   push  ACC
   ; Read DEC_B ...
   mov   DPTR, #DEC_B
   movx  A, @DPTR           ; read LSB
   push  ACC
   inc   DPTR
   movx  A, @DPTR
   push  ACC
   inc   DPTR
   movx  A, @DPTR
   push  ACC
   ; Read SDC_2 ...
   mov   DPTR, #SDC_2
   movx  A, @DPTR           ; read LSB
   push  ACC
   inc   DPTR
   movx  A, @DPTR
   push  ACC
   inc   DPTR
   movx  A, @DPTR
   push  ACC

   ; Pop the items in a leisurely fashion and store them.  First off the stack
   ; is SDC_2 (MSB).
   ; Store SDC_2 ...
   mov   DPTR, #mSDC_2+3
   clr   A
   movx  @DPTR, A
   mov   DPTR, #mSDC_2+2
   pop   ACC
   movx  @DPTR, A           ; write MSB
   mov   DPTR, #mSDC_2+1
   pop   ACC
   movx  @DPTR, A
   mov   DPTR, #mSDC_2+0
   pop   ACC
   movx  @DPTR, A
   ; Store DEC_B ...
   mov   DPTR, #mDEC_B+3
   clr   A
   movx  @DPTR, A
   mov   DPTR, #mDEC_B+2
   pop   ACC
   movx  @DPTR, A           ; write MSB
   mov   DPTR, #mDEC_B+1
   pop   ACC
   movx  @DPTR, A
   mov   DPTR, #mDEC_B+0
   pop   ACC
   movx  @DPTR, A
   ; Store FEC_B ...
   mov   DPTR, #mFEC_B+3
   clr   A
   movx  @DPTR, A
   mov   DPTR, #mFEC_B+2
   pop   ACC
   movx  @DPTR, A           ; write MSB
   mov   DPTR, #mFEC_B+1
   pop   ACC
   movx  @DPTR, A
   mov   DPTR, #mFEC_B+0
   pop   ACC
   movx  @DPTR, A

   ; Store SDC_1 ...
   mov   DPTR, #mSDC_1+3
   clr   A
   movx  @DPTR, A
   mov   DPTR, #mSDC_1+2
   pop   ACC
   movx  @DPTR, A           ; write MSB
   mov   DPTR, #mSDC_1+1
   pop   ACC
   movx  @DPTR, A
   mov   DPTR, #mSDC_1+0
   pop   ACC
   movx  @DPTR, A
   ; Store DEC_A ...
   mov   DPTR, #mDEC_A+3
   clr   A
   movx  @DPTR, A
   mov   DPTR, #mDEC_A+2
   pop   ACC
   movx  @DPTR, A           ; write MSB
   mov   DPTR, #mDEC_A+1
   pop   ACC
   movx  @DPTR, A
   mov   DPTR, #mDEC_A+0
   pop   ACC
   movx  @DPTR, A
   ; Store FEC_A ...
   mov   DPTR, #mFEC_A+3
   clr   A
   movx  @DPTR, A
   mov   DPTR, #mFEC_A+2
   pop   ACC
   movx  @DPTR, A           ; write MSB
   mov   DPTR, #mFEC_A+1
   pop   ACC
   movx  @DPTR, A
   mov   DPTR, #mFEC_A+0
   pop   ACC
   movx  @DPTR, A

read_counters_99:
   ret                      ; read_counters
;-----------------------------------------------------------------------------

FILL_TO_ADDRESS_CRC    8A00h, 0FFh, 08867h
reg5      equ   $
reg4len   equ   $-reg4

;*****************************************************************************
;*****************************************************************************
;                       Boot- and Operate-Mode Commands
;                            (use Register Bank 2)
;*****************************************************************************
;*****************************************************************************
; Subroutine -- DCE_LFDNOOP (DCE Opcode '80'X)
; Purpose -- Force the DCE to send a housekeeping packet and feed the
;            watchdog timer without incrementing the command counter.
; Inputs -- none
; Outputs -- NOP_FLAG=1
; Timing -- 11 machine cycles (0.017 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDNOOP:
   NOP6                     ; padding for future patch jumps
   setb  NOP_FLAG           ; main executive loop looks at this flag
   ret                      ; DCE_LFDNOOP
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDMADDR (DCE Opcode '81'X)
; Purpose -- Set memory monitor address for one of the 8 memory monitors
; Inputs -- 3 command parameters:
;    PARAM1 = memory monitor index  0..7
;    PARAM2 = source address        0..FFFF
;    PARAM3 = memory space          0..2 (0=XDATA, 1=CODE, 2=DATA)
; Parameter Range Checking --
;    (1) PARAM1 is masked modulo 8
;    (2) PARAM2 none
;    (3) PARAM3 is masked module 4 and value 3 is rejected
; Outputs -- Memory monitor parameters changed
; Timing -- 131 machine cycles (0.197 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDMADDR:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Store source address (PARAM2) in R3:R2 ...
   mov   DPTR, #PARAM2      ; Address LSB
   movx  A, @DPTR
   mov   R2, A              ; store address LSB
   inc   DPTR
   movx  A, @DPTR
   mov   R3, A              ; store address MSB

   ; Store memory specifier (PARAM3) in R4 ...
   mov   DPTR, #PARAM3      ; Memory Space LSB
   movx  A, @DPTR
   mov   R4, A              ; store memory space LSB

   ; Store memory monitor number (PARAM1) in R0 ...
   mov   DPTR, #PARAM1      ; Memory monitor index LSB
   movx  A, @DPTR
   anl   A, #7              ; modulo 8
   inc   ACC                ; increment because of DJNZ instruction below
   mov   R0, A              ; load loop counter with index+1

   ; Create bit mask (in ACC) and find offset into 'mLFDMADD' at which to
   ; store the address ...
   mov   DPTR, #mLFDMADD    ; beginning of memory monitor address table
   mov   A, #1              ; bit mask init to bit 0
   sjmp  DCE_LFDMADDR_20
DCE_LFDMADDR_10:   
   rl    A                  ; move bit left
   inc   DPTR               ; addresses are words
   inc   DPTR               ; so we advance pointer two bytes
DCE_LFDMADDR_20:
   djnz  R0, DCE_LFDMADDR_10

   mov   R5, A              ; save mask in R5
   cpl   A                  ; invert the bits to turn one off
   mov   R1, A              ; save inverted mask in R1
   
   ; Store PARAM2 in memory monitor address table
   mov   A, R2              ; store memory monitor address
   movx  @DPTR, A           ; put LSB of addr in table
   inc   DPTR
   mov   A, R3
   movx  @DPTR, A           ; put MSB of addr in table
   
   ; Use memory specifier (PARAM3) to set/clear bits in 'mLFDMDATA'
   ; and 'mLFDMEXTR'.
   mov   A, R4              ; get memory space select bits
   anl   A, #3              ; modulo 4
   rl    A                  ; multiply by two for jump table
   mov   DPTR, #MM_TBL      ; Use jump table below to select command
   jmp   @A+DPTR
MM_TBL:
   sjmp  MM_XDATA           ; XDATA space (external)
   sjmp  MM_CODE            ; CODE space
   sjmp  MM_DATA            ; DATA space (internal 8051)
   sjmp  DCE_LFDMADDR_99    ; Illegal.  Exit.

MM_XDATA:                   ; External data
   mov   DPTR, #mLFDMDATA   ; get bitmap for data/code
   movx  A, @DPTR
   anl   A, R1              ; CLEAR the appropriate bit using mask
   movx  @DPTR, A           ; save bitmap
   mov   DPTR, #mLFDMEXTR   ; get bitmap for internal/external data
   movx  A, @DPTR
   anl   A, R1              ; CLEAR the appropriate bit using mask
   movx  @DPTR, A           ; save bitmap
   sjmp  DCE_LFDMADDR_98    ; success

MM_CODE:                    ; Code
   mov   DPTR, #mLFDMDATA   ; get bitmap for data/code
   movx  A, @DPTR
   orl   A, R5              ; SET the appropriate bit using mask
   movx  @DPTR, A           ; save bitmap
   sjmp  DCE_LFDMADDR_98    ; success

MM_DATA:                    ; Internal data
   mov   DPTR, #mLFDMDATA   ; get bitmap for data/code
   movx  A, @DPTR
   anl   A, R1              ; CLEAR the appropriate bit using mask
   movx  @DPTR, A           ; save bitmap
   mov   DPTR, #mLFDMEXTR   ; get bitmap for internal/external data
   movx  A, @DPTR
   orl   A, R5              ; SET the appropriate bit using mask
   movx  @DPTR, A           ; save bitmap
    
DCE_LFDMADDR_98:
   setb  EXE_FLAG           ; command executed

DCE_LFDMADDR_99:
   pop   PSW                ; restore register bank to previous
   ret                      ; DCE_LFDMADDR
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDCRC (DCE Opcode '82'X)
; Purpose -- Setup a background task to compute the CRC of a specified range
;            of memory.  The CRC is computed by routine 'BG_CRC', the
;            background CRC task handler.  'Region 0' is used for the region
;            specified in this command.
; Inputs -- 3 command parameters:
;    PARAM1 = starting address  0..FFFF
;    PARAM2 = number of bytes   0..FFFF
;    PARAM3 = memory space      0..1 (0=XDATA, 1=CODE)
; Parameter Range Checking --
;    (1) PARAM1 N/A
;    (2) PARAM2 N/A
;    (3) PARAM3 only allowed the values '0' or '1'
; Outputs -- While computing the CRC, housekeeping will show CRC=0000h. Once
;            CRC computation is complete, housekeeping will show the computed
;            value in parameter 'LFMCRC'.
; Timing -- 165 machine cycles (0.248 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDCRC:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; This command merely populates element 0 of the background CRC table.
   ; The background CRC task is then initiated as for other regions.
   
   ; Read elements from the PARAM list, push them onto the 8051 stack,
   ; then write them to the CRC table ...

   ; Read the source specifier (PARAM3), perform range checking, and if OK
   ; push onto 8051 stack ...
   mov   DPTR, #PARAM3      ; memory space specifier
   movx  A, @DPTR           ; PARAM3 LSB
   mov   R2, A              ; store PARAM3 LSB
   inc   DPTR
   movx  A, @DPTR           ; PARAM3 MSB
   jnz   DCE_LFDCRC_99      ; parameter error -- exit
   mov   R0, #0             ; assume XDATA region
   mov   A, R2
   jz    DCE_LFDCRC_10      ; XDATA is correct region
   mov   R0, #1             ; assume CODE region
   dec   A
   jnz   DCE_LFDCRC_99      ; parameter error -- exit

DCE_LFDCRC_10:
   mov   A, R0
   push  ACC                ; push memory specifier on stack

   ; Read the length (PARAM2) and push onto 8051 stack ...
   mov   DPTR, #PARAM2      ; point to LSB of length
   movx  A, @DPTR
   push  ACC                ; push length LSB onto stack
   inc   DPTR
   movx  A, @DPTR
   push  ACC                ; push length MSB onto stack

   ; Read the starting address (PARAM1) and push onto 8051 stack ...
   mov   DPTR, #PARAM1      ; point to LSB of starting address
   movx  A, @DPTR
   push  ACC                ; push address LSB onto stack
   inc   DPTR
   movx  A, @DPTR
   push  ACC                ; push address MSB onto stack

   ; Read items from stack and put into the background CRC table ...
   mov   DPTR, #CRC_REGIONS ; base address of CRC regions (Region 0)
   pop   ACC                ; pop address MSB from stack
   movx  @DPTR, A           ; store address MSB
   inc   DPTR
   pop   ACC                ; pop address LSB from stack
   movx  @DPTR, A           ; store address LSB
   inc   DPTR
   pop   ACC                ; pop length MSB from stack
   movx  @DPTR, A           ; store length MSB
   inc   DPTR
   pop   ACC                ; pop length LSB from stack
   movx  @DPTR, A           ; store length LSB
   inc   DPTR
   pop   ACC                ; pop memory specifier from stack
   movx  @DPTR, A           ; store memory space specifier
   
   ; Clear 'current CRC' to show 0000h while computing this CRC ...
   mov   DPTR, #mLFMCRC
   clr   A
   movx  @DPTR, A
   inc   DPTR
   movx  @DPTR, A
   ; Clear 'previous CRC' to prevent CRC mis-compare when completed ...
   inc   DPTR
   movx  @DPTR, A
   inc   DPTR
   movx  @DPTR, A

   mov   REGION_CRC, #0     ; LFDCRC uses region '0'
   lcall INIT_CRC           ; initialize internal variables for the new region

DCE_LFDCRC_98:
   setb  EXE_FLAG           ; command executed

DCE_LFDCRC_99:
   pop   PSW                ; restore register bank to previous
   ret                      ; DCE_LFDCRC
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDCOPY (DCE Opcode '83'X)
; Purpose -- Copies a range of memory (CODE or XDATA) to XDATA.
; Input -- 4 command parameters:
;    PARAM1 = source address       0..FFFF
;    PARAM2 = destination address  0..FFFF
;    PARAM3 = number of bytes      0..FFFF (0=no copy, 1-FFFF allowed)
;    PARAM4 = source memory space  0..1 (0=XDATA, 1=CODE)
; Parameter Range Checking --
;    (1) PARAM1 N/A
;    (2) PARAM2 N/A
;    (3) PARAM3 N/A
;    (4) PARAM4 only allowed the values '0' or '1'
; Output -- XDATA space has bytes copied.
; Timing -- N=1024 bytes of XDATA space ...
;              Timing = 30,898 machine cycles (46.3 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDCOPY:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Store source address (PARAM1) in R3:R2
   mov   DPTR, #PARAM1      ; point to source address LSB
   movx  A, @DPTR
   mov   R2, A              ; store source LSB
   inc   DPTR
   movx  A, @DPTR
   mov   R3, A              ; store source MSB
      
   ; Store destination address (PARAM2) in R5:R4
   mov   DPTR, #PARAM2      ; point to destination address LSB
   movx  A, @DPTR
   mov   R4, A              ; store destination LSB
   inc   DPTR
   movx  A, @DPTR
   mov   R5, A              ; store destination MSB
      
   ; Store number of bytes (PARAM3) in R0:R7
   mov   DPTR, #PARAM3      ; point to LSB
   movx  A, @DPTR
   mov   R7, A              ; store LSB
   inc   DPTR
   movx  A, @DPTR
   mov   R0, A              ; store MSB

   ; If length (PARAM3) is zero, we are to exit ...
   cjne  R0, #0, DCE_LFDCOPY_20  ; OK if length MSB is non-zero
   cjne  R7, #0, DCE_LFDCOPY_20  ; OK if length LSB is non-zero
   sjmp  DCE_LFDCOPY_99     ; exit (failure)
   
DCE_LFDCOPY_20:
   ; Range check source specifier (PARAM4) and use it to setup 'SRC_FLAG'
   mov   DPTR, #PARAM4      ; point to LSB (only byte that counts)
   movx  A, @DPTR
   mov   R1, A              ; store LSB
   inc   DPTR
   movx  A, @DPTR
   jnz   DCE_LFDCOPY_99     ; parameter error -- exit
   clr   SRC_FLAG           ; assume source is XDATA
   mov   A, R1
   jz    DCE_LFDCOPY_30     ; start copy if source is XDATA space
   setb  SRC_FLAG           ; assume source is CODE
   dec   A
   jnz   DCE_LFDCOPY_99     ; parameter error -- exit

DCE_LFDCOPY_30:             ; copy data
   lcall MEM_COPY

DCE_LFDCOPY_98:
   setb  EXE_FLAG           ; command executed

DCE_LFDCOPY_99:
   pop   PSW                ; restore register bank to previous
   ret                      ; DCE_LFDCOPY
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDUPLOD (DCE Opcode 'AD'X)
; Purpose -- Copy an upload packet to the desired XDATA locations if the CRC
;            matches that specified in the parameter list.  If the specified
;            CRC = 0000h, then no CRC comparison is made prior to the memory
;            copy (and the CRC is simply not calculated, to save time).
; Inputs -- 3 command parameters:
;    PARAM1 = destination address  0..FFFF
;    PARAM2 = number of bytes      1..1024 (decimal)
;    PARAM3 = predicted CRC        0..FFFF (0=skip CRC checking)
; Parameter Range Checking --
;    (1) PARAM1 N/A
;    (2) PARAM2 only allowed values 0001h through 0400h
;    (3) PARAM3 N/A
; Outputs -- Specified XDATA locations overwritten.
; Timing -- Test case:  N=1024 bytes
;           w/o CRC --> Timing = 31,184 machine cycles (46.8 msec @ 8-MHz)
;           w/  CRC --> Timing = 71,308 machine cycles (107.0 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDUPLOD:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2
   
   ; PARAM3 tells us whether we are to calculate the CRC for this upload
   setb  CALC_CRC           ; assume we ARE to calculate the CRC
   mov   DPTR, #PARAM3      ; point to LSB
   movx  A, @DPTR
   jnz   DCE_LFDUPLOD_00    ; jump on non-zero (assumption correct)
   inc   DPTR
   movx  A, @DPTR
   jnz   DCE_LFDUPLOD_00    ; jump on non-zero (assumption correct)
   clr   CALC_CRC           ; assumption wrong -- we are NOT computing CRC

DCE_LFDUPLOD_00:

   ; Store length (PARAM2) in R3:R1 and range check ...
   mov   DPTR, #PARAM2      ; point to length parameter LSB
   movx  A, @DPTR           ; read length LSB
   mov   R1, A              ; store LSB in R1
   inc   DPTR               ; point to length parameter MSB
   movx  A, @DPTR           ; read length MSB
   mov   R3, A              ; store MSB in R3
   ; decrement length and ensure it is in the range 0..3FF (hex)
   ; (but do not store the decremented length)
   clr   C                  ; prepare for SUBB instruction
   mov   A, R1              ; LSB --> ACC
   subb  A, #1              ; subtract 1
   mov   A, R3              ; MSB --> ACC
   subb  A, #0              ; propagate carry (borrow)
   ; Now length should be in the range 0..3FF (hex), so if the MSB exceeds
   ; 3 then the length is too long or too short.  Subtract 4 from the MSB
   ; and if the MSB was >= 4, there is NO CARRY BIT.  Therefore, if the
   ; carry bit is set, we are OK.
   clr   C                  ; prepare for SUBB instruction
   subb  A, #4
   jc    DCE_LFDUPLOD_10    ; jump if length is OK (carry bit set)
   
   mov   A, #LOW(DIAG0001)  ; Diagnostic:  Bad upload length
   lcall report_diagnostic
   sjmp  DCE_LFDUPLOD_99    ; exit
   
DCE_LFDUPLOD_10:

   ; temporarily store length parameter (currently in R3:R1) in R0:R7 ...
   mov   A, R3
   mov   R0, A
   mov   A, R1
   mov   R7, A

   jnb   CALC_CRC, DCE_LFDUPLOD_50  ; jump if NOT computing CRC

DCE_LFDUPLOD_20:            ; Compute CRC and compare to specified value
   mov   CRCH, #0FFh        ; initialize CRC to 0FFFFh
   mov   CRCL, #0FFh
   mov   DPTR, #RS01        ; starting address of upload packet

   ; Fixup "number of bytes" if LSB is non-zero (to make DJNZ commands work)
   cjne  R1, #0, DCE_LFDUPLOD_22
   sjmp  DCE_LFDUPLOD_30    ; MSB is OK since LSB=0
DCE_LFDUPLOD_22:
   inc   R3                 ; if LSB<>0 we must increment MSB for looping

DCE_LFDUPLOD_30:            ; read data byte-by-byte and compute CRC
   movx  A, @DPTR
   lcall DOCRC              ; compute CRC for this byte
   inc   DPTR
   djnz  R1, DCE_LFDUPLOD_30 ; loop by decrementing counter LSB
   djnz  R3, DCE_LFDUPLOD_30 ; loop by decrementing counter MSB

   ; Save upload packet CRC (little-endian) ...
   mov   DPTR, #mLFMXFER
   mov   A, CRCL
   movx  @DPTR, A           ; store LSB
   inc   DPTR
   mov   A, CRCH
   movx  @DPTR, A           ; store MSB
   
   ; Compare computed CRC with value provided (PARAM3) ...
   mov   DPTR, #PARAM3      ; point to LSB
   movx  A, @DPTR           ; read CRC parameter LSB
   cjne  A, CRCL, DCE_LFDUPLOD_40  ; jump on miscompare
   inc   DPTR
   movx  A, @DPTR           ; read CRC parameter MSB
   cjne  A, CRCH, DCE_LFDUPLOD_40  ; jump on miscompare
   sjmp  DCE_LFDUPLOD_50    ; CRC compares OK, so proceed with copy ...

DCE_LFDUPLOD_40:
   mov   A, #LOW(DIAG0002)  ; Diagnostic:  Bad upload CRC
   lcall report_diagnostic
   sjmp  DCE_LFDUPLOD_99    ; exit
                            
DCE_LFDUPLOD_50:            ; Copy upload data to XDATA space
   ; setup registers for MEM_COPY routine ...
   ; R0:R7 already holds "number of bytes" to copy
   ; R3:R2 must hold source address ...
   mov   R3, #HIGH(RS01)    ; address of upload packet
   mov   R2, #LOW(RS01)
   ; R5:R4 must hold destination address (PARAM1) ...
   mov   DPTR, #PARAM1      ; point to LSB of destination address
   movx  A, @DPTR
   mov   R4, A
   inc   DPTR               ; point to MSB of destination address
   movx  A, @DPTR
   mov   R5, A
   clr   SRC_FLAG           ; source memory specifier (0=XDATA, 1=CODE)
   lcall MEM_COPY           ; copy the data

   lcall INIT_CRC           ; added to v01016 (prevent background CRC errors)
   lcall CLEAR_OLD_CRCS     ; clear all 'previous CRC' values to prevent CRC
                            ; comparison errors due to this upload

DCE_LFDUPLOD_98:
   setb  EXE_FLAG           ; command executed
   
DCE_LFDUPLOD_99:
   pop   PSW                ; restore register bank to previous
   ret                      ; DCE_LFDUPLOD
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDDNLOD (DCE Opcode 'AE'X)
; Purpose -- Copy a range of XDATA into the download packet space.  Compute a
;            CRC on the specified range of memory and report it in the
;            housekeeping packet.
; Inputs -- 2 command parameters:
;    PARAM1 = memory address   0..FFFF
;    PARAM2 = number of bytes  1..1024 (decimal)
; Parameter Range Checking --
;    (1) PARAM1 N/A
;    (2) PARAM2 only allowed values 0001h through 0400h
; Outputs -- Download packet of specified XDATA locations.
; Timing -- Test case:  N=1024 bytes
;              Timing = 71,050 machine cycles (106.6 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDDNLOD:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2
   
   ; Store length (PARAM2) in R3:R1 and range check ...
   mov   DPTR, #PARAM2      ; point to length parameter LSB
   movx  A, @DPTR           ; read length LSB
   mov   R1, A              ; store LSB in R1
   inc   DPTR               ; point to length parameter MSB
   movx  A, @DPTR           ; read length MSB
   mov   R3, A              ; store MSB in R3
   ; decrement length and ensure it is in the range 0..3FF (hex)
   ; (but do not store the decremented length)
   clr   C                  ; prepare for SUBB instruction
   mov   A, R1              ; LSB --> ACC
   subb  A, #1              ; Subtract '1'
   mov   A, R3              ; MSB --> ACC
   subb  A, #0              ; propagate carry (borrow)
   ; Now length should be in the range 0..3FF (hex), so if the MSB exceeds
   ; 3 then the length is too long or too short.  Subtract 4 from the MSB
   ; and if the MSB was >= 4, there is NO CARRY BIT.  Therefore, if the
   ; carry bit is set, we are OK.
   clr   C                  ; prepare for SUBB instruction
   subb  A, #4
   jc    DCE_LFDDNLOD_10    ; jump if length is OK (carry bit set)
   
   mov   A, #LOW(DIAG0003)  ; Diagnostic:  Bad download length
   lcall report_diagnostic
   sjmp  DCE_LFDDNLOD_99    ; exit

DCE_LFDDNLOD_10:

   ; temporarily store length parameter (currently in R3:R1) in R0:R7 ...
   mov   A, R3
   mov   R0, A
   mov   A, R1
   mov   R7, A

DCE_LFDDNLOD_20:            ; Compute CRC and compare to specified value
   mov   CRCH, #0FFh        ; initialize the CRC to 0FFFFh
   mov   CRCL, #0FFh
   
   ; Put starting address of download block (PARAM1) in DPTR ...
   mov   DPTR, #PARAM1      ; point to LSB of source address
   movx  A, @DPTR
   mov   R4, A
   inc   DPTR               ; point to MSB of source address
   movx  A, @DPTR
   mov   DPH, A
   mov   DPL, R4
   
   ; store this starting address temporarily (for MEM_COPY to follow) ...
   push  DPH
   push  DPL

   ; Fixup "number of bytes" if LSB is non-zero (to make DJNZ commands work)
   cjne  R1, #0, DCE_LFDDNLOD_22
   sjmp  DCE_LFDDNLOD_30    ; MSB is OK since LSB=0
DCE_LFDDNLOD_22:
   inc   R3                 ; if LSB<>0 we must increment MSB for looping

DCE_LFDDNLOD_30:            ; read data byte-by-byte and compute CRC
   movx  A, @DPTR
   lcall DOCRC              ; compute CRC for this byte
   inc   DPTR
   djnz  R1, DCE_LFDDNLOD_30 ; loop by decrementing counter LSB
   djnz  R3, DCE_LFDDNLOD_30 ; loop by decrementing counter MSB

   ; Save download packet CRC (little-endian) ...
   mov   DPTR, #mLFMXFER
   mov   A, CRCL
   movx  @DPTR, A
   inc   DPTR
   mov   A, CRCH
   movx  @DPTR, A

   ; flag main routine to send dump packet ...
   setb  SEND_DUMP_FLAG
   
   ; recall DPTR from stack (for the MEM_COPY operation)
   pop   DPL
   pop   DPH

DCE_LFDDNLOD_50:            ; Copy XDATA to the download space
   ; setup registers for MEM_COPY routine ...
   ; R0:R7 already holds "number of bytes" to copy
   ; R3:R2 must hold source address ...
   mov   R3, DPH            ; address of download packet
   mov   R2, DPL
   ; R5:R4 must hold destination address ...
   mov   R5, #HIGH(TS01)    ; address of download area
   mov   R4, #LOW(TS01)
   clr   SRC_FLAG           ; source memory specifier (0=XDATA, 1=CODE)
   lcall MEM_COPY           ; copy the data

   ; Fill the remainder of the download buffer with character 'DD' ...
   ; MEM_COPY advances R5:R4 to point to next destination byte.
   ; Put this in DPTR and load ACC with desired fill character ...
   mov   DPH, R5
   mov   DPL, R4
   mov   A, #0DDh           ; fill with character 'DD'
   sjmp  DCE_LFDDNLOD_61
DCE_LFDDNLOD_60:
   movx  @DPTR, A
   inc   DPTR
DCE_LFDDNLOD_61:
   ; loop until DPTR points to the byte immediately following the download
   ; buffer ...
   mov   R0, DPL
   cjne  R0, #LOW(TS01+1024), DCE_LFDDNLOD_60
   mov   R0, DPH
   cjne  R0, #HIGH(TS01+1024), DCE_LFDDNLOD_60

DCE_LFDDNLOD_98:
   setb  EXE_FLAG           ; command executed
   
DCE_LFDDNLOD_99:
   pop   PSW                ; restore register bank to previous
   ret                      ; DCE_LFDDNLOD
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDGOTO (DCE Opcode 'EA'X)
; Purpose -- Transfer 8051 control to the address specified by PARAM1.
;            This is a dangerous command!
; Inputs -- 1 command parameter:
;    PARAM1 = jump-to address   0..FFFF (hex)
; Parameter Range Checking --
;    (1) PARAM1 N/A
; Outputs -- Jump to the specified address (must RETURN from there)
; Timing -- 25 machine cycles (0.038 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDGOTO:
   NOP6                     ; padding for future patch jumps
   ; Put jump-to address (PARAM1) in DPTR ...
   mov   DPTR, #PARAM1      ; point to LSB of jump address
   movx  A, @DPTR
   push  ACC                ; store LSB
   inc   DPTR               ; point to MSB of jump address
   movx  A, @DPTR
   mov   DPH, A             ; DPH <-- MSB
   pop   DPL                ; DPL <-- LSB
   clr   A                  ; prepare for JMP command

DCE_LFDGOTO_98:
   setb  EXE_FLAG           ; flag command executed
   
DCE_LFDGOTO_99:
   jmp   @A+DPTR            ; DCE_LFDGOTO
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDRSTP (DCE Opcode 'F0'X)
; Purpose -- Initiate a power-on reset by commanding.
; Inputs -- none
; Outputs -- System reset.  The XDATA reset pattern will be modified so as to
;            indicate a power-on reset.
; Timing -- First toggle of 'HARDRESET' line within 5 machine cycles.
;           Succeeding toggles are 5 msec apart (@ 8-MHz).  Reset should occur
;           after second toggle (i.e., ~5 msec into this routine).
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDRSTP:
   NOP6                     ; padding for future patch jumps
   ; Change the reset pattern ...
   mov   DPTR, #mPATTERN
   mov   A, #0
   movx  @DPTR, A

   cpl   HARDRESET          ; toggle reset line to trip POR circuit
    
   ; The Reset doesn't happen until we complement the HARDRESET bit again.
   ; We will toggle the HARDRESET line up to 64 more times before reporting
   ; as a diagnostic the failure to reset.  Each toggle is timed for 5 ms
   ; based on an 8 MHz oscillator (10 msec period).
   
delay_5msec   equ   1914    ; yields 5 msec toggle period

   mov   R2, #64            ; toggle 64 more times before reporting diag
DCE_LFDRSTP_00:
   mov   R0, #LOW(delay_5msec)  ; load '#LOW(delay_5msec)'
   mov   R1, #HIGH(delay_5msec) ; load '#HIGH(delay_5msec)'
DCE_LFDRSTP_10:
   djnz  R0, DCE_LFDRSTP_10 ; decrement counter LSB
   djnz  R1, DCE_LFDRSTP_10 ; decrement counter MSB
   cpl   HARDRESET          ; toggle reset line for HARDRESET
   djnz  R2, DCE_LFDRSTP_00 ; decrement the "64-toggles" counter

   ; We could get here if the reset line is not connected, as in breadboard.
   mov   A, #LOW(DIAG001F)  ; Diagnostic:  Commanded reset failure
   lcall report_diagnostic

DCE_LFDRSTP_99:
   ret                      ; DCE_LFDRSTP
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDWDOG (DCE Opcode 'F1'X)
; Purpose -- Enables or Disables the Watchdog Timer.
; Inputs -- 1 command parameter:
;    PARAM1 = disable/enable watchdog  0..1  (0=disable, 1=enable)
; Parameter Range Checking --
;    (1) PARAM1 only allowed the values '0' or '1'
; Outputs -- Watchdog timer disable/enabled, as desired.
; Timing -- 56 machine cycles to Enable, 43 to Disable ( < 0.084 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDWDOG:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 ...
   mov   DPTR, #PARAM1      ; point to LSB of PARAM1
   movx  A, @DPTR
   mov   R0, A              ; store LSB
   inc   DPTR               ; point to MSB
   movx  A, @DPTR
   jnz   DCE_LFDWDOG_99     ; parameter error -- exit
   mov   A, R0              ; LSB of parameter 1
   jz    DCE_LFDWDOG_20     ; disable watchdog
   dec   A                  ; decrement (to look for '1')
   jz    DCE_LFDWDOG_30     ; enable watchdog
   sjmp  DCE_LFDWDOG_99     ; parameter error -- exit
   
DCE_LFDWDOG_20:
   lcall DISABLE_DOG
   sjmp  DCE_LFDWDOG_98
   
DCE_LFDWDOG_30:
   lcall ENABLE_DOG
   
DCE_LFDWDOG_98:
   setb  EXE_FLAG           ; command executed
   
DCE_LFDWDOG_99:
   pop   PSW                ; restore register bank to previous
   ret                      ; DCE_LFDWDOG
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDDIAGC (DCE Opcode 'F4'X)
; Purpose -- Clear the 32 entries of the diagnostic push-down stack.  It DOES
;            NOT clear the incrementing diagnostic serial number reported with
;            the diagnostics in housekeeping.  Nor does this routine affect
;            HST status buffer messages reported separately in housekeeping.
; Inputs -- none
; Outputs -- Diagnostic stack cleared.
; Timing -- 405 machine cycles (0.608 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDDIAGC:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2
   
   mov   R0, #64            ; 32 2-byte entries to clear
   mov   DPTR, #mLFDDIAGS   ; address of diagnostic push-down stack
   clr   A                  ; fill value
DCE_LFDDIAGC_00:
   movx  @DPTR, A           ; zero-out location
   inc   DPTR               ; advance pointer to next location
   djnz  R0, DCE_LFDDIAGC_00 ; loop until done

DCE_LFDDIAGC_98:
   setb  EXE_FLAG           ; command executed

   pop   PSW                ; restore register bank to previous
   ret                      ; DCE_LFDDIAGC
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDRSTW (DCE Opcode 'F5'X)
; Purpose -- Initiate a watchdog reset by commanding.
; Inputs -- none
; Outputs -- System reset.  The XDATA reset pattern is NOT modified so as to
;            indicate a watchdog reset.
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDRSTW:
   NOP6                     ; padding for future patch jumps
    
   ; Since we are imitating a watchdog reset, we shall use the watchdog
   ; hardware.  This is a modified form of 'ENABLE_DOG' ...
   clr   EA                 ; hold off interrupts briefly
   ; ensure Timer0 is running and interrupt enabled ...
   setb  TR0                ; start Timer0
   setb  ET0                ; enable Timer0 interrupt
   mov   TL0, #0            ; timeout in 256 machine cycles
   mov   TH0, #0FFh
   ; setup the PCA for watchdog ...
   mov   CCON, #0           ; stop the PCA timer
   mov   CL, #0             ; clear PCA timer LSB
   mov   CH, #0             ; clear PCA timer MSB
   mov   CCAP4L, #LOW(1)    ; set watchdog timeout LSB
   mov   CCAP4H, #HIGH(1)   ; set watchdog timeout MSB
   mov   CMOD, #44h         ; Enable WDT mode, use Timer0 rollover
   mov   CCAPM4, #48h       ; PCA module 4 as watchdog
   mov   CCON, #40h         ; start the PCA timer
   setb  EA                 ; re-enable interrupts

   ; Delay 300 msec to allow watchdog hardware to reset the system.
   mov   R2, #3
DCE_LFDRSTW_00:
   mov   R0, #LOW(33460)    ; LSB
   mov   R1, #HIGH(33460)   ; MSB
DCE_LFDRSTW_10:
   djnz  R0, DCE_LFDRSTW_10 ; decrement counter LSB
   djnz  R1, DCE_LFDRSTW_10 ; decrement counter MSB
   djnz  R2, DCE_LFDRSTW_00 ; decrement main counter

   ; If the watchdog fails to reset the system, issue a diagnostic ...
   mov   A, #LOW(DIAG001F)  ; Diagnostic:  Commanded reset failure
   lcall report_diagnostic

DCE_LFDRSTW_99:
   ret                      ; DCE_LFDRSTW
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDRSTA (DCE Opcode 'FA'X)
; Purpose -- Perform an ACTEL counter reset.
; Inputs -- none
; Outputs -- none
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDRSTA:
   NOP6                     ; padding for future patch jumps

   ; Reset the ACTEL devices on the TDCs and DCE-A by giving them a low pulse
   ; of at least 10us ...
   clr   ACTEL_RESET        ; active low reset
   mov   R0, #3             ; Loop for 10 us
   djnz  R0, $
   setb  ACTEL_RESET        ; enable the ACTEL devices on the TDCs and DCE-A

DCE_LFDRSTA_98:
   setb  EXE_FLAG           ; command executed

DCE_LFDRSTA_99:
   ret                      ; DCE_LFDRSTA
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDHKREQ (DCE Opcode 'FF'X)
; Purpose -- Causes COS DCE FSW to send a housekeeping packet (and increment
;            the commands-received counter).
; Inputs -- none
; Outputs -- none
; Timing -- 11 machine cycles (0.017 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDHKREQ:
   NOP6                     ; padding for future patch jumps
   setb  EXE_FLAG           ; command executed
   ret                      ; DCE_LFDHKREQ
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- CMD_ILLEGAL
; Purpose -- Issues diagnostic for an illegal command.
; Inputs -- none
; Outputs -- Diagnostic message
; Timing -- 982 machine cycles (1.47 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CMD_ILLEGAL:
   NOP6                     ; padding for future patch jumps
   clr   CMD_BEGUN          ; don't flag illegal commands as begun
   mov   A, #LOW(DIAG0011)  ; Diagnostic:  Illegal Command
   lcall report_diagnostic
   ret                      ; CMD_ILLEGAL
;-----------------------------------------------------------------------------

FILL_TO_ADDRESS_CRC    9000h, 0FFh, 0E14Eh
reg6      equ   $
reg5len   equ   $-reg5

;*****************************************************************************
;*****************************************************************************
;                            Operate-Mode Commands
;*****************************************************************************
;*****************************************************************************
; Subroutine -- DCE_LFGBWK (DCE Opcode '85'X)
; Purpose -- This digitizer command is used to tune FUV detector performance
;            for Segment A or Segment B.  Adjusts the Begin Walk setting.
; Inputs -- 3 command parameters:
;    PARAM1 = Begin Walk Setting Adjustment  0..FF
;    PARAM2 = Segment Selection              0..1 (0=Segment A, 1=Segment B)
;    PARAM3 = Axis                           0..1 (0=X Axis, 1=Y Axis)
; Parameter Range Checking -- Performed in subroutine 'WRITE_DAC' ...
;    (1) PARAM1 -- must be in the range 0..FF
;    (2) PARAM2 -- must be either '0' or '1'
;    (3) PARAM3 -- must be either '0' or '1'
; Outputs --
; Timing -- 455 machine cycles (0.683 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFGBWK:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Put the correct DAC address in 'DAC_ADDR' based on PARAM3 ...
   mov   DAC_ADDR, #12h     ; Assume Axis=X
   mov   DPTR, #PARAM3      ; Axis parameter LSB
   movx  A, @DPTR
   jz    DCE_LFGBWK_10      ; jump if Axis=X
   mov   DAC_ADDR, #32h     ; Axis=Y

DCE_LFGBWK_10:
   lcall WRITE_DAC          ; Perform range checking and write to DAC if
                            ; parameters are OK.

DCE_LFGBWK_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFGBWK
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFGEWK (DCE Opcode '89'X)
; Purpose -- This digitizer command is used to tune FUV detector performance
;            for Segment A or Segment B.  Adjusts the End Walk setting.
; Inputs -- 2 command parameters:
;    PARAM1 = End Walk Setting Adjustment  0..FF
;    PARAM2 = Segment Selection            0..1 (0=Segment A, 1=Segment B)
;    PARAM3 = Axis                         0..1 (0=X Axis, 1=Y Axis)
; Parameter Range Checking -- Performed in subroutine 'WRITE_DAC' ...
;    (1) PARAM1 -- must be in the range 0..FF
;    (2) PARAM2 -- must be either '0' or '1'
;    (3) PARAM3 -- must be either '0' or '1'
; Outputs --
; Timing -- 455 machine cycles (0.683 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFGEWK:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Put the correct DAC address in 'DAC_ADDR' based on PARAM3 ...
   mov   DAC_ADDR, #13h     ; Assume Axis=X
   mov   DPTR, #PARAM3      ; Axis parameter LSB
   movx  A, @DPTR
   jz    DCE_LFGEWK_10      ; jump if Axis=X
   mov   DAC_ADDR, #33h     ; Axis=Y

DCE_LFGEWK_10:
   lcall WRITE_DAC          ; Perform range checking and write to DAC if
                            ; parameters are OK.

DCE_LFGEWK_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFGEWK
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFHVMAX (DCE Opcode '8D'X)
; Purpose -- This command sets the maximum high voltage that is allowed by the
;            hardware for Segment A or Segment B.
; Inputs -- 2 command parameters:
;    PARAM1 = High Voltage Setting  0..FF
;    PARAM2 = Segment Selection     0..1 (0=Segment A, 1=Segment B)
; Parameter Range Checking --
;    (1) PARAM1 -- must be in the range 0..FF
;    (2) PARAM2 -- must be either '0' or '1'
; Outputs --
; Timing -- 222 machine cycles (0.333 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFHVMAX:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..FF) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(0FFh)    ; R2:R3 holds upper range limit
   mov   R3, #LOW(0FFh)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFHVMAX_99     ; parameter error -- cleanup and exit

   ; Range check PARAM2 (0..1) ...
   mov   DPTR, #PARAM2
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFHVMAX_99     ; parameter error -- cleanup and exit

   ; Check whether the desired VMAX is below current target voltage ...
   ; Point to 'LFHVTGTA' or 'LFHVTGTB' ...
   mov   DPTR, #PARAM2      ; read Segment
   movx  A, @DPTR
   mov   DPTR, #hv_segA+3   ; assume Segment A
   clr   USE_SEGMENT        ; USE_SEGMENT=0 for Segment A
   jz    DCE_LFHVMAX_10
   mov   DPTR, #hv_segB+3   ; actually Segment B
   setb  USE_SEGMENT        ; USE_SEGMENT=1 for Segment B
DCE_LFHVMAX_10:
   movx  A, @DPTR           ; read VTGT
   mov   DPTR, #mVERR+1     ; store VTGT in 'mVERR+1'
   movx  @DPTR, A
   mov   DPTR, #PARAM1      ; desired VMAX
   movx  A, @DPTR
   mov   DPTR, #mVERR+0     ; store desired VMAX in 'mVERR+0'
   movx  @DPTR, A
   ; Check if desired VMAX less than current VTGT ...
   lcall CHECK_VMAX         ; returns ACC=0 on success
   jnz   DCE_LFHVMAX_99     ; exit (failure)

   ; desired VMAX not less than current VTGT, so set the value ...
DCE_LFHVMAX_20:
   mov   DPTR, #PARAM2      ; read Segment
   movx  A, @DPTR
   jnz   DCE_LFHVMAX_20B    ; jump for Segment B

DCE_LFHVMAX_20A:
   ; Segment A ...
   mov   DPTR, #PARAM1      ; read setting
   movx  A, @DPTR
   mov   DPTR, #hv_segA+2   ; pointer to VMAXA
   movx  @DPTR, A
   mov   DPTR, #VMAXA       ; I/O address for Segment A maximum value
   movx  @DPTR, A           ; write the value
   sjmp  DCE_LFHVMAX_98     ; exit (success)

DCE_LFHVMAX_20B:
   ; Segment B ...
   mov   DPTR, #PARAM1      ; read setting
   movx  A, @DPTR
   mov   DPTR, #hv_segB+2   ; pointer to VMAXB
   movx  @DPTR, A
   mov   DPTR, #VMAXB       ; I/O address for Segment B maximum value
   movx  @DPTR, A           ; write the value

DCE_LFHVMAX_98:
   setb  EXE_FLAG           ; command executed

DCE_LFHVMAX_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFHVMAX
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFGLQT (DCE Opcode '95'X)
; Purpose -- This digitizer command is used to tune FUV detector performance
;            for Segment A or Segment B.  Adjusts the Lower Charge Threshold.
; Inputs -- 2 command parameters:
;    PARAM1 = Lower Charge Threshold Setting  0..FF
;    PARAM2 = Segment Selection               0..1 (0=Segment A, 1=Segment B)
; Parameter Range Checking -- Performed in subroutine 'WRITE_DAC' ...
;    (1) PARAM1 -- must be in the range 0..FF
;    (2) PARAM2 -- must be either '0' or '1'
; Outputs --
; Timing -- 375 machine cycles (0.563 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFGLQT:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Put the correct DAC address in 'DAC_ADDR' ...
   mov   DAC_ADDR, #21h     ; DAC address
   lcall WRITE_DAC_10       ; Perform range checking and write to DAC if
                            ; parameters are OK.

DCE_LFGLQT_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFGLQT
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFGTT (DCE Opcode '97'X)
; Purpose -- This digitizer command is used to tune FUV detector performance
;            for Segment A or Segment B.  Adjusts the Timing Threshold in the
;            dispersion or cross-dispersion direction that defines the
;            criteria for valid events.
; Inputs -- 3 command parameters:
;    PARAM1 = Dispersion Timing Threshold     0..FF
;    PARAM2 = Segment Selection               0..1 (0=Segment A, 1=Segment B)
;    PARAM3 = Direction for Timing Threshold  0..1 (0=DISP, 1=XDISP)
; Parameter Range Checking -- Performed in subroutine 'WRITE_DAC' ...
;    (1) PARAM1 -- must be in the range 0..FF
;    (2) PARAM2 -- must be either '0' or '1'
;    (3) PARAM3 -- must be either '0' or '1'
; Outputs --
; Timing -- 455 machine cycles (0.683 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFGTT:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Put the correct DAC address in 'DAC_ADDR' based on PARAM3 ...
   mov   DAC_ADDR, #23h     ; Assume Axis=X
   mov   DPTR, #PARAM3      ; Axis parameter LSB
   movx  A, @DPTR
   jz    DCE_LFGTT_10       ; jump if Axis=X
   mov   DAC_ADDR, #20h     ; Axis=Y

DCE_LFGTT_10:
   lcall WRITE_DAC          ; Perform range checking and write to DAC if
                            ; parameters are OK.

DCE_LFGTT_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFGTT
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFRLSOVD (DCE Opcode '9C'X)
; Purpose -- This command overrides the protection against driving the door
;            against endstops.
; Inputs -- 1 command parameter:
;    PARAM1 = Enable/Disable the Endswitch Override (0=Disable, 1=Enable)
; Parameter Range Checking --
;    (1) PARAM1 -- must be either '0' or '1'
; Outputs -- write to CONTROLS2.7
; Timing -- 96 machine cycles (0.144 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFRLSOVD:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..1) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFRLSOVD_99    ; parameter error -- exit

   ; Parameter within range -- write to CONTROLS2.7 ...
   mov   DPTR, #PARAM1
   movx  A, @DPTR
   jz    DCE_LFRLSOVD_10    ; jump for Disable (bit='0')

   ; Enable ...
   clr   C2_LOCK            ; unlock CONTROLS2
   mov   DPTR, #CONTROLS2
   movx  A, @DPTR           ; read CONTROLS2
   setb  ACC.7              ; set the bit
   movx  @DPTR, A
   setb  C2_LOCK
   sjmp  DCE_LFRLSOVD_98    ; exit (success)

DCE_LFRLSOVD_10:
   ; Disable ...
   clr   C2_LOCK            ; unlock CONTROLS2
   mov   DPTR, #CONTROLS2
   movx  A, @DPTR           ; read CONTROLS2
   clr   ACC.7              ; clear the bit
   movx  @DPTR, A
   setb  C2_LOCK

DCE_LFRLSOVD_98:
   setb  EXE_FLAG           ; command executed

DCE_LFRLSOVD_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFRLSOVD
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFGSTIM (DCE Opcode '9E'X)
; Purpose -- This command configures the internal stim (stimulus) pulses.
; Inputs -- 2 command parameters:
;    PARAM1 = Stim Pulse Rate    0..3 (0=Off, 1=2pps, 2=30pps, 3=2000pps)
;    PARAM2 = Segment Selection  0..1 (0=Segment A, 1=Segment B)
; Parameter Range Checking --
;    (1) PARAM1 -- must be in the range 0..3
;    (2) PARAM2 -- must be either '0' or '1'
; Outputs --
; Timing -- 393 machine cycles (0.590 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFGSTIM:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..3) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(3)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(3)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFGSTIM_99     ; parameter error -- exit

   mov   DAC_ADDR, #70h     ; TDC MUX address for Stim Pulse command
   lcall WRITE_DAC_20

   ; Update state information for housekeeping (if command was executed)
   ; since there is no hardware talkback ...
   jnb   EXE_FLAG, DCE_LFGSTIM_99  ; command did not execute, exit
   mov   DPTR, #PARAM1      ; read PARAM1
   movx  A, @DPTR
   push  ACC                ; store temporarily
   mov   DPTR, #PARAM2      ; read PARAM2
   movx  A, @DPTR
   mov   DPTR, #mLFGSTIMA   ; assume Segment A
   jz    DCE_LFGSTIM_10
   mov   DPTR, #mLFGSTIMB   ; actually Segment B
DCE_LFGSTIM_10:
   pop   ACC                ; recall state information
   movx  @DPTR, A

DCE_LFGSTIM_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFGSTIM
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFGUQT (DCE Opcode 'A2'X)
; Purpose -- This digitizer command is used to tune FUV detector performance
;            for Segment A or Segment B.  Adjusts the Upper Charge Threshold.
; Inputs -- 2 command parameters:
;    PARAM1 = Upper Charge Threshold Setting  0..FF
;    PARAM2 = Segment Selection               0..1 (0=Segment A, 1=Segment B)
; Parameter Range Checking -- Performed in subroutine 'WRITE_DAC' ...
;    (1) PARAM1 -- must be in the range 0..FF
;    (2) PARAM2 -- must be either '0' or '1'
; Outputs --
; Timing -- 375 machine cycles (0.563 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFGUQT:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Put the correct DAC address in 'DAC_ADDR' ...
   mov   DAC_ADDR, #22h     ; DAC address
   lcall WRITE_DAC_10       ; Perform range checking and write to DAC if
                            ; parameters are OK.

DCE_LFGUQT_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFGUQT
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFGSHFT (DCE Opcode 'A4'X)
; Purpose -- This digitizer command is used to tune FUV detector performance
;            for Segment A or Segment B.  Shifts the Segment image space in
;            the dispersion or cross-dispersion direction.
; Inputs -- 3 command parameters:
;    PARAM1 = Image Shift Setting       0..FF
;    PARAM2 = Segment Selection         0..1 (0=Segment A, 1=Segment B)
;    PARAM3 = Direction of Image Shift  0..1 (0=DISP, 1=XDISP)
; Parameter Range Checking -- Performed in subroutine 'WRITE_DAC' ...
;    (1) PARAM1 -- must be in the range 0..FF
;    (2) PARAM2 -- must be either '0' or '1'
;    (3) PARAM3 -- must be either '0' or '1'
; Outputs --
; Timing -- 451 machine cycles (0.677 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFGSHFT:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Put the correct DAC address in 'DAC_ADDR' based on PARAM3 ...
   mov   DAC_ADDR, #10h     ; Assume Axis=X
   mov   DPTR, #PARAM3      ; Axis parameter LSB
   movx  A, @DPTR
   jz    DCE_LFGSHFT_10     ; jump if Axis=X
   mov   DAC_ADDR, #30h     ; Axis=Y

DCE_LFGSHFT_10:
   lcall WRITE_DAC          ; Perform range checking and write to DAC if
                            ; parameters are OK.

DCE_LFGSHFT_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFGSHFT
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFGSTR (DCE Opcode 'A6'X)
; Purpose -- This digitizer command is used to tune FUV detector performance
;            for Segment A or Segment B.  Stretches Segment image space in
;            the dispersion or cross-dispersion direction.
; Inputs -- 3 command parameters:
;    PARAM1 = Image Stretch Setting       0..FF
;    PARAM2 = Segment Selection           0..1 (0=Segment A, 1=Segment B)
;    PARAM3 = Direction of Image Stretch  0..1 (0=DISP, 1=XDISP)
; Parameter Range Checking -- Performed in subroutine 'WRITE_DAC' ...
;    (1) PARAM1 -- must be in the range 0..FF
;    (2) PARAM2 -- must be either '0' or '1'
;    (3) PARAM3 -- must be either '0' or '1'
; Outputs --
; Timing -- 451 machine cycles (0.677 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFGSTR:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Put the correct DAC address in 'DAC_ADDR' based on PARAM3 ...
   mov   DAC_ADDR, #11h     ; Assume Axis=X
   mov   DPTR, #PARAM3      ; Axis parameter LSB
   movx  A, @DPTR
   jz    DCE_LFGSTR_10      ; jump if Axis=X
   mov   DAC_ADDR, #31h     ; Axis=Y

DCE_LFGSTR_10:
   lcall WRITE_DAC          ; Perform range checking and write to DAC if
                            ; parameters are OK.

DCE_LFGSTR_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFGSTR
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFHVLOW (DCE Opcode 'B6'X)
; Purpose -- This command establishes the high voltage setting that will be
;            used when commanding the segment to the HV_LOW state using
;            the LFHSTATE command.
; Inputs -- 2 command parameters:
;    PARAM1 = High Voltage Setting  0..FF
;    PARAM2 = Segment Selection     0..1 (0=Segment A, 1=Segment B)
; Parameter Range Checking --
;    (1) PARAM1 -- must be in the range 0..FF
;    (2) PARAM2 -- must be either '0' or '1'
; Outputs --
; Timing -- 177 machine cycles (0.266 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFHVLOW:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Point to 'LFHVLOA' or 'LFHVLOB' ...
   mov   DPTR, #PARAM2      ; read Segment
   movx  A, @DPTR
   mov   DPTR, #hv_segA     ; assume Segment A
   jz    DCE_LFHVLOW_10
   mov   DPTR, #hv_segB     ; actually Segment B

DCE_LFHVLOW_10:
   lcall WRITE_SETTING      ; Range check parameters and store value if
                            ; parameters are OK.

DCE_LFHVLOW_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFHVLOW
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFPCRP (DCE Opcode 'BA'X)
; Purpose -- This command initializes the global Count Rate Protection (CRP)
;            processing for Segments A or B.
; Inputs -- 3 command parameters:
;    PARAM1 = Smoothing Interval               0..FF (0=Disable, 1-255 secs)
;    PARAM2 = Segment Selection                0..1 (0=Segment A, 1=Segment B)
;    PARAM3 = Maximum Event Counts per Second  0..FFFF
; Parameter Range Checking --
;    (1) PARAM1 -- must be in the range 0..FF
;    (2) PARAM2 -- must be either '0' or '1'
;    (3) PARAM3 -- N/A
; Outputs --
; Timing -- 341 machine cycles (0.512 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFPCRP:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..FF) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(0FFh)    ; R2:R3 holds upper range limit
   mov   R3, #LOW(0FFh)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFPCRP_99      ; parameter error -- exit

   ; Range check PARAM2 (0..1) ...
   mov   DPTR, #PARAM2
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFPCRP_99      ; parameter error -- exit

   ; Store parameters ...
   mov   DPTR, #PARAM1      ; PARAM1 into R5 ('LFPINT')
   movx  A, @DPTR
   mov   R5, A
   mov   DPTR, #PARAM3      ; PARAM3 into R7:R6 ('LFPCNT')
   movx  A, @DPTR
   mov   R6, A
   inc   DPTR
   movx  A, @DPTR
   mov   R7, A
   ; Select between Segment A or B using PARAM2 ...
   mov   DPTR, #PARAM2
   movx  A, @DPTR
   mov   DPTR, #crp_segA    ; assume Segment A
   jz    DCE_LFPCRP_10
   mov   DPTR, #crp_segB    ; actually Segment B
DCE_LFPCRP_10:
   mov   A, R5
   movx  @DPTR, A           ; store 'LFPINT' value
   inc   DPTR
   mov   A, R6
   movx  @DPTR, A           ; store 'LFPCNT' LSB
   inc   DPTR
   mov   A, R7
   movx  @DPTR, A           ; store 'LFPCNT' MSB

   ; Clear appropriate circular buffer of past count rates ...

   ; However, DON'T CLEAR RAM IF LFPINT = 0 ...
   mov   A, R5              ; LFPINT
   jz    DCE_LFPCRP_98      ; jump if LFPINT = 0

   ; LFPINT <> 0, so clear the buffer
   inc   DPTR               ; advance DPTR to buffer offset
   mov   A, #0
   movx  @DPTR, A           ; set buffer offset to 0
   inc   DPTR               ; advance DPTR to buffer pointer
   movx  A, @DPTR           ; pointer MSB
   push  ACC
   inc   DPTR
   movx  A, @DPTR           ; pointer LSB
   mov   DPL, A
   pop   DPH
   ; DPTR now points to start of CRP circular buffer and R5 holds the
   ; number of entries we must clear ('LFPINT').  Multiply 'LFPINT' by
   ; 3 to get the number of bytes to clear ...
   mov   A, R5
   mov   B, #3              ; 3-byte entries
   mul   AB                 ; product in B:A
   ; Put number of bytes to clear in R6:R7 ...
   mov   R7, A
   mov   R6, B
   lcall CLEAR_RAM

DCE_LFPCRP_98:
   setb  EXE_FLAG           ; command executed

DCE_LFPCRP_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFPCRP
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFHVPWR (DCE Opcode 'C0'X)
; Purpose -- This command turns on or off the high voltage power supply.
; Inputs -- 1 command parameter:
;    PARAM1 = HV Power Supply Control  0..1 (0=Off, 1=On)
; Parameter Range Checking --
;    (1) PARAM1 -- must be either '0' or '1'
; Outputs -- write to CONTROLS1.0
; Timing -- 147 machine cycles to turn OFF (0.221 msec @ 8-MHz)
;           37,847 machine cycles to turn ON (56.8 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFHVPWR:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..1) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFHVPWR_99     ; parameter error -- exit

   ; Parameter within range -- write to CONTROLS1.0 ...
   mov   DPTR, #PARAM1
   movx  A, @DPTR
   jz    DCE_LFHVPWR_20    ; jump for Off (bit='0')

   ; Turn On if LFHVENA allows ...
   jb    LFHVENA, DCE_LFHVPWR_21  ; jump if commanding allowed and
                                  ; requesting ON

   ; Otherwise, send a diagnostic ...
   mov   A, #LOW(DIAG0008)  ; Diagnostic:  HV commanding not enabled
   lcall report_diagnostic
   sjmp  DCE_LFHVPWR_99     ; exit (failure)

DCE_LFHVPWR_20:

   ; For Turn OFF, reset VMAX for both Segments ...
   mov   A, #0              ; zero out the settings
   mov   DPTR, #VMAXA
   movx  @DPTR, A
   mov   DPTR, #VMAXB
   movx  @DPTR, A
   mov   DPTR, #hv_segA+2   ; LFHVMAXA
   movx  @DPTR, A
   mov   DPTR, #hv_segB+2   ; LFHVMAXB
   movx  @DPTR, A
   sjmp  DCE_LFHVPWR_22     ; jump to items common to both ON and OFF command

DCE_LFHVPWR_21:

   ; For Turn ON, clear the circular buffers and histograms for the various
   ; data collection functions (version 01.017).  This is the reason for the
   ; large disparity in timing between turning ON and OFF.
   mov   DPTR, #start_history_seg
   mov   R6, #HIGH(length_history_seg)
   mov   R7, #LOW(length_history_seg)
   lcall CLEAR_RAM

DCE_LFHVPWR_22:

   ; For both Turn ON and OFF, reset V(target) and V(setting) ...
   mov   A, #0              ; zero out the settings
   mov   DPTR, #VSETA
   movx  @DPTR, A
   mov   DPTR, #VSETB
   movx  @DPTR, A
   mov   DPTR, #hv_segA+3   ; LFHVTGTA
   movx  @DPTR, A
   inc   DPTR               ; LFHVDACA
   movx  @DPTR, A
   mov   DPTR, #hv_segB+3   ; LFHVTGTB
   movx  @DPTR, A
   inc   DPTR               ; LFHVDACB
   movx  @DPTR, A
   
   ; PARAM1 determines whether we are turning power on or off ...
   mov   DPTR, #PARAM1
   movx  A, @DPTR
   jz    DCE_LFHVPWR_30    ; jump for Off (bit='0')

   ; Turning HV on ...

   ; Set the voltage state for housekeeping ...
   mov   DPTR, #hv_segAB+0
   mov   A, #7              ; set LFHSTATE = 7
   movx  @DPTR, A

DCE_LFHVPWR_25:
   ; Turn On HV power (CONTROLS1.0 = '1') ...
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR           ; read CONTROLS1
   setb  ACC.0              ; set the bit
   movx  @DPTR, A
   sjmp  DCE_LFHVPWR_98     ; exit (success)

DCE_LFHVPWR_30:
   ; Turning HV off ...
   
   ; Set the voltage state for housekeeping ...
   mov   DPTR, #hv_segAB+0
   mov   A, #0              ; LFHSTATE = 0
   movx  @DPTR, A

   ; Turn Off HV power (CONTROLS1.0 = '0') ...
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR           ; read CONTROLS1
   clr   ACC.0              ; clear the bit
   movx  @DPTR, A

DCE_LFHVPWR_98:
   setb  EXE_FLAG           ; command executed

DCE_LFHVPWR_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFHVPWR
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFRACT1 (DCE Opcode 'C1'X)
; Purpose -- LFRACT1 and LFDACT2 are redundant commands to open the FUV
;            detector door when the primary door open method fails.  Both
;            LFRACT1 and LFRACT2 must be preceded by the LFRACTEN command
;            which enables the actuator commands for opening the door.
; Inputs -- 1 command parameter:
;    PARAM1 = Power State for HOP Actuator  0..1 (0=Off, 1=On)
; Parameter Range Checking --
;    (1) PARAM1 -- must be either '0' or '1'
; Outputs -- write to CONTROLS2.1
; Timing -- 96 machine cycles (0.144 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFRACT1:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..1) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFRACT1_99     ; parameter error -- exit

   ; Parameter within range -- write to CONTROLS2.1 ...
   mov   DPTR, #PARAM1
   movx  A, @DPTR
   jz    DCE_LFRACT1_20     ; jump for Off (bit='0')

   ; Before turning on actuator 1,
   ; Check if actuator has been enabled (CONTROLS2.4 = '1') ...
   clr   C2_LOCK            ; unlock CONTROLS2
   mov   DPTR, #CONTROLS2
   movx  A, @DPTR           ; read CONTROLS2
   setb  C2_LOCK
   jb    ACC.4, DCE_LFRACT1_10  ; jump if enabled

   mov   A, #LOW(DIAG0010)  ; Diagnostic:  Door actuators not enabled
   lcall report_diagnostic
   sjmp  DCE_LFRACT1_99     ; exit (failure)

DCE_LFRACT1_10:
   ; Turn On ...
   clr   C2_LOCK            ; unlock CONTROLS2
   mov   DPTR, #CONTROLS2
   movx  A, @DPTR           ; read CONTROLS2
   setb  ACC.1              ; set the bit
   movx  @DPTR, A
   setb  C2_LOCK
   mov   DPTR, #DOOR_TIMEOUT  ; load countdown with proper value
   movx  A, @DPTR
   mov   DOOR_TIMER, A
   sjmp  DCE_LFRACT1_98     ; exit (success)

DCE_LFRACT1_20:
   ; Turn Off ...
   clr   C2_LOCK            ; unlock CONTROLS2
   mov   DPTR, #CONTROLS2
   movx  A, @DPTR           ; read CONTROLS2
   clr   ACC.1              ; clear the bit
   movx  @DPTR, A
   setb  C2_LOCK

DCE_LFRACT1_98:
   setb  EXE_FLAG           ; command executed

DCE_LFRACT1_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFRACT1
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFRACT2 (DCE Opcode 'C2'X)
; Purpose -- LFRACT1 and LFDACT2 are redundant commands to open the FUV
;            detector door when the primary door open method fails.  Both
;            LFRACT1 and LFRACT2 must be preceded by the LFRACTEN command
;            which enables the actuator commands for opening the door.
; Inputs -- 1 command parameter:
;    PARAM1 = Power State for HOP Actuator  0..1 (0=Off, 1=On)
; Parameter Range Checking --
;    (1) PARAM1 -- must be either '0' or '1'
; Outputs -- write to CONTROLS2.2
; Timing -- 96 machine cycles (0.144 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFRACT2:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..1) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFRACT2_99     ; parameter error -- exit

   ; Parameter within range -- write to CONTROLS2.2 ...
   mov   DPTR, #PARAM1
   movx  A, @DPTR
   jz    DCE_LFRACT2_20     ; jump for Off (bit='0')

   ; Before turning on actuator 2,
   ; Check if actuator has been enabled (CONTROLS2.4 = '1') ...
   clr   C2_LOCK            ; unlock CONTROLS2
   mov   DPTR, #CONTROLS2
   movx  A, @DPTR           ; read CONTROLS2
   setb  C2_LOCK
   jb    ACC.4, DCE_LFRACT2_10  ; jump if enabled

   mov   A, #LOW(DIAG0010)  ; Diagnostic:  Door actuators not enabled
   lcall report_diagnostic
   sjmp  DCE_LFRACT2_99     ; exit (failure)

DCE_LFRACT2_10:
   ; Turn On ...
   clr   C2_LOCK            ; unlock CONTROLS2
   mov   DPTR, #CONTROLS2
   movx  A, @DPTR           ; read CONTROLS2
   setb  ACC.2              ; set the bit
   movx  @DPTR, A
   setb  C2_LOCK
   mov   DPTR, #DOOR_TIMEOUT  ; load countdown with proper value
   movx  A, @DPTR
   mov   DOOR_TIMER, A
   sjmp  DCE_LFRACT2_98     ; exit (success)

DCE_LFRACT2_20:
   ; Turn Off ...
   clr   C2_LOCK            ; unlock CONTROLS2
   mov   DPTR, #CONTROLS2
   movx  A, @DPTR           ; read CONTROLS2
   clr   ACC.2              ; clear the bit
   movx  @DPTR, A
   setb  C2_LOCK

DCE_LFRACT2_98:
   setb  EXE_FLAG           ; command executed

DCE_LFRACT2_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFRACT2
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFRACTEN (DCE Opcode 'C3'X)
; Purpose -- This command enables or disables the secondary method of opening
;            the detector door, i.e., the High Output Parafin (HOP) actuators.
; Inputs -- 1 command parameter:
;    PARAM1 = HOP Actuator Enable/Disable Setting  0..1 (0=Disable, 1=Enable)
; Parameter Range Checking --
;    (1) PARAM1 -- must be either '0' or '1'
; Outputs -- write to CONTROLS2.4
; Timing -- 96 machine cycles (0.144 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFRACTEN:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..1) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFRACTEN_99    ; parameter error -- exit

   ; Parameter within range -- write to CONTROLS2.4 ...
   mov   DPTR, #PARAM1
   movx  A, @DPTR
   jz    DCE_LFRACTEN_10    ; jump for Disable (bit='0')

   ; Enable ...
   clr   C2_LOCK            ; unlock CONTROLS2
   mov   DPTR, #CONTROLS2
   movx  A, @DPTR           ; read CONTROLS2
   setb  ACC.4              ; set the bit
   movx  @DPTR, A
   setb  C2_LOCK
   sjmp  DCE_LFRACTEN_98    ; exit (success)

DCE_LFRACTEN_10:
   ; Disable ...
   clr   C2_LOCK            ; unlock CONTROLS2
   mov   DPTR, #CONTROLS2
   movx  A, @DPTR           ; read CONTROLS2
   clr   ACC.4              ; clear the bit
   movx  @DPTR, A
   setb  C2_LOCK

DCE_LFRACTEN_98:
   setb  EXE_FLAG           ; command executed

DCE_LFRACTEN_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFRACTEN
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFRAXPWR (DCE Opcode 'C4'X)
; Purpose -- This command controls power to the auxillary power supply.
; Inputs -- 1 command parameter:
;    PARAM1 = Auxillary Power Control  0..1 (0=Off, 1=On)
; Parameter Range Checking --
;    (1) PARAM1 -- must be either '0' or '1'
; Outputs -- write to CONTROLS2.0
; Timing -- 96 machine cycles (0.144 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFRAXPWR:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..1) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFRAXPWR_99    ; parameter error -- exit

   ; Parameter within range -- write to CONTROLS2.0 ...
   mov   DPTR, #PARAM1
   movx  A, @DPTR
   jz    DCE_LFRAXPWR_10    ; jump for Off (bit='0')

   ; Turn On ...
   clr   C2_LOCK            ; unlock CONTROLS2
   mov   DPTR, #CONTROLS2
   movx  A, @DPTR           ; read CONTROLS2
   setb  ACC.0              ; set the bit
   movx  @DPTR, A
   setb  C2_LOCK
   sjmp  DCE_LFRAXPWR_98    ; exit (success)

DCE_LFRAXPWR_10:
   ; Turn Off ...
   clr   C2_LOCK            ; unlock CONTROLS2
   mov   DPTR, #CONTROLS2
   movx  A, @DPTR           ; read CONTROLS2
   clr   ACC.0              ; clear the bit
   movx  @DPTR, A
   setb  C2_LOCK

DCE_LFRAXPWR_98:
   setb  EXE_FLAG           ; command executed

DCE_LFRAXPWR_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFRAXPWR
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFRACTRS (DCE Opcode 'C5'X)
; Purpose -- This command is used to relatch the FUV detector door to the door
;            motor after opening the door using the HOP actuators.
; Inputs -- 1 command parameter:
;    PARAM1 = Actuator Reset Function  0..1 (0=Abort, 1=Relatch)
; Parameter Range Checking --
;    (1) PARAM1 -- must be either '0' or '1'
; Outputs --
; Timing -- 118 machine cycles (0.177 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFRACTRS:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..1) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFRACTRS_99    ; parameter error -- exit

   mov   DPTR, #PARAM1
   movx  A, @DPTR
   jz    DCE_LFRACTRS_20    ; jump for Abort (bit='0')

   ; Attempt to reset door.
   ; Check if motor power is enabled (LFRMENA='1') ...
   jb    LFRMENA, DCE_LFRACTRS_10  ; jump if enabled

   mov   A, #LOW(DIAG000B)  ; Diagnostic:  Motor commands not enabled
   lcall report_diagnostic
   sjmp  DCE_LFRACTRS_99    ; exit (failure)

DCE_LFRACTRS_10:
   ; Check door latch status (I/O bit 4000.0) ...
   mov   DPTR, #HV_STATUS
   movx  A, @DPTR           ; bit 0 is door latch status
   ; Following line changed from 'jnb' to 'jb' because logic level was
   ; inverted ...
   jb   ACC.0, DCE_LFRACTRS_15 ; jump if door is unlatched

   ; since door status shows 'latched', we presume that the the actuators
   ; never fired ...
   mov   A, #LOW(DIAG0015)  ; Diagnostic:  Actuator already latched.
   lcall report_diagnostic
   sjmp  DCE_LFRACTRS_99    ; exit (failure)

DCE_LFRACTRS_15:
   ; door is unlatched, start the re-latch sequence ...
   ; Assert the door motor override (CONTROLS2.7 = '1') ...
   clr   C2_LOCK            ; unlock CONTROLS2
   mov   DPTR, #CONTROLS2
   movx  A, @DPTR
   setb  ACC.7
   movx  @DPTR, A
   setb  C2_LOCK
   ; Set door direction OPEN (CONTROLS1.4&3 = '01') ...
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR           ; read CONTROLS1
   anl   A, #11100111b      ; clear bits 3 & 4
   setb  ACC.3              ; write '01' to bits 4&3 (note change in polarity)
   movx  @DPTR, A
   ; Enable motor power (CONTROLS1.5 = '1') ...
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR           ; read CONTROLS1
   setb  ACC.5              ; set the bit
   movx  @DPTR, A
   ; Motor power is enabled, so restart three-minute timer (approx) ...
   mov   DPTR, #DOOR_TIMEOUT  ; load countdown with proper value
   movx  A, @DPTR
   mov   DOOR_TIMER, A
   sjmp  DCE_LFRACTRS_98    ; exit (success)

DCE_LFRACTRS_20:
   lcall DOOR_STOP          ; Perform 'Door Stop'

DCE_LFRACTRS_98:
   setb  EXE_FLAG           ; command executed

DCE_LFRACTRS_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFRACTRS
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFHVNOM (DCE Opcode 'C6'X)
; Purpose -- This command establishes the high voltage setting that will be
;            used when commanding the segment to the HV_NOM state using the
;            LFHSTATE command.
; Inputs -- 2 command parameters:
;    PARAM1 = High Voltage Setting  0..FF
;    PARAM2 = Segment Selection     0..1 (0=Segment A, 1=Segment B)
; Parameter Range Checking --
;    (1) PARAM1 -- must be in the range 0..FF
;    (2) PARAM2 -- must be either '0' or '1'
; Outputs --
; Timing -- 177 machine cycles (0.266 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFHVNOM:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Point to 'LFHVNOMA' or 'LFHVNOMB' ...
   mov   DPTR, #PARAM2      ; read Segment
   movx  A, @DPTR
   mov   DPTR, #hv_segA+1   ; assume Segment A
   jz    DCE_LFHVNOM_10
   mov   DPTR, #hv_segB+1   ; actually Segment B

DCE_LFHVNOM_10:
   lcall WRITE_SETTING      ; Range check parameters and store value if
                            ; parameters are OK.

DCE_LFHVNOM_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFHVNOM
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFHVSET (DCE Opcode 'CB'X)
; Purpose -- Set the detector segment's target voltage to that specified and
;            enable ramping.
; Inputs -- 2 command parameters:
;    PARAM1 = Target Voltage Value  0..FF
;    PARAM2 = Segment Selection     0..1 (0=Segment A, 1=Segment B)
; Parameter Range Checking --
;    (1) PARAM1 -- must be in the range 0..FF
;    (2) PARAM2 -- must be either '0' or '1'
; Outputs --
; Timing -- 232 machine cycles (0.348 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFHVSET:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..FF) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(0FFh)    ; R2:R3 holds upper range limit
   mov   R3, #LOW(0FFh)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFHVSET_99     ; parameter error -- cleanup and exit

   ; Range check PARAM2 (0..1) ...
   mov   DPTR, #PARAM2
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFHVSET_99     ; parameter error -- cleanup and exit

   ; Check whether the desired target voltage exceeds the current
   ; VMAX setting ...
   ; Point to 'LFHVMAXA' or 'LFHVMAXB' ...
   mov   DPTR, #PARAM2      ; read Segment
   movx  A, @DPTR
   mov   DPTR, #hv_segA+2   ; assume Segment A
   clr   USE_SEGMENT        ; USE_SEGMENT=0 for Segment A
   jz    DCE_LFHVSET_10
   mov   DPTR, #hv_segB+2   ; actually Segment B
   setb  USE_SEGMENT        ; USE_SEGMENT=1 for Segment B
DCE_LFHVSET_10:
   movx  A, @DPTR           ; read VMAX
   mov   DPTR, #mVERR+0     ; store VMAX in 'mVERR+0'
   movx  @DPTR, A
   mov   DPTR, #PARAM1      ; desired target voltage
   movx  A, @DPTR
   mov   DPTR, #mVERR+1     ; store target voltage in 'mVERR+1'
   movx  @DPTR, A
   ; Check if desired target is at or below VMAX ...
   lcall CHECK_VTGT         ; returns ACC=0 on success
   jnz   DCE_LFHVSET_99     ; exit (failure)

   ; target voltage does not exceed VMAX, so set the value ...

   ; Check if HV is enabled ...
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR
   jb    ACC.0, DCE_LFHVSET_20  ; jump if HV enabled

   ; HV not enabled, so make LFHSTATE = 0 ...
   clr   A
   mov   DPTR, #hv_segAB+0  ; LFHSTATE
   movx  @DPTR, A
   sjmp  DCE_LFHVSET_99     ; exit (failure)

DCE_LFHVSET_20:
   ; set LFHSTATE = 6 ...
   mov   A, #6
   mov   DPTR, #hv_segAB+0  ; LFHSTATE
   movx  @DPTR, A

   ; Set the target voltage and enable ramping ...
   mov   DPTR, #PARAM2      ; read Segment
   movx  A, @DPTR
   jnz   DCE_LFHVSET_20B    ; jump for Segment B

DCE_LFHVSET_20A:
   ; Segment A ...
   mov   DPTR, #PARAM1      ; read setting
   movx  A, @DPTR
   mov   DPTR, #hv_segA+3   ; pointer to 'target' voltage
   movx  @DPTR, A
   setb  LFHRMPA            ; enable ramping
   sjmp  DCE_LFHVSET_98     ; exit (success)

DCE_LFHVSET_20B:
   ; Segment B ...
   mov   DPTR, #PARAM1      ; read setting
   movx  A, @DPTR
   mov   DPTR, #hv_segB+3   ; pointer to 'target' voltage
   movx  @DPTR, A
   setb  LFHRMPB            ; enable ramping

DCE_LFHVSET_98:
   setb  EXE_FLAG           ; command executed

DCE_LFHVSET_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFHVSET
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFHRAMPT (DCE Opcode 'CE'X)
; Purpose -- This command specifies the time between steps when ramping up
;            the high voltage.  Units are 0.1 second.  Zero means take a
;            single step to the target voltage.
; Inputs -- 1 command parameter:
;    PARAM1 = Time Between Ramp Steps  0..FFFF
; Parameter Range Checking --
;    (1) PARAM1 -- N/A
; Outputs --
; Timing -- 33 machine cycles (0.050 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFHRAMPT:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; save the parameter value in 'LFHRAMPT' ...
   mov   DPTR, #PARAM1+0
   movx  A, @DPTR           ; read Parameter LSB
   mov   DPTR, #hv_segAB+1  ; storage location LSB
   movx  @DPTR, A           ; store LSB
   mov   DPTR, #PARAM1+1
   movx  A, @DPTR           ; read Parameter MSB
   mov   DPTR, #hv_segAB+2  ; storage location MSB
   movx  @DPTR, A           ; store LSB

DCE_LFHRAMPT_98:
   setb  EXE_FLAG           ; command executed

DCE_LFHRAMPT_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFHRAMPT
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFHQPWR (DCE Opcode 'CF'X)
; Purpose -- This command controls power to the two grids (one for each
;            segment) positioned above the Micro Channel Plate (MCP) detector
;            segments.
; Inputs -- 1 command parameter:
;    PARAM1 = Grid Power Option  0..1 (0=Off, 1=On)
; Parameter Range Checking --
;    (1) PARAM1 -- must be either '0' or '1'
; Outputs -- write to CONTROLS1.1
; Timing -- 96 machine cycles (0.144 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFHQPWR:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..1) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFHQPWR_99     ; parameter error -- exit

   mov   DPTR, #PARAM1
   movx  A, @DPTR
   jz    DCE_LFHQPWR_20     ; jump if requesting power OFF

   ; Requesting grid power ON ...
   jnb   LFHVENA, DCE_LFHQPWR_10 ; not allowed to turn power on

   ; Turn grid power ON (CONTROLS1.1 = '1') ...
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR           ; read CONTROLS1
   setb  ACC.1              ; set the bit
   movx  @DPTR, A
   sjmp  DCE_LFHQPWR_98     ; exit (success)

DCE_LFHQPWR_10:
   ; Grid Power not allowed ...
   mov   A, #LOW(DIAG0008)  ; Diagnostic:  HV commanding not enabled
   lcall report_diagnostic
   sjmp  DCE_LFHQPWR_99     ; exit (failure)

DCE_LFHQPWR_20:
   ; Turn grid power OFF (CONTROLS1.1 = '0') ...
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR           ; read CONTROLS1
   clr   ACC.1              ; clear the bit
   movx  @DPTR, A

DCE_LFHQPWR_98:
   setb  EXE_FLAG           ; command executed

DCE_LFHQPWR_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFHQPWR
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFHVILIM (DCE Opcode 'D0'X)
; Purpose -- This command establishes the high voltage current limit for each
;            detector segment.
; Inputs -- 1 command parameter:
;    PARAM1 = High Voltage Current Limit  0..FF (0=Disable, 1-255 limit value)
; Parameter Range Checking --
;    (1) PARAM1 -- must be in the range 0..FF
; Outputs --
; Timing -- 96 machine cycles (0.144 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFHVILIM:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..FF) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(0FFh)    ; R2:R3 holds upper range limit
   mov   R3, #LOW(0FFh)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFHVILIM_99    ; parameter error -- exit

   ; save the parameter value in 'mLFHVILIM' ...
   mov   DPTR, #PARAM1
   movx  A, @DPTR           ; read Parameter LSB
   mov   DPTR, #mLFHVILIM   ; storage location
   movx  @DPTR, A           ; store LSB
   ; also store it in each of the current-limit task structures (one for
   ; each segment) ...
   mov   DPTR, #ilim_param_hva+0  ; Segment A
   movx  @DPTR, A
   mov   DPTR, #ilim_param_hvb+0  ; Segment B
   movx  @DPTR, A

DCE_LFHVILIM_98:
   setb  EXE_FLAG           ; command executed

DCE_LFHVILIM_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFHVILIM
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFHVENA (DCE Opcode 'D1'X)
; Purpose -- Controls the software interlock on commands affecting high
;            voltage.
; Inputs -- 1 command parameter:
;    PARAM1 = High Voltage Control  0..1 (0=Disable, 1=Enable)
; Parameter Range Checking --
;    (1) PARAM1 -- must be either '0' or '1'
; Outputs -- bit LFHVENA set/cleared if command executed
; Timing -- 87 machine cycles (0.131 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFHVENA:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..1) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFHVENA_99     ; parameter error -- exit

   mov   DPTR, #PARAM1
   movx  A, @DPTR
   clr   LFHVENA            ; ASSUME we are to disable HV
   jz    DCE_LFHVENA_98     ; jump if assumption correct
   setb  LFHVENA            ; ACTUALLY supposed to enable HV

DCE_LFHVENA_98:
   setb  EXE_FLAG           ; command executed

DCE_LFHVENA_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFHVENA
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFRMDIR (DCE Opcode 'D9'X)
; Purpose -- This command defines which direction (open or close) the door
;            will move when command LFRMPWR initiates the door motion.
; Inputs -- 1 command parameter:
;  OLD ---> PARAM1 = Door Direction  0..2 (0=Safe, 1=Close, 2=Open)
;    PARAM1 = Door Direction  0..2 (0=Safe, 1=Open, 2=Close)
; Parameter Range Checking --
;    (1) PARAM1 -- must be in the range 0..2
; Outputs -- write to CONTROLS1.4 & 3
; Timing -- 100 machine cycles (0.150 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFRMDIR:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..2) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(2)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(2)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFRMDIR_99     ; parameter error -- exit

   ; Check if motor power is enabled (LFRMENA='1') ...
   jb    LFRMENA, DCE_LFRMDIR_10  ; jump if enabled

   mov   A, #LOW(DIAG000B)  ; Diagnostic:  Motor commands not enabled
   lcall report_diagnostic
   sjmp  DCE_LFRMDIR_99     ; exit (failure)

DCE_LFRMDIR_10:
   ; Motor power is enabled, so restart three-minute timer (approx) and issue
   ; desired command ...
   mov   DPTR, #DOOR_TIMEOUT  ; load countdown with proper value
   movx  A, @DPTR
   mov   DOOR_TIMER, A

   ; Parameter within range -- write to CONTROLS1.4&3 ...
   mov   DPTR, #PARAM1
   movx  A, @DPTR
   mov   R0, A              ; store parameter value

   mov   DPTR, #CONTROLS1
   movx  A, @DPTR           ; read CONTROLS1
   anl   A, #11100111b      ; clear bits 3 & 4
   cjne  R0, #0, DCE_LFRMDIR_20
   movx  @DPTR, A
   sjmp  DCE_LFRMDIR_98     ; exit (success)

DCE_LFRMDIR_20:
   cjne  R0, #1, DCE_LFRMDIR_30
   setb  ACC.3              ; write '01' to bits 4&3
   movx  @DPTR, A
   sjmp  DCE_LFRMDIR_98     ; exit (success)

DCE_LFRMDIR_30:
   setb  ACC.4              ; write '10' to bits 4&3
   movx  @DPTR, A

DCE_LFRMDIR_98:
   setb  EXE_FLAG           ; command executed

DCE_LFRMDIR_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFRMDIR
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFRMPWR (DCE Opcode 'DA'X)
; Purpose -- This command initiates the FUV detector door motion.
; Inputs -- 1 command parameter:
;    PARAM1 = Start or Stop Door Movement  0..1 (0=Stop, 1=Start)
; Parameter Range Checking --
;    (1) PARAM1 -- must be either '0' or '1'
; Outputs -- write to CONTROLS1.5
; Timing -- 98 machine cycles (0.147 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFRMPWR:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..1) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFRMPWR_99     ; parameter error -- exit

   ; Parameter within range -- write to CONTROLS1.5 ...
   mov   DPTR, #PARAM1
   movx  A, @DPTR
   jz    DCE_LFRMPWR_20     ; jump for Stop (bit='0')

   ; Start ...
   ; First, check if motor power is enabled (LFRMENA='1') ...
   jb    LFRMENA, DCE_LFRMPWR_10  ; jump if enabled

   mov   A, #LOW(DIAG000B)  ; Diagnostic:  Motor commands not enabled
   lcall report_diagnostic
   sjmp  DCE_LFRMPWR_99     ; exit (failure)

DCE_LFRMPWR_10:
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR           ; read CONTROLS1
   setb  ACC.5              ; set the bit
   movx  @DPTR, A
   ; Motor power is enabled, so restart three-minute timer (approx) ...
   mov   DPTR, #DOOR_TIMEOUT  ; load countdown with proper value
   movx  A, @DPTR
   mov   DOOR_TIMER, A
   sjmp  DCE_LFRMPWR_98     ; exit (success)

DCE_LFRMPWR_20:
   ; Stop (no need to check LFRMENA) ...
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR           ; read CONTROLS1
   clr   ACC.5              ; clear the bit
   movx  @DPTR, A

DCE_LFRMPWR_98:
   setb  EXE_FLAG           ; command executed

DCE_LFRMPWR_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFRMPWR
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFRMENA (DCE Opcode 'DB'X)
; Purpose -- This command enables or disables the primary method of opening
;            the detector door; i.e., the door motor.
; Inputs -- 1 command parameter:
;    PARAM1 = Enable/Disable Door Motor  0..1 (0=Disable, 1=Enable)
; Parameter Range Checking --
;    (1) PARAM1 -- must be either '0' or '1'
; Outputs --
; Timing -- 89 machine cycles (0.134 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFRMENA:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..1) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFRMENA_99     ; parameter error -- exit

   mov   DPTR, #PARAM1
   movx  A, @DPTR
   clr   LFRMENA            ; ASSUME we are to disable motor
   jz    DCE_LFRMENA_98     ; jump if assumption correct
   setb  LFRMENA            ; ACTUALLY supposed to enable motor
   mov   DPTR, #DOOR_TIMEOUT  ; load countdown with proper value
   movx  A, @DPTR
   mov   DOOR_TIMER, A

DCE_LFRMENA_98:
   setb  EXE_FLAG           ; command executed

DCE_LFRMENA_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFRMENA
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFRILIM (DCE Opcode 'DC'X)
; Purpose -- Set the auxillary power supply current limit.
; Inputs -- 1 command parameter:
;    PARAM1 = Auxillary Power Supply Current Limit  0..FF
; Parameter Range Checking --
;    (1) PARAM1 -- must be in the range 0..FF
; Outputs --
; Timing -- 87 machine cycles (0.131 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFRILIM:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (0..FF) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(0FFh)    ; R2:R3 holds upper range limit
   mov   R3, #LOW(0FFh)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFRILIM_99     ; parameter error -- exit

   ; Store parameter in 'LFRILIM' for comparison with H/W talkback ...
   mov   DPTR, #PARAM1
   movx  A, @DPTR
   mov   DPTR, #ilim_param_aux+0
   movx  @DPTR, A

DCE_LFRILIM_98:
   setb  EXE_FLAG           ; command executed

DCE_LFRILIM_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFRILIM
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFSAFE (DCE Opcode 'DD'X)
; Purpose -- Calls the 'DOOR_STOP' and 'SAFE_CONTROLS' functions.
; Inputs -- none
; Outputs -- none
; Timing -- 523 machine cycles (0.785 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFSAFE:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   lcall DOOR_STOP
   lcall SAFE_CONTROLS

DCE_LFSAFE_98:
   setb  EXE_FLAG           ; command executed

DCE_LFSAFE_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFSAFE
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFHSTATE (DCE Opcode 'E6'X)
; Purpose -- This command transitions the high voltage for both Segments A
;            and B from their current setting to one of four pre-programmed
;            states.
; Inputs -- 1 command parameter:
;    PARAM1 = High Voltage State  1..4 (1=NOMA, 2=NOMB, 3=NOMAB, 4=LOW)
; Parameter Range Checking --
;    (1) PARAM1 -- must be in the range 1..4
; Outputs --
; Timing -- 234 machine cycles (0.351 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFHSTATE:
   NOP6                     ; padding for future patch jumps
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 (1..4) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(1)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(1)
   mov   R2, #HIGH(4)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(4)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   DCE_LFHSTATE_99    ; parameter error -- exit

   ; First perform range checking that desired target voltage does not
   ; exceed the current VMAX for each segment ...

   ; Segment A first -- read current VMAX from table ...
   mov   DPTR, #hv_segA+2   ; VMAXA
   movx  A, @DPTR
   mov   DPTR, #mVERR+0
   movx  @DPTR, A           ; store current VMAXA in 'mVERR+0'
   ; Bit ACC.0 determines state of Segment A -- 0=LOW, 1=NOM
   mov   DPTR, #PARAM1      ; desired state
   movx  A, @DPTR
   mov   DPTR, #hv_segA+0   ; VLOWA
   jnb   ACC.0, DCE_LFHSTATE_20
   mov   DPTR, #hv_segA+1   ; VNOMA
DCE_LFHSTATE_20:
   movx  A, @DPTR           ; read new target value
   mov   R7, A              ; temporary storage of VTGTA
   mov   DPTR, #mVERR+1
   movx  @DPTR, A           ; store target in 'mVERR+1'
   clr   USE_SEGMENT        ; 0 = Segment A, 1 = Segment B
   lcall CHECK_VTGT         ; returns ACC=0 on success
   jnz   DCE_LFHSTATE_99    ; exit (failure)

   ; Segment B second -- read current VMAX from table ...
   mov   DPTR, #hv_segB+2   ; VMAXB
   movx  A, @DPTR
   mov   DPTR, #mVERR+0
   movx  @DPTR, A           ; store current VMAXB in 'mVERR+0'
   ; Bit ACC.1 determines state of Segment B -- 0=LOW, 1=NOM
   mov   DPTR, #PARAM1
   movx  A, @DPTR
   mov   DPTR, #hv_segB+0   ; VLOWB
   jnb   ACC.1, DCE_LFHSTATE_30
   mov   DPTR, #hv_segB+1   ; VNOMB
DCE_LFHSTATE_30:
   movx  A, @DPTR           ; read new target value
   mov   R6, A              ; temporary storage of VTGTB
   mov   DPTR, #mVERR+1
   movx  @DPTR, A           ; store target in 'mVERR+1'
   setb  USE_SEGMENT        ; 0 = Segment A, 1 = Segment B
   lcall CHECK_VTGT         ; returns ACC=0 on success
   jnz   DCE_LFHSTATE_99    ; exit (failure)

   ; No range errors, ensure HV is enabled before changing state ...
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR
   jb    ACC.0, DCE_LFHSTATE_40  ; jump if HV enabled
   ; HV not enabled, so make LFHSTATE = 0 ...
   clr   A
   mov   DPTR, #hv_segAB+0  ; LFHSTATE
   movx  @DPTR, A
   sjmp  DCE_LFHSTATE_99    ; exit (failure)

DCE_LFHSTATE_40:
   ; No range errors, so write the target values for both segments ...
   mov   A, R7              ; recall Segment A target voltage
   mov   DPTR, #hv_segA+3   ; store in VTGTA
   movx  @DPTR, A

   mov   A, R6              ; recall Segment B target voltage
   mov   DPTR, #hv_segB+3   ; store in VTGTB
   movx  @DPTR, A

   ; Store voltage state information in LFHSTATE ...
   mov   DPTR, #PARAM1
   movx  A, @DPTR
   mov   DPTR, #hv_segAB+0  ; LFHSTATE
   movx  @DPTR, A

   ; Enable ramping for both segments ...
   setb  LFHRMPA            ; Segment A
   setb  LFHRMPB            ; Segment B

DCE_LFHSTATE_98:
   setb  EXE_FLAG           ; command executed

DCE_LFHSTATE_99:
   pop   PSW                ; restore register bank
   ret                      ; DCE_LFHSTATE
;-----------------------------------------------------------------------------

FILL_TO_ADDRESS_CRC    9A00h, 0FFh, 0ACCFh
reg7      equ   $
reg6len   equ   $-reg6

;*****************************************************************************
;*****************************************************************************
;                        Operate-mode Support Routines
;*****************************************************************************
;*****************************************************************************
; Subroutine -- RANGECHK_PARAM
; Purpose -- Verifies that the 16-bit (little-endian) parameter pointed to
;            by DPTR falls between values held in R0:R1(low) and R2:R3(high),
;            inclusive.
; Inputs -- DPTR points to the 16-bit parameter to check (little-endian),
;           R0:R1 holds the smallest allowed value,
;           R2:R3 the largest allowed value
; Outputs -- Accumulator cleared if no error, non-zero on error.
; Timing -- 
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
RANGECHK_PARAM:
   NOP6                     ; padding for future patch jumps
   push  DPH                ; save pointer for later use
   push  DPL
   ; Subtract R0:R1 from parameter value.  Error if negative ...
   clr   C                  ; prepare for SUBB operation
   movx  A, @DPTR           ; LSB of parameter
   subb  A, R1              ; subtract LSB of lower limit
   inc   DPTR
   movx  A, @DPTR           ; MSB of parameter
   subb  A, R0              ; subtract MSB of lower limit
   mov   A, #1              ; assume there is an error ==> set ACC non-zero
   pop   DPL                ; recover pointer
   pop   DPH
   jc    RANGECHK_PARAM_99  ; jump if parameter less than lower limit

   push  DPH                ; save pointer for later restoration
   push  DPL
   ; Subtract parameter value from R2:R3.  Error if negative ...
   ; (Carry is already clear from above)
   movx  A, @DPTR           ; LSB of parameter
   mov   B, A               ; store parameter LSB temporarily
   mov   A, R3              ; LSB of upper limit
   subb  A, B               ; subtract parameter LSB from upper limit LSB
   inc   DPTR
   movx  A, @DPTR           ; MSB of parameter
   mov   B, A               ; store temporarily
   mov   A, R2              ; MSB of upper limit
   subb  A, B               ; subtract parameter MSB from upper limit MSB
   mov   A, #1              ; assume there is an error ==> set ACC non-zero
   pop   DPL                ; recover pointer
   pop   DPH
   jc    RANGECHK_PARAM_99  ; jump if parameter less than lower limit

   clr   A                  ; ACC zero denotes no errors
   
RANGECHK_PARAM_99:
   ret                      ; RANGECHK_PARAM
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- WRITE_DAC
; Purpose -- Supports all commands which write to DACs on either TDC A or B.
;            This subroutine performs range checking on the three supplied
;            parameters and, if in range, writes the data value (8-bits) to
;            the specified DAC address.
; Inputs -- 3 command parameters plus DAC address:
;    PARAM1 = Setting            0..FF
;    PARAM2 = Segment Selection  0..1 (0=Segment A, 1=Segment B)
;    PARAM3 = Axis               0..1 (0=X Axis, 1=Y Axis)
;    DAC_ADDR holds the D/A converter address on either TDC.
; Parameter Range Checking -- (these are the DEFAULT range settings for all
;                              TDC DAC commands)
;    (1) PARAM1 -- must be in the range 0..FF
;    (2) PARAM2 -- must be either '0' or '1'
;    (3) PARAM3 -- must be either '0' or '1'
; Outputs --
; Timing -- 
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
WRITE_DAC:                  ; Entry point for 3-parameter DAC calls which need
                            ; all three parameters range checked for the
                            ; DEFAULT range settings.
   NOP6                     ; padding for future patch jumps

   ; Range check PARAM3 (0..1) ...
   mov   DPTR, #PARAM3
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   WRITE_DAC_99       ; parameter error -- exit

WRITE_DAC_10:               ; Entry point for 2-parameter DAC calls which need
                            ; both parameters range checked for the
                            ; DEFAULT range settings.
   NOP6                     ; padding for future patch jumps

   ; Range check PARAM1 (0..FF) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(0FFh)    ; R2:R3 holds upper range limit
   mov   R3, #LOW(0FFh)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   WRITE_DAC_99       ; parameter error -- exit

WRITE_DAC_20:               ; Entry point for 2-parameter DAC calls which only
                            ; need PARAM2 checked for the DEFAULT range.
   NOP6                     ; padding for future patch jumps

   ; Range check PARAM2 (0..1) ...
   mov   DPTR, #PARAM2
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   WRITE_DAC_99       ; parameter error -- exit

   ; Prepare to call either subroutine 'TDC_A' or 'TDC_B'.
   ; Put 16-bit TDC value in R0:R1 ...
   mov   R0, DAC_ADDR       ; D/A address on TDC goes in R0
   mov   DPTR, #PARAM1      ; Setting value (LSB) goes in R1
   movx  A, @DPTR
   mov   R1, A              ; setting value

   ; Select the proper TDC Segment ...
   mov   DPTR, #PARAM2      ; Segment parameter LSB
   movx  A, @DPTR
   jb    ACC.0, WRITE_DAC_30 ; jump for Segment B
   lcall TDC_A              ; Segment A
   sjmp  WRITE_DAC_98       ; done

WRITE_DAC_30:
   lcall TDC_B              ; Segment B

WRITE_DAC_98:
   setb  EXE_FLAG           ; command executed

WRITE_DAC_99:
   ret                      ; WRITE_DAC
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- WRITE_SETTING
; Purpose -- Supports commands which write to voltage settings (low, nominal,
;            max) on either TDC A or B.
;            This subroutine performs range checking on the two supplied
;            parameters and, if in range, writes the data value (8-bits) to
;            the specified memory location (pointed to by DPTR).
; Inputs -- 2 command parameters:
;    PARAM1 = Setting            0..FF
;    PARAM2 = Segment Selection  0..1 (0=Segment A, 1=Segment B)
; Parameter Range Checking -- (these are the DEFAULT range settings)
;    (1) PARAM1 -- must be in the range 0..FF
;    (2) PARAM2 -- must be either '0' or '1'
; Outputs --
; Timing -- 
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
WRITE_SETTING:
   NOP6                     ; padding for future patch jumps

   push  DPH
   push  DPL

   ; Range check PARAM1 (0..FF) ...
   mov   DPTR, #PARAM1
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(0FFh)    ; R2:R3 holds upper range limit
   mov   R3, #LOW(0FFh)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   WRITE_SETTING_10   ; parameter error -- cleanup and exit

   ; Range check PARAM2 (0..1) ...
   mov   DPTR, #PARAM2
   mov   R0, #HIGH(0)       ; R0:R1 holds lower range limit
   mov   R1, #LOW(0)
   mov   R2, #HIGH(1)       ; R2:R3 holds upper range limit
   mov   R3, #LOW(1)
   lcall RANGECHK_PARAM     ; if error, returns with ACC non-zero
   jnz   WRITE_SETTING_10   ; parameter error -- cleanup and exit

   sjmp  WRITE_SETTING_20   ; parameters OK, continue

WRITE_SETTING_10:
   ; Parameter error(s).  We must pop the stack twice to remove unused
   ; DPTR value ...
   pop   ACC
   pop   ACC
   sjmp  WRITE_SETTING_99   ; exit (failure)

WRITE_SETTING_20:
   ; Parameters OK, so read setting value and store it in DPTR location ...
   mov   DPTR, #PARAM1      ; read PARAM1 -- setting value
   movx  A, @DPTR
   pop   DPL                ; retrieve DPTR from stack
   pop   DPH
   movx  @DPTR, A           ; store setting in appropriate location

WRITE_SETTING_98:
   setb  EXE_FLAG           ; command executed

WRITE_SETTING_99:
   ret                      ; WRITE_SETTING
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- CHECK_VTGT, CHECK_VMAX
; Purpose -- CHECK_VTGT checks that proposed VTGT value is less than or
;            equal to current VMAX value, else it generates a diagnostic.
;            CHECK_VMAX checks that proposed VMAX value is greater than or
;            equal to current VTGT value, else it generates a diagnostic.
; Inputs -- VMAX and VTGT stored as the first and second bytes, respectively,
;           of the variable 'mVERR'.  Bit variable USE_SEGMENT determines
;           the segment in question.
; Outputs -- If 'mVERR+0' < 'mVERR+1', issue diagnostic and return nonzero
;            value in ACC.  Else, return ACC=0.
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CHECK_VTGT:
   NOP6                     ; padding for future patch jumps
   mov   A, #LOW(DIAG0033)  ; Diagnostic:  Attempt to set VTGT > VMAX (Seg A)
   jnb   USE_SEGMENT, CHECK_VTGT_05
   mov   A, #LOW(DIAG0035)  ; Diagnostic:  Attempt to set VTGT > VMAX (Seg B)
CHECK_VTGT_05:
   push  ACC                ; store temporarily
   sjmp  CHECK_VMAX_10

CHECK_VMAX:
   NOP6                     ; padding for future patch jumps
   mov   A, #LOW(DIAG0034)  ; Diagnostic:  Attempt to set VMAX < VTGT (Seg A)
   jnb   USE_SEGMENT, CHECK_VMAX_05
   mov   A, #LOW(DIAG0036)  ; Diagnostic:  Attempt to set VMAX < VTGT (Seg B)
CHECK_VMAX_05:
   push  ACC                ; store temporarily

CHECK_VMAX_10:
   mov   DPTR, #mVERR
   movx  A, @DPTR           ; read VMAX value
   mov   B, A
   inc   DPTR
   movx  A, @DPTR           ; read VTGT value
   xch   A, B               ; ACC=VMAX, B=VTGT
   clr   C                  ; prepare for SUBB operation
   subb  A, B               ; subtract VMAX - VTGT
   pop   ACC                ; pop diagnostic number
   jc    CHECK_VMAX_20      ; jump on error
   clr   A                  ; indicate no error
   sjmp  CHECK_VMAX_99      ; exit, success

CHECK_VMAX_20:
   ; VTGT > VMAX, so issue diagnostic ...
   lcall report_diagnostic
   mov   A, #0FFh           ; indicate error

CHECK_VMAX_99:
   ret                      ; CHECK_VTGT
;-----------------------------------------------------------------------------

FILL_TO_ADDRESS_CRC    9E00h, 0FFh, 0A7D8h
reg8      equ   $
reg7len   equ   $-reg7

;*****************************************************************************
;*****************************************************************************
;                           Operate-Mode Activities
;*****************************************************************************
;*****************************************************************************
; Subroutine -- ACTIVITY_ILIM
; Purpose -- The current limit activity checks three current readings (HV A,
;            HV B, and Auxillary supply current) against their stored limits.
;            If a current exceeds its limit, a diagnostic is issued.  If the
;            current stays above the limit for 'N' consecutive samples, this
;            activity will turn off the HV.  Every transition of a current
;            from at or below its limit to over-the-limit will generate the
;            relevant diagnostic.  If a current reading ever shows 255 (max
;            for an 8-bit reading), the HV is shut off.  This activity also
;            stores the last 1024 samples of each of the currents in circular
;            buffers and a histogram of each.
;
;   <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
;   Pseudo-Code listing ...
;
;   # Set ILIM_STATE = 1
;   << Top of 'case' statement here >>
;   # if (ILIM_STATE == 0)
;      # exit
;   # elseif (ILIM_STATE == 1)
;      # if (auxillary current limit > 0)
;         # read and store auxillary current
;      # endif
;      # if (HV-A current limit > 0)
;         # read and store HV-A current
;      # endif
;      # if (HV-B current limit > 0)
;         # read and store HV-B current
;      # endif
;      # if (HV enabled)
;         # ILIM_STATE = 2  (check currents against limits)
;      # else
;         # ILIM_STATE = 5  (merely store currents in buffers)
;      # endif
;      # jump to top of 'case' statement
;   # elseif (ILIM_STATE == 2)
;      # if (auxillary current limit > 0)
;         # test auxillary current against limit (TEST)
;      # endif
;      # ILIM_STATE = 3
;      # jump to top of 'case' statement
;   # elseif (ILIM_STATE == 3)
;      # if (HV-A current limit > 0)
;         # test HV-A current against limit (TEST)
;      # endif
;      # ILIM_STATE = 4
;      # jump to top of 'case' statement
;   # elseif (ILIM_STATE == 4)
;      # if (HV-B current limit > 0)
;         # test HV-B current against limit (TEST)
;      # endif
;      # ILIM_STATE = 5
;      # jump to top of 'case' statement
;   # elseif (ILIM_STATE == 5)
;      # if (auxillary current limit > 0)
;         # store auxillary current in circular buffer and histogram (STORE)
;      # endif
;      # ILIM_STATE = 6
;      # jump to top of 'case' statement
;   # elseif (ILIM_STATE == 6)
;      # if (HV-A current limit > 0)
;         # store HV-A current in circular buffer and histogram (STORE)
;      # endif
;      # ILIM_STATE = 7
;      # jump to top of 'case' statement
;   # elseif (ILIM_STATE == 7)
;      # if (HV-B current limit > 0)
;         # store HV-B current in circular buffer and histogram (STORE)
;      # endif
;      # ILIM_STATE = 8
;      # jump to top of 'case' statement
;   # else
;      # ILIM_STATE = 0  (force an exit from this task)
;      # jump to top of 'case' statement
;   # endif
;
;   Pseudo-Code listing of "TEST" ...
;   
;   
;   
;   
;   Pseudo-Code listing of "STORE" ...
;   
;   
;   
;
;   <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
;
; Inputs -- This activity has a state variable, ILIM_STATE:
;              ILIM_STATE=0 --> end of activity
;              ILIM_STATE=1 --> Read three current values (from ADC)
;              ILIM_STATE=2 --> Test AUX current
;              ILIM_STATE=3 --> Test HV-A current
;              ILIM_STATE=4 --> Test HV-B current
;              ILIM_STATE=5 --> Store AUX data
;              ILIM_STATE=6 --> Store HV-A data
;              ILIM_STATE=7 --> Store HV-B data
;              else         --> end of activity

;   Based on 'ILIM_STATE', one of three parameter structures is
;   used:  'ilim_param_hva', 'ilim_param_hvb', or 'ilim_param_aux'.
;   The structure has the following form ...
;   <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
;   ilim_param_hva:
;      DB   128             ; Current limit
;      DB   20              ; Over-current count limit
;      DB   0               ; Over-current counts
;      DB   0               ; present reading
;      DB   0               ; previous reading
;      DW   0               ; offset into circular buffer (little-endian)
;      DW   mHV_IMON_BUFA   ; pointer to circular buffer (big-endian)
;      DW   mHV_IMON_PDFA   ; pointer to histogram (big-endian)
;   <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
;
; Outputs --
; Timing -- 
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ACTIVITY_ILIM:
   NOP6                     ; padding for future patch jumps

   mov   ILIM_STATE, #1     ; begin ILIM activity in state '1'

   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ; Top of 'case' statement ...
ACTIVITY_ILIM_00:
   mov   A, ILIM_STATE
   cjne  A, #0, ACTIVITY_ILIM_00_END
   ljmp  ACTIVITY_ILIM_99   ; exit
ACTIVITY_ILIM_00_END:
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ; State 1 --> rapidly read all three current sources (if HV is enabled) ...
ACTIVITY_ILIM_10:
   cjne  A, #1, ACTIVITY_ILIM_10_END

   mov   ILIM_STATE, #5     ; ASSUME HV disabled --> next state = 5
   ; Check for HV enabled ...
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR
   jnb   ACC.0, ACTIVITY_ILIM_101 ; jump if HV disabled
   mov   ILIM_STATE, #2     ; HV enabled --> next state = 2

   ; Read Auxillary Power Supply current ...
ACTIVITY_ILIM_101:
   mov   DPTR, #ilim_param_aux+0  ; point to 'Current limit'
   movx  A, @DPTR
   jz    ACTIVITY_ILIM_102  ; jump if current limit = 0 (disable task)
   mov   DPTR, #MUX
   mov   A, #LVPS_AUX_MON   ; MUX address of LVPS Aux. power supply current
   movx  @DPTR, A
   lcall ADC_READ
   mov   DPTR, #ilim_param_aux+3  ; point to 'present reading'
   movx  @DPTR, A

   ; Read HV-A current ...
ACTIVITY_ILIM_102:
   mov   DPTR, #ilim_param_hva+0  ; point to 'Current limit'
   movx  A, @DPTR
   jz    ACTIVITY_ILIM_103  ; jump if current limit = 0 (disable task)
   mov   DPTR, #MUX
   mov   A, #HV_IMON_A      ; MUX address HV-A current
   movx  @DPTR, A
   lcall ADC_READ
   mov   DPTR, #ilim_param_hva+3  ; point to 'present reading'
   movx  @DPTR, A

   ; Read HV-B current ...
ACTIVITY_ILIM_103:
   mov   DPTR, #ilim_param_hvb+0  ; point to 'Current limit'
   movx  A, @DPTR
   jz    ACTIVITY_ILIM_104  ; jump if current limit = 0 (disable task)
   mov   DPTR, #MUX
   mov   A, #HV_IMON_B      ; MUX address HV-B current
   movx  @DPTR, A
   lcall ADC_READ
   mov   DPTR, #ilim_param_hvb+3  ; point to 'present reading'
   movx  @DPTR, A

ACTIVITY_ILIM_104:
   ljmp  ACTIVITY_ILIM_00   ; Back to top of 'case' statement

ACTIVITY_ILIM_10_END:
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ; State 2 --> test AUX current ...
ACTIVITY_ILIM_20:
   cjne  A, #2, ACTIVITY_ILIM_20_END
   mov   DPTR, #ilim_param_aux  ; point to structure
   movx  A, @DPTR
   jz    ACTIVITY_ILIM_201  ; jump if current limit = 0 (disable task)
   ljmp  ACTIVITY_ILIM_80
ACTIVITY_ILIM_201:
   inc   ILIM_STATE         ; transition to next state
   ljmp  ACTIVITY_ILIM_00   ; Back to top of 'case' statement
ACTIVITY_ILIM_20_END:
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ; State 3 --> test HV-A current ...
ACTIVITY_ILIM_30:
   cjne  A, #3, ACTIVITY_ILIM_30_END
   mov   DPTR, #ilim_param_hva  ; point to structure
   movx  A, @DPTR
   jz    ACTIVITY_ILIM_301  ; jump if current limit = 0 (disable task)
   ljmp  ACTIVITY_ILIM_80
ACTIVITY_ILIM_301:
   inc   ILIM_STATE         ; transition to next state
   ljmp  ACTIVITY_ILIM_00   ; Back to top of 'case' statement
ACTIVITY_ILIM_30_END:
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ; State 4 --> test HV-B current ...
ACTIVITY_ILIM_40:
   cjne  A, #4, ACTIVITY_ILIM_40_END
   mov   DPTR, #ilim_param_hvb ; point to structure
   movx  A, @DPTR
   jz    ACTIVITY_ILIM_401  ; jump if current limit = 0 (disable task)
   ljmp  ACTIVITY_ILIM_80
ACTIVITY_ILIM_401:
   inc   ILIM_STATE         ; transition to next state
   ljmp  ACTIVITY_ILIM_00   ; Back to top of 'case' statement
ACTIVITY_ILIM_40_END:
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ; State 5 --> store AUX current in circular buffer and histogram ...
ACTIVITY_ILIM_50:
   cjne  A, #5, ACTIVITY_ILIM_50_END
   mov   DPTR, #ilim_param_aux  ; point to structure
   movx  A, @DPTR
   jz    ACTIVITY_ILIM_501  ; jump if current limit = 0 (disable task)
   ljmp  ACTIVITY_ILIM_90
ACTIVITY_ILIM_501:
   inc   ILIM_STATE         ; transition to next state
   ljmp  ACTIVITY_ILIM_00   ; Back to top of 'case' statement
ACTIVITY_ILIM_50_END:
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ; State 6 --> store HV-A current in circular buffer and histogram ...
ACTIVITY_ILIM_60:
   cjne  A, #6, ACTIVITY_ILIM_60_END
   mov   DPTR, #ilim_param_hva  ; point to structure
   movx  A, @DPTR
   jz    ACTIVITY_ILIM_601  ; jump if current limit = 0 (disable task)
   ljmp  ACTIVITY_ILIM_90
ACTIVITY_ILIM_601:
   inc   ILIM_STATE         ; transition to next state
   ljmp  ACTIVITY_ILIM_00   ; Back to top of 'case' statement
ACTIVITY_ILIM_60_END:
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ; State 7 --> store HV-B current in circular buffer and histogram ...
ACTIVITY_ILIM_70:
   cjne  A, #7, ACTIVITY_ILIM_70_END
   mov   DPTR, #ilim_param_hvb  ; point to structure
   movx  A, @DPTR
   jz    ACTIVITY_ILIM_701  ; jump if current limit = 0 (disable task)
   ljmp  ACTIVITY_ILIM_90
ACTIVITY_ILIM_701:
   inc   ILIM_STATE         ; transition to next state
   ljmp  ACTIVITY_ILIM_00   ; Back to top of 'case' statement
ACTIVITY_ILIM_70_END:
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

   ; Invalid state -- task goes dormant until restarted by timer0
   mov   ILIM_STATE, #0     ; transition to dormant state
   ljmp  ACTIVITY_ILIM_00   ; Back to top of 'case' statement

   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ; Perform current limit testing ...
ACTIVITY_ILIM_80:
   ; Store pointer to start of XDATA space structure in R6:R7 ...
   mov   R6, DPH
   mov   R7, DPL
   ; Store first 5 bytes from XDATA space structure in internal registers
   ; for speed and flexibility (use R1, R2, R3, R4, R5) ...
   movx  A, @DPTR
   mov   R1, A              ; store 'Current limit'
   inc   DPTR
   movx  A, @DPTR
   mov   R2, A              ; store 'Over-current count limit'
   inc   DPTR
   movx  A, @DPTR
   mov   R3, A              ; store 'Over-current counts'
   inc   DPTR
   movx  A, @DPTR
   mov   R4, A              ; store 'present reading'
   inc   DPTR
   movx  A, @DPTR
   mov   R5, A              ; store 'previous reading'

$if 1
   ; We are no longer checking for current reading = 255, so jump ...
   sjmp  ACTIVITY_ILIM_803
$else
   ; First, compare 'present reading' (R4) to 255 ...
   cjne  R4, #255, ACTIVITY_ILIM_803  ; jump if current <> 255
   ; 'present reading' = 255 --> shut off power and issue diagnostic ...
$endif

ACTIVITY_ILIM_801:
   ; Shut off HV power ...
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR
   clr   ACC.0              ; shut off HV power
   clr   ACC.1              ; shut off Grid power
   movx  @DPTR, A

   ; Reset VMAX and VSET for both Segments ...
   mov   A, #0              ; zero out the settings
   mov   DPTR, #VSETA
   movx  @DPTR, A
   mov   DPTR, #VSETB
   movx  @DPTR, A
   mov   DPTR, #VMAXA
   movx  @DPTR, A
   mov   DPTR, #VMAXB
   movx  @DPTR, A

   ; Update the HV ramping structures to reflect these hardware settings ...
   mov   DPTR, #hv_segA+2   ; LFHVMAXA
   movx  @DPTR, A
   inc   DPTR               ; LFHVTGTA
   movx  @DPTR, A
   inc   DPTR               ; LFHVDACA
   movx  @DPTR, A
   mov   DPTR, #hv_segB+2   ; LFHVMAXB
   movx  @DPTR, A
   inc   DPTR               ; LFHVTGTB
   movx  @DPTR, A
   inc   DPTR               ; LFHVDACB
   movx  @DPTR, A

   ; Disable voltage commands ...
   clr   LFHVENA

   ; set LFHSTATE = 0 ...
   clr   A
   mov   DPTR, #hv_segAB+0  ; LFHSTATE
   movx  @DPTR, A

   ; Set ILIM_COUNTDOWN to countdown value ...
   mov   DPTR, #ilim_param_all+0 ; offset to number of post-shutdown samples
   movx  A, @DPTR
   mov   ILIM_COUNTDOWN, A

   ; Load appropriate diagnostic code ...
   mov   A, ILIM_STATE
   mov   R0, #LOW(DIAG002E) ; Diagnostic:  AUX power disabled
   add   A, #-2
   jz    ACTIVITY_ILIM_802  ; jump if state=2 (AUX power)
   mov   R0, #LOW(DIAG002A) ; Diagnostic:  HV-A disabled
   add   A, #-1
   jz    ACTIVITY_ILIM_802  ; jump if state=3 (HV-A power)
   mov   R0, #LOW(DIAG002B) ; Diagnostic:  HV-B disabled

ACTIVITY_ILIM_802:
   ; store the diagnostic ...
   mov   A, R0
   lcall report_diagnostic
   sjmp  ACTIVITY_ILIM_807  ; copy back and go to next current measurement

ACTIVITY_ILIM_803:
   ; Check for 'present reading' (R4) >= 'Current limit' (R1) ...
   mov   A, R4              ; 'present reading'
   clr   C                  ; prepare for SUBB operation
   subb  A, R1              ; subtract 'Current limit'
   jnc   ACTIVITY_ILIM_804  ; jump if limit equalled or exceeded (Carry = '0')

   ; Since limit not reached nor exceeded, reset the
   ; 'Over-current counts' (R3) ...
   mov   R3, #0
   sjmp  ACTIVITY_ILIM_807  ; copy back and go to next current measurement

ACTIVITY_ILIM_804:
   ; limit equalled or exceeded --> increment 'Over-current counts' (R3) ...
   inc   R3

   ; if the 'Over-current counts' (R3) = 1, then we have transitioned from
   ; within-limits to out-of-limits ...
   cjne  R3, #1, ACTIVITY_ILIM_806  ; jump if NO transition

   ; We have transitioned from within-limits to out-of-limits.
   ; Send a diagnostic ...
   mov   A, ILIM_STATE
   mov   R0, #LOW(DIAG0016) ; Diagnostic:  AUX power disabled
   add   A, #-2
   jz    ACTIVITY_ILIM_805  ; jump if state=2 (AUX power)
   mov   R0, #LOW(DIAG0027) ; Diagnostic:  HV-A disabled
   add   A, #-1
   jz    ACTIVITY_ILIM_805  ; jump if state=3 (HV-A power)
   mov   R0, #LOW(DIAG0028) ; Diagnostic:  HV-B disabled

ACTIVITY_ILIM_805:
   mov   A, R0
   lcall report_diagnostic
   sjmp  ACTIVITY_ILIM_807  ; copy back and go to next current measurement

ACTIVITY_ILIM_806:
   ; If 'Over-current counts' (R3) >= 'Over-current count limit' (R2), 
   ; turn off power and issue diagnostic ...
   mov   A, R3              ; 'Over-current counts'
   clr   C                  ; prepare for SUBB operation
   subb  A, R2              ; subtract 'Over-current count limit'
   jc    ACTIVITY_ILIM_807  ; jump if 'Over-current count limit' NOT exceeded
                            ; (Carry = '1')

   ; 'Over-current count limit' was met or exceeded.  Kill power and issue
   ; diagnostic ...
   sjmp  ACTIVITY_ILIM_801

ACTIVITY_ILIM_807:
   ; store 2 bytes of scratchpad back to CODE space (XDATA) ...
   mov   DPH, R6
   mov   DPL, R7
   inc   DPTR               ; don't store 'Current limit'
   inc   DPTR               ; don't store 'Over-current count limit'
   mov   A, R3
   movx  @DPTR, A           ; (1) store 'Over-current counts'
   inc   DPTR               ; don't store 'present reading'
   inc   DPTR
   mov   A, R4              ; 'present reading' becomes 'previous reading'
   movx  @DPTR, A           ; (2) store 'previous reading'

   ; Go to next state (next current source) ...
   inc   ILIM_STATE         ; transition to next state
   ljmp  ACTIVITY_ILIM_00   ; Back to top of 'case' statement
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ; Store reading in its circular buffer and historgram ...
ACTIVITY_ILIM_90:
   ; Store pointer to start of XDATA space structure in R6:R7 ...
   mov   R6, DPH
   mov   R7, DPL
   ; Store 'present reading' in R4 ...
   inc   DPTR
   inc   DPTR
   inc   DPTR
   movx  A, @DPTR
   mov   R4, A              ; store 'present reading'

   ; Retrieve offset into circular buffer (bytes 5&6, little-endian, from
   ; structure) and add to circular buffer pointer.  Structure's base address
   ; held in R6:R7 ...
   mov   A, R7              ; LSB of structure pointer
   add   A, #5              ; move ahead to byte 5
   mov   DPL, A
   mov   A, R6              ; MSB of structure pointer
   addc  A, #0              ; propagate carry
   mov   DPH, A
   movx  A, @DPTR           ; read LSB of offset
   mov   B, A               ; temporary storage
   inc   DPTR
   movx  A, @DPTR           ; read MSB of offset
   push  ACC                ; push offset MSB
   push  B                  ; push offset LSB
   inc   DPTR               ; point to circular buffer pointer
   movx  A, @DPTR           ; read MSB of buffer pointer
   mov   B, A               ; temporary storage
   inc   DPTR
   movx  A, @DPTR           ; read LSB of buffer pointer
   mov   DPL, A
   pop   ACC                ; recover offset LSB from stack
   add   A, DPL             ; add LSB to DPTR
   mov   DPL, A
   pop   ACC                ; recover offset MSB from stack
   addc  A, B               ; add buffer pointer LSB (w/ carry)
   mov   DPH, A
   ; DPTR now pointer to the particular byte within the circular buffer where
   ; the 'present reading' is to be stored ...
   mov   A, R4              ; recover 'present reading'
   movx  @DPTR, A           ; store 'present reading' in circular buffer
   ; increment circular buffer pointer, but not beyond 1023 (3FFh) ...
   ; point DPTR to byte 5 in XDATA structure (base address in R6:R7) ...
   mov   A, R7
   add   A, #5
   mov   DPL, A
   mov   A, R6
   addc  A, #0
   mov   DPH, A
   movx  A, @DPTR
   add   A, #1              ; increment LSB of offset
   movx  @DPTR, A           ; store LSB of incremented offset
   inc   DPTR
   movx  A, @DPTR
   addc  A, #0              ; propagate carry to MSB
   anl   A, #3              ; MSB is module 4 for wrap-around in buffer
   movx  @DPTR, A           ; store MSB of incremented offset
   
   ; Update the histogram ...
   ; Retrieve pointer to histogram (bytes 9&10, big-endian, from
   ; structure).  Structure's base address held in R6:R7 ...
   mov   A, R7              ; LSB of structure pointer
   add   A, #9              ; move ahead to byte 9
   mov   DPL, A
   mov   A, R6              ; MSB of structure pointer
   addc  A, #0              ; propagate carry
   mov   DPH, A
   movx  A, @DPTR           ; read MSB of pointer
   push  ACC                ; push pointer MSB
   inc   DPTR
   movx  A, @DPTR           ; read LSB of pointer
   push  ACC                ; push pointer LSB
   mov   A, R4              ; recover 'present reading' (histogram offset)
   mov   B, #2              ; two bytes per entry
   mul   AB                 ; product (offset) stored in B:A
   mov   DPL, A
   pop   ACC                ; pop pointer LSB
   add   A, DPL             ; add LSB of offset
   mov   DPL, A
   pop   ACC                ; pop pointer MSB
   addc  A, B               ; add MSB of offset
   mov   DPH, A
   ; DPTR now holds the offset to the first byte (LSB) of the histogram entry
   ; which must be incremented ...
   push  DPH                ; store this address
   push  DPL
   movx  A, @DPTR           ; read LSB of histogram entry
   add   A, #1
   movx  @DPTR, A           ; store LSB of incremented histogram entry
   inc   DPTR
   movx  A, @DPTR           ; read MSB of histogram entry
   addc  A, #0              ; propagate carry bit to MSB
   movx  @DPTR, A           ; store MSB of incremented histogram entry
   pop   DPL
   pop   DPH
   jnc   ACTIVITY_ILIM_901  ; jump if count <= FFFFh
   ; Histogram entry too large (more than 16 bits), so limit it to FFFFh ...
   mov   A, #0FFh           ; write histogram count = FFFFh
   movx  @DPTR, A           ; store LSB
   inc   DPTR
   movx  @DPTR, A           ; store MSB
ACTIVITY_ILIM_901:
   ; Go to next state (next current source) ...
   inc   ILIM_STATE         ; transition to next state
   ljmp  ACTIVITY_ILIM_00   ; Back to top of 'case' statement
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

ACTIVITY_ILIM_99:
   ret                      ; ACTIVITY_ILIM
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ACTIVITY_CRP
; Purpose -- The Count-Rate-Protection (CRP) activity.
;
;   <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
;   Pseudo-Code listing ...
;   # if (LFPINT == 0)
;      # exit
;   # else
;      # Read current FEC counter from hardware, store locally.
;      # Load pointer to next location in circular buffer.
;      # Write FEC counter value to circular buffer.
;      # Multiply LFPINT by LFPCNT to find limit value.  (LIMIT)
;      # Add all values from circular buffer.            (SUM)
;      # Increment (and wrap, if necessary) buffer pointer.
;      # if (SUM > LIMIT)
;         >>>>> Worst Case Timing Event <<<<<
;         # Set flag indicating the CRP failure.
;         # Disable Ramping.
;         # Set voltage to VLOW.
;         # Issue Diagnostic 22.  (most processing time here)
;      # else
;         # exit
;      # endif
;   # endif
;   <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
;
; Inputs -- Bit parameter 'USE_SEGMENT' (0=Segment A, 1=Segment B)
;
;   Based on 'USE_SEGMENT', either structure 'crp_segA' or
;   'crp_segB' is used.  The structure has the following
;   form ...
;   <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
;   crp_segA:
;      DB   3           ; LFPINTA
;      DW   0           ; LFPCNTA (little-endian)
;      DB   0           ; offset into circular buffer
;      DW   mCRP_BUFA   ; pointer to circular buffer (big-endian)
;   <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
;
; Outputs -- 
; Timing -- Nominal:     
;           Worst Case:  
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ACTIVITY_CRP:
   NOP6                     ; padding for future patch jumps

   ; Test to see whether CRP is enabled (LFPINT<>0) for this segment ...
   mov   DPTR, #crp_segA+0  ; LFPINTA
   jnb   USE_SEGMENT, ACTIVITY_CRP_10
   mov   DPTR, #crp_segB+0  ; LFPINTB
ACTIVITY_CRP_10:
   movx  A, @DPTR
   mov   R0, A              ; store 'LFPINT'
   jnz   ACTIVITY_CRP_20    ; jump if CRP is enabled
   ljmp  ACTIVITY_CRP_99    ; jump if CRP NOT enabled

ACTIVITY_CRP_20:
   ; while DPTR is positioned, grab the next 2 bytes, 'LFPCNT', and store
   ; in R1:R2 ...
   inc   DPTR
   movx  A, @DPTR
   mov   R2, A              ; store 'LFPCNT' LSB
   inc   DPTR
   movx  A, @DPTR
   mov   R1, A              ; store 'LFPCNT' MSB

   ; read current FEC count value from memory ...
   mov   DPTR, #mFEC_A      ; Segment A
   jnb   USE_SEGMENT, ACTIVITY_CRP_30
   mov   DPTR, #mFEC_B      ; Segment B
ACTIVITY_CRP_30:
   ; read counter (little-endian) and store in R5:R6:R7 ...
   movx  A, @DPTR           ; read LSB
   mov   R7, A
   inc   DPTR
   movx  A, @DPTR
   mov   R6, A
   inc   DPTR
   movx  A, @DPTR
   mov   R5, A

   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ; NOTE -- if the FEC value is to be considered 'scaled' in some sense
   ; (probably decreased by a nice multiple of 2), this is the place to do
   ; it.  We could shift the 3-byte count value right by 'N' bits before
   ; storing it in the circular buffer.
   ; Answer from Ken B. -- NO SCALING.  Leave it alone.  12/14/00
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

   ; Write new value into circular buffer (little-endian).  First determine
   ; the current offset into the buffer ...
   mov   DPTR, #crp_segA+4  ; mCRP_BUFA
   jnb   USE_SEGMENT, ACTIVITY_CRP_40
   mov   DPTR, #crp_segB+4  ; mCRP_BUFB
ACTIVITY_CRP_40:

   ; push base address onto stack ...
   movx  A, @DPTR
   push  ACC                ; address MSB
   inc   DPTR
   movx  A, @DPTR
   push  ACC                ; address LSB

   mov   DPTR, #crp_segA+3  ; offset into circular buffer A
   jnb   USE_SEGMENT, ACTIVITY_CRP_50
   mov   DPTR, #crp_segB+3  ; offset into circular buffer B
ACTIVITY_CRP_50:

   movx  A, @DPTR
   mov   B, #3              ; 3-bytes per entry
   mul   AB
   pop   DPL                ; pop LSB of buffer address
   add   A, DPL             ; add LSB of offset
   mov   DPL, A             ; store it
   pop   ACC                ; pop MSB of buffer address
   addc  A, B               ; add MSB of offset
   mov   DPH, A             ; store it

   ; Write current counter value to circular buffer (little-endian) ...
   mov   A, R7              ; LSB
   movx  @DPTR, A
   inc   DPTR
   mov   A, R6
   movx  @DPTR, A
   inc   DPTR
   mov   A, R5              ; MSB
   movx  @DPTR, A

   ; Test whether or not we have met or exceeded our average count rate
   ; limit ...
   
   ; We'll start by multiplying 'LFPINT' by 'LFPCNT' and storing the
   ; NEGATIVE of the 3-byte value in R5:R6:R7.  Then we'll add all
   ; 'LFPINT' items from the circular buffer 'mCRP_BUF' onto R5:R6:R7.
   ; If the result is positive, we have exceeded our count limit.
   
   ; Multiply 'LFPINT' by 'LFPCNT'.  R0 * (R1:R2) = R0*R1*256 + R0*R2
   ; (also, we'll complement each byte as we go along so we end up with
   ; the negative value when done) ...
   mov   A, R0              ; 'LFPINT' was stored in R0, above
   mov   B, R1              ; MSB of 'LFPCNT' was stored in R1, above
   mul   AB                 ; 2-byte result in B:A
   mov   R6, A              ; put partial product in R5:R6
   mov   R5, B
   mov   A, R0
   mov   B, R2              ; LSB of 'LFPCNT' was stored in R2, above
   mul   AB
   cpl   A                  ; complement byte before storing
   mov   R7, A              ; add partial product to R6:R7
   mov   A, B               ; put MSB in ACC
   add   A, R6
   cpl   A                  ; complement byte before storing
   mov   R6, A
   mov   A, R5
   addc  A, #0              ; propagate carry into R5
   cpl   A                  ; complement byte before storing
   mov   R5, A
   ; Finally, to TRULY have the negative of the limit value, we must
   ; add 1 to the value in R5:R6:R7 ...
   mov   A, R7              ; LSB
   add   A, #1              ; add '1'
   mov   R7, A
   mov   A, R6
   addc  A, #0              ; propagate carry bit
   mov   R6, A
   mov   A, R5
   addc  A, #0              ; propagate carry bit
   mov   R5, A
   ; R5:R6:R7 now holds the negative of the product of 'LFPINT' and 'LFPCNT'.

   ; put copy of 'LFPINT' (still in R0) in R1 for looping ...
   mov   A, R0
   mov   R1, A

   ; Now add in the first 'LFPINT' 3-byte elements from 'mCRP_BUF' ...
   ; base address of circular buffer ...
   mov   DPTR, #crp_segA+4  ; mCRP_BUFA
   jnb   USE_SEGMENT, ACTIVITY_CRP_60
   mov   DPTR, #crp_segB+4  ; mCRP_BUFB
ACTIVITY_CRP_60:
   ; read base pointer and push onto stack ...
   movx  A, @DPTR
   push  ACC                ; push MSB
   inc   DPTR
   movx  A, @DPTR
   mov   DPL, A
   pop   DPH
   ; now DPTR points to start of circular buffer
ACTIVITY_CRP_65:
   movx  A, @DPTR
   add   A, R7
   mov   R7, A
   inc   DPTR
   movx  A, @DPTR
   addc  A, R6
   mov   R6, A
   inc   DPTR
   movx  A, @DPTR
   addc  A, R5              ; propagate carry
   mov   R5, A
   inc   DPTR
   jc    ACTIVITY_CRP_66    ; exit loop early if we exceed the limit
   djnz  R1, ACTIVITY_CRP_65  ; loop until all 'LFPINT' entries are summed

ACTIVITY_CRP_66:
   ; Carry = '1' means we have exceeded our average count rate ...
   push  PSW                ; save state of Carry bit temporarily

   ; before testing for over-limit, advance the buffer offset and wrap
   ; it around to zero if it reaches 'LFPINT' ...

   mov   DPTR, #crp_segA+3  ; offset into circular buffer A
   jnb   USE_SEGMENT, ACTIVITY_CRP_70
   mov   DPTR, #crp_segB+3  ; offset into circular buffer B
ACTIVITY_CRP_70:
   ; increment buffer offset and wrap-around if greater than or equal to
   ; the integration interval ...
   movx  A, @DPTR           ; read offset
   inc   A                  ; increment buffer offset
   movx  @DPTR, A           ; store it (temporarily)
   clr   C                  ; prepare for SUBB
   subb  A, R0              ; subtract:  offset - LFPINT
   jc    ACTIVITY_CRP_80    ; jump if offset is OK
   mov   A, #0              ; zero the offset
   movx  @DPTR, A           ; store the zeroed offset

ACTIVITY_CRP_80:
   ; now test to see if we met or exceeded the average count rate limit ...
   pop   PSW                ; restore state of carry bit
   jc    ACTIVITY_CRP_90    ; jump if we have met or exceeded limit

   ; limit NOT met or exceeded, so exit ...
   sjmp  ACTIVITY_CRP_99    ; exit

   ; limit met or exceeded -- ramp HV to VLOW (both segments) ...
ACTIVITY_CRP_90:
   ; set LFHSTATE = 4 ...
   mov   A, #4
   mov   DPTR, #hv_segAB+0  ; LFHSTATE
   movx  @DPTR, A

   ; Ramp HV to HV_LOW due to CRP violation (BOTH segments) ...

ACTIVITY_CRP_90A:
   ; Use ramping task to ramp Segment A to the lesser of its current DAC
   ; setting and VLOW ...
   mov   DPTR, #hv_segA+0   ; read VLOW setting
   movx  A, @DPTR
   mov   B, A
   mov   DPTR, #hv_segA+4   ; read VDAC setting
   movx  A, @DPTR
   clr   C                  ; prepare for subb operation
   subb  A, B               ; VDAC - VLOW
   mov   A, B               ; put VLOW in ACC (preparation)
   jnc   ACTIVITY_CRP_90A_1 ; jump if VDAC >= VLOW
   movx  A, @DPTR           ; read VDAC setting again
ACTIVITY_CRP_90A_1:
   mov   DPTR, #hv_segA+3   ; VTGT setting
   movx  @DPTR, A           ; write lesser of VDAC or VLOW to VTGT
   setb  LFHRMPA            ; ensure ramping is enabled

ACTIVITY_CRP_90B:
   ; Use ramping task to ramp Segment B to the lesser of its current DAC
   ; setting and VLOW ...
   mov   DPTR, #hv_segB+0   ; read VLOW setting
   movx  A, @DPTR
   mov   B, A
   mov   DPTR, #hv_segB+4   ; read VDAC setting
   movx  A, @DPTR
   clr   C                  ; prepare for subb operation
   subb  A, B               ; VDAC - VLOW
   mov   A, B               ; put VLOW in ACC (preparation)
   jnc   ACTIVITY_CRP_90B_1 ; jump if VDAC >= VLOW
   movx  A, @DPTR           ; read VDAC setting again
ACTIVITY_CRP_90B_1:
   mov   DPTR, #hv_segB+3   ; VTGT setting
   movx  @DPTR, A           ; write lesser of VDAC or VLOW to VTGT
   setb  LFHRMPB            ; ensure ramping is enabled

   ; Issue the diagnostic message ...
   ; First, write the HST error parameter ...
   mov   DPTR, #crp_segA+4  ; mCRP_BUFA -- start of circular buffer
   jnb   USE_SEGMENT, ACTIVITY_CRP_95
   mov   DPTR, #crp_segB+4  ; mCRP_BUFB -- start of circular buffer
ACTIVITY_CRP_95:
   movx  A, @DPTR           ; read MSB of circular buffer address
   mov   B, A               ; temporary storage
   inc   DPTR
   movx  A, @DPTR           ; read LSB of circular buffer address
   mov   DPTR, #mBADCRP     ; HST error parameter for this error
   movx  @DPTR, A           ; write LSB first
   inc   DPTR
   mov   A, B
   movx  @DPTR, A           ; write MSB last
   ; Now issue the diagnostic ...
   mov   A, #LOW(DIAG0022)  ; Diagnostic:  CRP failure
   lcall report_diagnostic

ACTIVITY_CRP_99:
   ret                      ; ACTIVITY_CRP
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ACTIVITY_HV_RAMP
; Purpose -- High-Voltage Ramping activity.
;
;   <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
;   Pseudo-Code listing ...
;   # Read current DAC setting from stored variable.     (DAC)
;   # Read current Target voltage from stored variable.  (TARGET)
;   # if (TARGET == DAC)
;      # Turn off ramping.
;      # Zero the ramping up-counter.
;   # elseif (TARGET < DAC)
;      # Command HV to TARGET value
;      # Turn off ramping.
;      # Zero the ramping up-counter.
;   # else
;      # if (LFHRAMPT==0)
;         # Command HV to TARGET value
;         # Turn off ramping.
;         # Zero the ramping up-counter.
;      # else
;         # Increment up-counter.
;         # if (up-counter >= LFHRAMPT)
;            # if (only one step remains to reach TARGET)
;               >>>>> Worst Case Timing Event <<<<<
;               # Command HV to TARGET value
;               # Turn off ramping.
;               # Zero the ramping up-counter.
;            # else
;               # Increment DAC setting in hardware and memory.
;               # Zero the ramping up-counter.
;         # else
;            # exit
;         # endif
;      # endif
;   # endif
;   <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
;
; Inputs -- Bit parameter 'USE_SEGMENT' (0=Segment A, 1=Segment B)
;
;   Based on 'USE_SEGMENT', either structure 'hv_segA' or
;   'hv_segB' is used.  The structure has the following
;   form ...
;   <><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
;   hv_segA:
;      DB   32   ; LFHVLOA  -- low voltage
;      DB   195  ; LFHVNOMA -- nominal voltage
;      DB   217  ; LFHVMAXA -- holds value written to VMAX register
;      DB   195  ; LFHVTGTA -- target voltage
;      DB   100  ; LFHVDACA -- holds value written to DAC (VSET)
;      DW   0    ; count-up to next ramp step (little-endian)
;   <><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
;
;   Another structure, 'hv_segAB', is common to both segments
;   A and B and has the form ...
;   <><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
;   hv_segAB:
;      DB   0    ; LFHSTATE
;      DW   0    ; LFHRAMPT (little-endian)
;   <><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
;
; Outputs --
; Timing -- 110 machine cycles (0.165 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ACTIVITY_HV_RAMP:
   NOP6                     ; padding for future patch jumps

   ; If target voltage is less than current voltage, set the target value
   ; and turn off ramping ...

   ; Read the target voltage setting (LFHVTGT) ...
   mov   DPTR, #hv_segA+3   ; assume Segment A
   jnb   USE_SEGMENT, ACTIVITY_HV_RAMP_10
   mov   DPTR, #hv_segB+3   ; actually Segment B
ACTIVITY_HV_RAMP_10:
   movx  A, @DPTR
   mov   R7, A              ; store LFHVTGT in R7

   ; Read the current DAC voltage setting (LFHVDAC) ...
   inc   DPTR
   movx  A, @DPTR           ; read LFHVDAC
   mov   B, A               ; store LFHVDAC in B

   ; Subtract (LFHVTGT - LFHVDAC).  If carry='1', target is
   ; below the current DAC setting ...
   clr   C
   mov   A, R7
   subb  A, B
   mov   R6, A              ; save difference temporarily
   jz    ACTIVITY_HV_RAMP_30  ; jump if target = DAC setting
   jnc   ACTIVITY_HV_RAMP_40  ; jump if target > DAC setting

ACTIVITY_HV_RAMP_15:
   ; Set the voltage immediately (it is going lower) ...
   ; Point to hardware address for setting value ...
   mov   DPTR, #VSETA       ; assume Segment A
   jnb   USE_SEGMENT, ACTIVITY_HV_RAMP_20
   mov   DPTR, #VSETB       ; actually Segment B
ACTIVITY_HV_RAMP_20:
   mov   A, R7              ; recall LFHVTGT setting
   movx  @DPTR, A           ; write LFHVTGT to hardware

ACTIVITY_HV_RAMP_30:
   ; If target <= DAC setting, turn off ramping and clean up ...
   jb    USE_SEGMENT, ACTIVITY_HV_RAMP_30B
ACTIVITY_HV_RAMP_30A:
   clr   LFHRMPA
   mov   DPTR, #hv_segA+4   ; point to DAC setting
   sjmp  ACTIVITY_HV_RAMP_30AB
ACTIVITY_HV_RAMP_30B:
   clr   LFHRMPB
   mov   DPTR, #hv_segB+4   ; point to DAC setting
ACTIVITY_HV_RAMP_30AB:
   ; store the target voltage in the 'current voltage' setting ...
   mov   A, R7              ; recall 'target' value
   movx  @DPTR, A           ; write to memory as 'DAC setting' value
   inc   DPTR               ; advance DPTR to up-counter
   clr   A
   movx  @DPTR, A           ; clear up-counter to zero
   inc   DPTR
   movx  @DPTR, A
   sjmp  ACTIVITY_HV_RAMP_99  ; exit

ACTIVITY_HV_RAMP_40:
   ; Since Target > DAC Setting, we must ramp up.
   ; First check to see if we have timed-out so we may take another step ...

   ; Read the up-count limit (LFHRAMPT) ...
   mov   DPTR, #hv_segAB+1
   movx  A, @DPTR           ; read up-counter limit LSB
   mov   R1, A
   inc   DPTR
   movx  A, @DPTR           ; read up-counter limit MSB
   mov   R0, A
   ; If LFHRAMPT=0, we jump immediately to the target voltage regardless
   ; of whether it is higher or lower ...
   cjne  R0, #0, ACTIVITY_HV_RAMP_45
   cjne  R1, #0, ACTIVITY_HV_RAMP_45
   sjmp  ACTIVITY_HV_RAMP_15  ; jump if LFHRAMPT=0

ACTIVITY_HV_RAMP_45:
   ; Increment up-counter value ...
   mov   DPTR, #hv_segA+5   ; assume Segment A
   jnb   USE_SEGMENT, ACTIVITY_HV_RAMP_50
   mov   DPTR, #hv_segB+5   ; actually Segment B
ACTIVITY_HV_RAMP_50:
   movx  A, @DPTR
   add   A, #1              ; increment LSB
   movx  @DPTR, A
   mov   R3, A              ; store up-counter LSB
   inc   DPTR
   movx  A, @DPTR
   addc  A, #0              ; propagate carry to MSB
   movx  @DPTR, A
   mov   R2, A              ; store up-counter MSB
   ; subtract  up-counter(R2:R3) - limit(R0:R1) ...
   clr   C                  ; prepare for SUBB operation
   mov   A, R3              ; up-counter LSB
   subb  A, R1              ; subtract limit LSB
   mov   A, R2              ; up-counter MSB
   subb  A, R0              ; subtract limit MSB
   ; if Carry='1', the up-counter HAS NOT reached the limit (necessary to
   ; take a step) ...
   jc    ACTIVITY_HV_RAMP_99  ; exit

   ; take one upward step in voltage for this segment ...
ACTIVITY_HV_RAMP_60:
   ; if target is only one step above current setting, treat this as a jump
   ; to the target value
   mov   A, R6              ; R6 holds (target - DAC setting)
   dec   A
   jz    ACTIVITY_HV_RAMP_15  ; jump if only one step remains

   ; more than one step remains, so increment the DAC setting value ...
   mov   DPTR, #hv_segA+4   ; assume Segment A
   jnb   USE_SEGMENT, ACTIVITY_HV_RAMP_65
   mov   DPTR, #hv_segB+4   ; actually Segment B
ACTIVITY_HV_RAMP_65:
   movx  A, @DPTR           ; read DAC setting from stored variable
   inc   A                  ; increment DAC setting
   movx  @DPTR, A           ; store the incremented value

   ; write incremented value to the DAC hardware ...
   mov   DPTR, #VSETA       ; assume Segment A
   jnb   USE_SEGMENT, ACTIVITY_HV_RAMP_70
   mov   DPTR, #VSETB       ; actually Segment B
ACTIVITY_HV_RAMP_70:
   movx  @DPTR, A           ; write DAC setting to hardware

   ; reset up-counter to zero ...
   mov   DPTR, #hv_segA+5   ; assume Segment A
   jnb   USE_SEGMENT, ACTIVITY_HV_RAMP_75
   mov   DPTR, #hv_segB+5   ; actually Segment B
ACTIVITY_HV_RAMP_75:
   clr   A
   movx  @DPTR, A
   inc   DPTR
   movx  @DPTR, A
   ; exit ...

ACTIVITY_HV_RAMP_99:
   ret                      ; ACTIVITY_HV_RAMP
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ACTIVITY_DOOR
; Purpose -- Perform 'Door Stop' if conditions indicate problems.
; Inputs -- none
; Outputs --
; Timing -- 
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ACTIVITY_DOOR:
   NOP6                     ; padding for future patch jumps

   ; Reasons to halt door activity ...

   ; Reason (1)  Door timer timed out -- shutdown called from Timer0 ISR.

   ; Reason (2)  Door latch status (I/O bit 4000.0) changed ...
   mov   DPTR, #HV_STATUS
   movx  A, @DPTR           ; bit 0 is door latch status
   jnb   ACC.0, ACTIVITY_DOOR_10
   cpl   DOOR_LATCH         ; xor DOOR_LATCH with current status
ACTIVITY_DOOR_10:
   jnb   DOOR_LATCH, ACTIVITY_DOOR_20 ; jump if NO CHANGE in status
   ; Door latch status changed.  Load 2 second time-out ...
   mov   DOOR_TIMER, #2

ACTIVITY_DOOR_20:
   ; Reason (3)  Motor current meets or exceeds limit (if limit nonzero) ...
   mov   DPTR, #ilim_param_aux  ; point to 'LFRILIM'
   movx  A, @DPTR           ; read the limit value
   jz    ACTIVITY_DOOR_98   ; jump if current limit zero (disabled)
   mov   B, A               ; store limit value
   mov   DPTR, #MUX
   mov   A, #LVPS_AUX_MON   ; MUX for Aux current sensor
   movx  @DPTR, A
   lcall ADC_READ           ; current sensor reading in ACC
   mov   R0, A              ; temporarily store current sensor reading
   clr   C                  ; prepare for SUBB operation
   subb  A, B               ; subtract Reading - Limit.  Carry=1 if OK.
   jc    ACTIVITY_DOOR_98   ; jump if NO VIOLATION
   ; Write the HST error parameter ...
   mov   DPTR, #mBADAUXI
   mov   A, R0              ; sensor reading causing shutdown
   movx  @DPTR, A
   inc   DPTR
   clr   A                  ; high byte always zero
   movx  @DPTR, A
   ; Then issue the diagnostic ...
   mov   A, #LOW(DIAG0026)  ; Diagnostic:  AUX current exceeded door current limit.
   lcall report_diagnostic

   lcall DOOR_STOP          ; Perform 'Door Stop'

ACTIVITY_DOOR_98:
   ; Set DOOR_LATCH to current door latch status ...
   mov   DPTR, #HV_STATUS
   movx  A, @DPTR           ; bit 0 is current door latch status
   clr   DOOR_LATCH         ; assume '0'
   jnb   ACC.0, ACTIVITY_DOOR_99 ; jump if assumption correct
   setb  DOOR_LATCH

ACTIVITY_DOOR_99:
   ret                      ; ACTIVITY_DOOR
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DOOR_STOP
; Purpose -- Perform 'Door Stop'
; Inputs -- none
; Outputs -- Door stopped and safed
; Timing -- 24 machine cycles (0.036 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DOOR_STOP:

   ; Motor power off (CONTROLS1.5,4,3 = '0') ...
   mov   DPTR, #CONTROLS1
   movx  A, @DPTR
   anl   A, #11000111b    ; clear bits 5,4,3 (motor power, direction -- 2 bits)
   clr   ACC.5
   movx  @DPTR, A

   clr   LFRMENA            ; disable motor commands

   ; Actuator enable(s) off (CONTROLS2.1,2,4,7 = '0') ...
   clr   C2_LOCK            ; unlock CONTROLS2
   mov   DPTR, #CONTROLS2
   movx  A, @DPTR
   anl   A, #01101001b     ; clear bits 7,4,2,1 (override, enable, Act2, Act1)
   movx  @DPTR, A
   setb  C2_LOCK

   ; Clear 3-minute timer ...
   mov   DOOR_TIMER, #0

DOOR_STOP_99:
   ret                      ; DOOR_STOP
;-----------------------------------------------------------------------------

FILL_TO_ADDRESS_CRC    0A400h, 0FFh, 05D28h
reg9      equ   $
reg8len   equ   $-reg8

;*****************************************************************************
;*****************************************************************************
;                          General Support Routines
;*****************************************************************************
;*****************************************************************************
; Subroutine -- COMMAND_VECTOR
; Purpose -- Uses the lookup table 'COS_OPER_TBL' to find the address of the
;            operate mode function specified and jumps to that routine.
; Inputs -- none
; Outputs -- JUMPS to the specified subroutine.  Returning from that routine
;            sends the code back to the place where this function was called
;            (subroutine 'PROCESS_COMMAND').
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
COMMAND_VECTOR:
   setb  CMD_BEGUN          ; 1=command format OK --> we WILL jump to it
   
   mov   DPTR, #RS17        ; point to Command opcode in receive buffer
   movx  A, @DPTR           ; load opcode
   anl   A, #7Fh            ; mask off msb
   mov   DPTR, #COS_OPER_TBL ; Point to Command jump table
   lcall CODE_TLU_WORD      ; 2-byte look-up value returned in B:A
   mov   DPH, B             ; put B:A into DPTR for code jump to follow
   mov   DPL, A
   clr   A                  ; prepare for code jump
   jmp   @A+DPTR            ; COMMAND_VECTOR
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- INCREMENT_WORD
; Purpose -- Increment a word pointed to by DPTR.  Assumed little-endian.
; Inputs -- DPTR to word to be incremented as little-endian value
; Outputs -- DPTR incremented twice; ACC changed.
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
INCREMENT_WORD:
   movx  A, @DPTR           ; fetch LSB
   add   A, #1              ; add 1
   movx  @DPTR, A           ; store LSB
   inc   DPTR               ; point to MSB
   movx  A, @DPTR           ; fetch MSB
   addc  A, #0              ; propagate carry bit
   movx  @DPTR, A           ; store MSB
   inc   DPTR               ; point to byte beyond
   ret                      ; INCREMENT_WORD
;-----------------------------------------------------------------------------   

;*****************************************************************************
; Subroutine -- INCREMENT_LWORD
; Purpose -- Increment a long word pointed to by DPTR.  Assumed little-endian.
; Inputs -- DPTR to word to be incremented as little-endian value
; Outputs -- DPTR incremented 2 or 4 times, depending on carry; ACC changed.
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
INCREMENT_LWORD:
   lcall INCREMENT_WORD     ; increment lower 2 bytes
   jnc   INCREMENT_LWORD_99 ; if no carry, we're done
   lcall INCREMENT_WORD     ; increment upper 2 bytes
INCREMENT_LWORD_99:
   ret                      ; INCREMENT_LWORD
;-----------------------------------------------------------------------------   

;*****************************************************************************
; Subroutine -- MEM_COPY
; Purpose -- Copies a range of memory (CODE or XDATA) to XDATA.
; Inputs -- source address =      R3:R2
;           destination address = R5:R4
;           number of bytes =     R0:R7
;           source memory space = SRC_FLAG (0=XDATA, 1=CODE)
; Outputs -- XDATA memory changed
; Timing -- Test case:  N=1024 bytes
;              Timing = 30,859 machine cycles (46.3 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
MEM_COPY:

   ; fixup number of bytes if LSB is non-zero (to make DJNZ commands work)
   cjne  R7, #0, MEM_COPY_10
   sjmp  MEM_COPY_30
MEM_COPY_10:
   inc   R0                 ; if LSB<>0 we must increment MSB for looping

MEM_COPY_30:                ; copy data
   mov   DPH, R3            ; get source pointer
   mov   DPL, R2
   jb    SRC_FLAG, MEM_COPY_34  ; jump if source is CODE space
   ; source is XDATA
   movx  A, @DPTR           ; read byte from XDATA
   sjmp  MEM_COPY_36

MEM_COPY_34:                ; copy from CODE to XDATA
   clr   A                  ; prepare for MOVC instruction
   movc  A, @A+DPTR         ; read byte from CODE

MEM_COPY_36:                ; write the data
   inc   DPTR               ; advance source pointer and re-save it
   mov   R3, DPH
   mov   R2, DPL
   mov   DPH, R5            ; get destination pointer
   mov   DPL, R4
   movx  @DPTR, A           ; write byte to XDATA
   inc   DPTR               ; advance destination pointer and re-save it
   mov   R5, DPH
   mov   R4, DPL
   djnz  R7, MEM_COPY_30    ; loop by decrementing counter LSB
   djnz  R0, MEM_COPY_30    ; loop by decrementing counter MSB

MEM_COPY_99:
   ret                      ; MEM_COPY
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- CHECK_CMD_PACKET
; Purpose -- Verifies proper syntax of a command packet.
; Inputs -- Command packet at XDATA location 'RS17'
; Outputs -- ACC=0 if packet is OK, ACC=0FFh if error(s)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CHECK_CMD_PACKET:

   mov   DPTR, #RS17        ; point to first byte of command packet
   movx  A, @DPTR           ; load opcode
   jb    ACC.7, CHECK_CMD_PACKET_05 ; Command MS-bit must be '1'

   mov   A, #LOW(DIAG0004)  ; Diagnostic:  Command MS-bit must be '1'
   lcall report_diagnostic
   mov   A, #0FFh            ; indicate error
   sjmp  CHECK_CMD_PACKET_99 ; done

CHECK_CMD_PACKET_05:

   ; Opcode structure is special.  High and low bytes must be the same.
   movx  A, @DPTR           ; Opcode low byte
   mov   CMDBUFF+0, A
   inc   DPTR
   movx  A, @DPTR           ; Opcode high byte
   cjne  A, CMDBUFF+0, CHECK_CMD_PACKET_10   ; bad packet structure

   ; Opcode structure is OK, so check for command complements
   mov   R0, #7              ; check 7 pairs of words for complements
   mov   DPTR, #RS17         ; reset DPTR pointer
   sjmp  CHECK_CMD_PACKET_20

CHECK_CMD_PACKET_10:
   mov   A, #LOW(DIAG0005)  ; Diagnostic:  Command byte must have twin
   lcall report_diagnostic
   mov   A, #0FFh            ; indicate error
   sjmp  CHECK_CMD_PACKET_99 ; done

CHECK_CMD_PACKET_20:
   movx  A, @DPTR           ; parameter LSB
   mov   CMDBUFF+0, A
   inc   DPTR
   movx  A, @DPTR           ; parameter MSB
   mov   CMDBUFF+1, A
   inc   DPTR
   movx  A, @DPTR           ; parameter complement LSB
   mov   CMDBUFF+2, A
   inc   DPTR
   movx  A, @DPTR           ; parameter complement MSB
   inc   DPTR
   cpl   A                  ; complement the parameter complement MSB
   cjne  A, CMDBUFF+1, CHECK_CMD_PACKET_30   ; bad packet structure
   mov   A, CMDBUFF+2       ; parameter complement LSB
   cpl   A                  ; complement the parameter complement LSB
   cjne  A, CMDBUFF+0, CHECK_CMD_PACKET_30   ; bad packet structure
   djnz  R0, CHECK_CMD_PACKET_20
   mov   A, #0              ; indicate no errors
   sjmp  CHECK_CMD_PACKET_99

CHECK_CMD_PACKET_30:
   mov   A, #LOW(DIAG0006)  ; Diagnostic:  Command Parameters bad complements
   lcall report_diagnostic
   mov   A, #0FFh           ; indicate error

CHECK_CMD_PACKET_99:
   ret
;-----------------------------------------------------------------------------   

;*****************************************************************************
; Subroutine -- ENABLE_DOG
; Purpose -- Setup and enable the watchdog timer using the Programmable
;    Counter Array (PCA) module.  PCA will use the Timer0 rollover to clock
;    its timer, so Timer0 must be running (the default state in operate).
;    Calling this routine also has the effect of feeding the
;    watchdog timer (i.e., resetting its timeout to the full value).
; Inputs -- PCA registers
; Outputs -- PCA registers
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ENABLE_DOG:
watchdog_count  equ   10*ticks_per_sec
   clr   EA                 ; hold off interrupts briefly
   mov   CCON, #0           ; stop the PCA timer
   mov   CL, #0             ; clear PCA timer LSB
   mov   CH, #0             ; clear PCA timer MSB
   ; load '#LOW(watchdog_count)' ...
   mov   CCAP4L, #LOW(watchdog_count)
  ; load '#HIGH(watchdog_count)' ...
   mov   CCAP4H, #HIGH(watchdog_count)
   mov   CMOD, #44h         ; Enable WDT mode, use Timer0 rollover
   mov   CCAPM4, #48h       ; PCA module 4 as watchdog
   mov   CCON, #40h         ; start the PCA timer
   setb  WDOG_ENABLED       ; flag bit
   setb  EA                 ; re-enable interrupts
ENABLE_DOG_99:   
   ret                      ; ENABLE_DOG
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DISABLE_DOG
; Purpose -- Disable the watchdog timer using the Programmable Counter Array
;    (PCA) module.  This routine simply stops the PCA timer so that it cannot
;    timeout.
; Inputs -- PCA registers
; Outputs -- PCA registers
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DISABLE_DOG:
   mov   CMOD, #0           ; Disable WDT mode
   mov   CCAPM4, #0         ; PCA module 4 -- NO watchdog
   mov   CCON, #0           ; stop the PCA timer
   clr   WDOG_ENABLED       ; flag bit
DISABLE_DOG_99:
   ret                      ; DISABLE_DOG
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- SETUP_TO_BUILD_HKP
; Purpose -- Initializes various memory locations (e.g., status bits) prior to
;            building the housekeeping packet.
; Inputs -- none
; Outputs -- none
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
SETUP_TO_BUILD_HKP:

   ; Status Bits '0' LSB ... 
   clr   A                  ; clear ACC before calling macro
   SET_STATUS  ACC.6, LFHVENA
   mov   DPTR, #mLFSBITS0+0 ; LSB of status word 0
   movx  @DPTR, A
   ; Status Bits '0' MSB ... 
   clr   A                  ; clear ACC before calling macro
   SET_STATUS  ACC.5, LFRMENA
   mov   DPTR, #mLFSBITS0+1 ; MSB of status word 0
   movx  @DPTR, A

   ; Status Bits '1' LSB ... 
   clr   A                  ; clear ACC
   setb  ACC.3              ; indicate OPERATE mode -- LFDOPERT='1'
   mov   DPTR, #mLFSBITS1+0 ; LSB of status word 1
   movx  @DPTR, A
   ; Status Bits '1' MSB (all bits zero) ... 
   clr   A                  ; clear ACC
   mov   DPTR, #mLFSBITS1+1 ; MSB of status word 1
   movx  @DPTR, A

   ; Status Bits '2' LSB ... 
   clr   A                  ; clear ACC before calling macro
   SET_STATUS  ACC.6, LFHRMPA
   SET_STATUS  ACC.7, LFHRMPB
   mov   DPTR, #mLFSBITS2+0 ; LSB of status word 2
   movx  @DPTR, A
   ; Status Bits '2' MSB ... 
   clr   A                  ; clear ACC before calling macro
   SET_STATUS  ACC.6, WDOG_ENABLED
   mov   DPTR, #mLFSBITS2+1 ; MSB of status word 2
   movx  @DPTR, A

   ; Status Bits '3' ... Two bilevel bytes.
   mov   DPTR, #HV_STATUS   ; HV_STATUS = 4000h
   movx  A, @DPTR           ; read HV_STATUS
   mov   DPTR, #mLFSBITS3+0 ; Address of LSB of Status Bits '3'
   movx  @DPTR, A           ; write HV_STATUS to Status Bits '3' LSB
   mov   DPTR, #BILEVEL     ; BILEVEL = 4500h
   movx  A, @DPTR           ; read BILEVEL
   mov   DPTR, #mLFSBITS3+1 ; Address of MSB of Status Bits '3'
   movx  @DPTR, A           ; write BILEVEL to Status Bits '3' MSB

   ; Status Bits '4' ... CONTROLS2.
   clr   C2_LOCK            ; Unlock CONTROLS2
   mov   DPTR, #CONTROLS2   ; CONTROLS2 = 4300h
   movx  A, @DPTR           ; read CONTROLS2 value
   setb  C2_LOCK            ; Lock CONTROLS2
   mov   DPTR, #mLFSBITS4   ; Address of Status Bits '4'
   movx  @DPTR, A           ; Status Bits '4' <== CONTROLS2 reading

   ; Status Bits '5' ... CONTROLS1.
   mov   DPTR, #CONTROLS1   ; CONTROLS1 = 4200h
   movx  A, @DPTR           ; read CONTROLS1 value
   mov   DPTR, #mLFSBITS5   ; Address of Status Bits '5'
   movx  @DPTR, A           ; Status Bits '5' <== CONTROLS1 reading

   ; LFDDETID ... Concatenation of LFDDETID1 and LFDDETID0 bits.
   clr   A
   push  ACC
   mov   DPTR, #HV_STATUS   ; HV_STATUS = 4000h
   movx  A, @DPTR
   jnb   ACC.7, SETUP_TO_BUILD_HKP_20
   pop   ACC
   setb  ACC.0
   push  ACC
SETUP_TO_BUILD_HKP_20:
   mov   DPTR, #BILEVEL     ; BILEVEL = 4500h
   movx  A, @DPTR
   jnb   ACC.7, SETUP_TO_BUILD_HKP_21
   pop   ACC
   setb  ACC.1
   push  ACC
SETUP_TO_BUILD_HKP_21:
   pop   ACC
   mov   DPTR, #mLFDDETID
   movx  @DPTR, A

   ; We want the 5 housekeeping mnemonics LFDDIAG0, LFDDIAG1, etc, to show
   ; only the diagnostics stored since the last housekeeping packet was
   ; sent.  If any exist, we'll copy them to the array 'mCURR_5' for the
   ; routine 'BUILD_HKP' to then copy into housekeeping.
   ; Copy the top 'mCURR_NUM' diagnostics from the 32-element push-down
   ; stack onto the list of current 5 (up to 5) diagnostics for housekeeping.
   ; First clear the 5 locations ...
   mov   DPTR, #mCURR_5     ; top of our 'list of 5'
   mov   R6, #HIGH(10)      ; 2 bytes for each = 10 bytes to clear
   mov   R7, #LOW(10)
   lcall CLEAR_RAM
   ; Now find out how many diagnostics were stored during this 1-second
   ; interval ...
   mov   DPTR, #mCURR_NUM   ; number of diagnostics stored
   movx  A, @DPTR
   jz    SETUP_TO_BUILD_HKP_40 ; no diagnostics this 1-second interval
   add   A, #-6             ; subtract 6 (which is one too many)
   jnc   SETUP_TO_BUILD_HKP_30 ; carry indicates > 5 (too many)
   mov   A, #-1             ; prepare to add 6 ==> -1 + 6 = 5
SETUP_TO_BUILD_HKP_30:
   add   A, #6              ; ACC holds number of diagnostics to copy
   rl    A                  ; double value (2 bytes per word)
   ; Setup for call to MEM_COPY ...
   ; Number of BYTES to copy goes in R0:R7 ...
   mov   R7, A
   mov   R0, #0
   ; Source address goes in R3:R2 ...
   mov   R3, #HIGH(mLFDDIAGS)   ; top of push-down diagnostics stack
   mov   R2, #LOW(mLFDDIAGS)
   ; Destination address goes in R5:R4 ...
   mov   R5, #HIGH(mCURR_5)     ; list of 5 current diagnostics
   mov   R4, #LOW(mCURR_5)
   ; Set SRC_FLAG='0' for XDATA copy ...
   clr   SRC_FLAG
   ; make the call ...
   lcall MEM_COPY

SETUP_TO_BUILD_HKP_40:

   ; Refresh the memory monitors ...
   lcall MEMORY_MONITORS

   ; Initialize a pointer to the beginning of the housekeeping memory
   ; (for housekeeping data storage) ...
   mov   HKP_PTR+0, #HIGH(TS17) ; MSB of housekeeping packet start
   mov   HKP_PTR+1, #LOW(TS17)  ; LSB

   ; Initialize a pointer to the beginning of the housekeeping script
   ; which directs the building of the housekeeping packet ...
   mov   SCRIPT_PTR+0, #HIGH(hkp_script)  ; MSB of housekeeping script
   mov   SCRIPT_PTR+1, #LOW(hkp_script)   ; LSB

   ; Indicate that no items are left over from previous runs of the script ...
   mov   HKP_PARAMS+3, #0
   
   setb  HKP_BUILD_FLAG     ; flag housekeeping construction started
   
SETUP_TO_BUILD_HKP_99:
   ret                      ; SETUP_TO_BUILD_HKP
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- BUILD_HKP
; Purpose -- Builds the housekeeping packet in accordance with the build
;            script (see "patchcst.a51" for the script and "common.inc" for
;            a list of the script opcodes.)
; Inputs -- SCRIPT_PTR+0:SCRIPT_PTR+1 points to the next line of the
;           housekeeping script.
;           HKP_PARAMS+0, +1, +2, & +3 hold the 4 bytes of the current line
;           of the script.  If HKP_PARAMS+3 (the repetition counter) is zero,
;           it is time to read a new line of the script.
;           HKP_PTR+0:HKP_PTR+1 points to the address within the
;           housekeeping buffer at which to write the next piece of data
;           collected by this routine.
;           If a particular line of the housekeeping script has a repetition
;           counter greater than 1 (i.e., write more than 1 byte to
;           housekeeping), the rep. counter is decremented once for each
;           data item collected and the HKP_PTR is incremented once.
;           NOTE:  When preparing to start housekeeping building, making
;           HKP_PARAMS+3 zero will force a reading of the next (i.e. first)
;           line of the housekeeping script.  Of course, both SCRIPT_PTR
;           and HKP_PTR should have been set up prior to this first call.
;           NOTE:  Since each line of housekeeping script may write up to
;           256 bytes into housekeeping, it is possible to scribble beyond
;           the end of the housekeeping area by as much as 255 bytes (if
;           the starting address were at the end of the housekeeping area
;           and 256 bytes were requested).
; Outputs -- The housekeeping buffer is populated in accordance with the
;            opcodes and parameters found in the housekeeping script.
; Timing -- (default housekeeping packet) ...
;              ~46,900 machine cycles ( 70.4 msec @ 8-MHz)
;*****************************************************************************
$SET (STAY_IN_BUILD_HKP=1)

FILL_TO_BOUNDARY   16, 0FFh
BUILD_HKP:

$if STAY_IN_BUILD_HKP
BUILD_HKP_NEXT:
   ; Reasons to leave this background task prior to completion ...
   jb    RXD_FLAG,    BUILD_HKP_10  ; opcode received
   ; No reasons found to leave this background task, so continue ...
   sjmp  BUILD_HKP_20
BUILD_HKP_10:
   ljmp  BUILD_HKP_99                                           
BUILD_HKP_20:
$endif

   ; If more items remain in a multi-item reading, do not get more data
   ; from the housekeeping script.  Simply jump over the lines below
   ; which read from the script and continue processing the current line,
   ; which is stored in HKP_PARAMS+0, +1, +2, and +3.
   mov   A, HKP_PARAMS+3
   cjne  A, #0, BUILD_HKP_30

   ; Load current address of housekeeping script pointer into DPTR for
   ; reading the next script line ...
   mov   DPH, SCRIPT_PTR+0  ; restore MSB of pointer to housekeeping script
   mov   DPL, SCRIPT_PTR+1  ; restore LSB
   
   ; Read the next line of the script ...
   movx  A, @DPTR           ; read script opcode
   mov   HKP_PARAMS+0, A    ; store script opcode
   inc   DPTR
   movx  A, @DPTR           ; read script parameter byte 1
   mov   HKP_PARAMS+1, A    ; store param 1
   inc   DPTR
   movx  A, @DPTR           ; read script parameter byte 2
   mov   HKP_PARAMS+2, A    ; store param 2
   inc   DPTR
   movx  A, @DPTR           ; read script parameter byte 3 (the repeat count)
   mov   HKP_PARAMS+3, A    ; store param 3
   inc   DPTR               ; advance DPTR to start of next script line
   
   ; Save script pointer in SCRIPT_PTR+0:SCRIPT_PTR+1 for next iteration.
   mov   SCRIPT_PTR+0, DPH  ; store MSB of pointer to housekeeping script
   mov   SCRIPT_PTR+1, DPL  ; store LSB

BUILD_HKP_30:

   ; Ensure the next housekeeping pointer is in the range TS17 to
   ; TS17+03FFh, inclusive.  To do this, subtract TS17 from the desired
   ; address and ensure the result is in the range 000-3FFh (i.e., that
   ; resulting MSB is less than 4) ...
   clr   C                  ; prepare for SUBB operation
   mov   A, HKP_PTR+1       ; LSB of next housekeeping address
   subb  A, #LOW(TS17)      ; no need to store value
   mov   A, HKP_PTR+0       ; MSB of next housekeeping address
   subb  A, #HIGH(TS17)
   jc    BUILD_HKP_001      ; Carry=1 --> new address LESS than TS17 (too low)
   subb  A, #4              ; verify MSB is less than 04h
   jnc   BUILD_HKP_001      ; Carry=0 --> new address too high
   ; new address is OK, so proceed ...

   ; Start a large CASE statement based on the opcode, held in HKP_PARAMS+0
   mov   A, HKP_PARAMS+0

   ;--------------------------------------------------------------------------
BUILD_HKP_00:               ; OPCODE = 0
   cjne  A, #0, BUILD_HKP_00_END ; jump if opcode <> 0
   ; opcode=0 --> done ...
BUILD_HKP_001:              ; entry point for leaving this routine
   clr   HKP_BUILD_FLAG     ; housekeeping construction is done
   setb  SEND_HKP_FLAG      ; time to send housekeeping buffer
   ljmp  BUILD_HKP_99       ; exit
BUILD_HKP_00_END:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
BUILD_HKP_01:               ; OPCODE = 1
   cjne  A, #1, BUILD_HKP_01_END ; jump if opcode <> 1
   ; opcode=1 --> read CODE space and copy into housekeeping packet
   ; Prepare registers for call to MEM_COPY.
   setb  SRC_FLAG           ; SRC_FLAG=1 --> copy from CODE space
   sjmp  BUILD_HKP_021      ; same code as XDATA copy (except SRC_FLAG)
BUILD_HKP_01_END:
   ;--------------------------------------------------------------------------
   
   ;--------------------------------------------------------------------------
BUILD_HKP_02:               ; OPCODE = 2
   cjne  A, #2, BUILD_HKP_02_END ; jump if opcode <> 2
   ; opcode=2 --> read XDATA space and copy into housekeeping packet
   ; Prepare registers for call to MEM_COPY.
   clr   SRC_FLAG           ; SRC_FLAG=0 --> copy from XDATA space

BUILD_HKP_021:   
   ; Put Source address in R3:R2 ...
   mov   R3, HKP_PARAMS+1
   mov   R2, HKP_PARAMS+2
   ; Put Destination address in R5:R4 ...
   mov   R5, HKP_PTR+0      ; MSB
   mov   R4, HKP_PTR+1      ; LSB
   ; Number of bytes to copy in R0:R7 ...
   mov   R0, #0
   mov   R7, HKP_PARAMS+3
   ; Parameter "SRC_FLAG" was already set/cleared (above)

   ; If the source address (R3:R2) is in the I/O space, unlock/enable
   ; appropriate devices (if reading such warrants this action) ...
BUILD_HKP_0221:
   cjne  R3, #HIGH(PH_A), BUILD_HKP_0222
   setb  PHA_DIS            ; Enable reading of PH_A (by disabling ACTEL)
BUILD_HKP_0222:
   cjne  R3, #HIGH(PH_B), BUILD_HKP_0223
   setb  PHA_DIS            ; Enable reading of PH_B (by disabling ACTEL)
BUILD_HKP_0223:
   cjne  R3, #HIGH(CONTROLS2), BUILD_HKP_0224
   clr   C2_LOCK            ; Unlock CONTROLS2
BUILD_HKP_0224:

   ; Copy the values ...
BUILD_HKP_0230:
   lcall MEM_COPY
   
   ; Lock/disable PH_A, PH_B, and CONTROLS2, regardless
   ; of whether we unlocked/enabled them above (it doesn't hurt) ...
   clr   PHA_DIS            ; Disable reading PH_A & PH_B (reenable ACTEL)
   setb  C2_LOCK            ; Lock CONTROLS2

   ; update destination address -- HKP_PTR+0:HKP_PTR+1
   mov   HKP_PTR+0, R5      ; store updated HKP address MSB
   mov   HKP_PTR+1, R4      ; store updated HKP address LSB
   mov   HKP_PARAMS+3, #0   ; flag this line of the script as done.
   ljmp  BUILD_HKP_NEXT     ; loop to top
BUILD_HKP_02_END:
   ;--------------------------------------------------------------------------
   
   ;--------------------------------------------------------------------------
BUILD_HKP_03:               ; OPCODE = 3
   cjne  A, #3, BUILD_HKP_03_END ; jump if opcode <> 3
   ; opcode=3 --> read DATA (internal) RAM and copy into housekeeping packet
   ; Put source address in R0 for indirect addressing ...
   mov   R0, HKP_PARAMS+2
   ; Put Destination address in DPTR ...
   mov   DPH, HKP_PTR+0     ; MSB
   mov   DPL, HKP_PTR+1     ; LSB
   ; Number of bytes is in 'HKP_PARAMS+3'
BUILD_HKP_031:
   mov   A, @R0             ; read data byte (indirect read)
   movx  @DPTR, A           ; save
   inc   DPTR               ; point to next destination address
   inc   R0                 ; point to next DATA item
   djnz  HKP_PARAMS+3, BUILD_HKP_031  ; loop until done
   ; update destination address -- HKP_PTR+0:HKP_PTR+1
   mov   HKP_PTR+0, DPH     ; store updated HKP address MSB
   mov   HKP_PTR+1, DPL     ; store updated HKP address LSB
   ljmp  BUILD_HKP_NEXT     ; loop to top
BUILD_HKP_03_END:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
BUILD_HKP_06:               ; OPCODE = 6
   cjne  A, #6, BUILD_HKP_06_END ; jump if opcode <> 6
   ; opcode=6 --> Analog reading (w/ MUX setup and delay)

   ; This housekeeping opcode utilized the A/D converter, as does the ILIM
   ; task.  Therefore, we must prevent interruptions by ILIM (invoked by the
   ; Timer0 ISR) by temporarily disabling Timer0 interrupts.
   clr   ET0                ; disable Timer0 interrupt

   ; Setup MUX for A/D reading ...
   mov   A, HKP_PARAMS+1    ; device MUX address for this sensor
   mov   DPTR, #MUX         ; I/O port for writing MUX address
   movx  @DPTR, A           ; write MUX address to port

BUILD_HKP_06_ENTRY_POINT:   ; entry point for other A/D opcodes

   ; Read analog value (return value in ACC) ...
   lcall ADC_READ

   ; Write the sensor value to housekeeping packet ...
   mov   DPH, HKP_PTR+0     ; recall HKP address
   mov   DPL, HKP_PTR+1
   movx  @DPTR, A           ; write the value to HKP
   inc   DPTR               ; advance HKP DPTR
   mov   HKP_PTR+0, DPH     ; store updated HKP address MSB
   mov   HKP_PTR+1, DPL     ; store updated HKP address LSB

   ; Prepare for next A/D reading by advancing HKP_PARAMS values ...
   inc   HKP_PARAMS+1       ; increment MUX address
   dec   HKP_PARAMS+3       ; decrement Number of Bytes to read
   ; Opcode and delay time parameters remain unchanged.

   setb  ET0                ; re-enable Timer0 interrupt
   ljmp  BUILD_HKP_NEXT     ; loop to top
BUILD_HKP_06_END:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
BUILD_HKP_07:               ; OPCODE = 7
   cjne  A, #7, BUILD_HKP_07_END ; jump if opcode <> 7
   ; opcode=7 --> move housekeeping pointer to any absolute location within
   ; the housekeeping buffer.

BUILD_HKP_07_ENTRY_POINT:
   ; Ensure the requested housekeeping pointer is in the range TS17 to
   ; TS17+03FFh, inclusive.  To do this, subtract TS17 from the desired
   ; address and ensure the result is in the range 000-3FFh (i.e., that
   ; resulting MSB is less than 4) ...
   clr   C                  ; prepare for SUBB operation
   mov   A, HKP_PARAMS+2    ; LSB of requested address
   subb  A, #LOW(TS17)      ; no need to store value
   mov   A, HKP_PARAMS+1    ; MSB of requested address
   subb  A, #HIGH(TS17)
   jc    BUILD_HKP_071      ; Carry=1 --> new address LESS than TS17 (too low)
   subb  A, #4              ; verify MSB is less than 04h
   jnc   BUILD_HKP_071      ; Carry=0 --> new address too high
   ; new address is OK, so store it ...
   mov   HKP_PTR+0, HKP_PARAMS+1  ; store new MSB
   mov   HKP_PTR+1, HKP_PARAMS+2  ; store new LSB
BUILD_HKP_071:
   mov   HKP_PARAMS+3, #0   ; flag this line of the script as done.
   ljmp  BUILD_HKP_NEXT     ; loop to top
BUILD_HKP_07_END:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
BUILD_HKP_08:               ; OPCODE = 8
   cjne  A, #8, BUILD_HKP_08_END ; jump if opcode <> 8
   ; opcode=8 --> skip space in the housekeeping buffer.  This is a relative
   ; change in the housekeeping buffer pointer, advancing it by the number of
   ; bytes found in HKP_PARAMS+3.
   
   ; In order to take advantage of the range checking performed by opcode 7
   ; (absolute pointer move), we will use that same code by changing the
   ; supplied parameters from relative to absolute and then call opcode 7.
   mov   A, HKP_PARAMS+3    ; number of byte to move ahead in buffer
   add   A, HKP_PTR+1       ; add LSB of current buffer pointer
   mov   HKP_PARAMS+2, A    ; store LSB as HKP_PARAMS+2 (absolute address LSB)
   mov   A, HKP_PTR+0
   addc  A, #0              ; propagate carry bit to MSB
   mov   HKP_PARAMS+1, A    ; store MSB as HKP_PARAMS+1 (absolute address MSB)
   ljmp  BUILD_HKP_07_ENTRY_POINT ; jump to opcode 7 processing
BUILD_HKP_08_END:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
BUILD_HKP_09:               ; OPCODE = 9
   cjne  A, #9, BUILD_HKP_09_END ; jump if opcode <> 9
   ; opcode=9 --> zero-out space in the housekeeping buffer
   mov   DPH, HKP_PTR+0     ; pointer into housekeeing buffer
   mov   DPL, HKP_PTR+1
   clr   A
BUILD_HKP_091:
   movx  @DPTR, A           ; clear current location
   inc   DPTR
   djnz  HKP_PARAMS+3, BUILD_HKP_091  ; loop until done
   ; update destination address -- HKP_PTR+0:HKP_PTR+1
   mov   HKP_PTR+0, DPH     ; store updated HKP address MSB
   mov   HKP_PTR+1, DPL     ; store updated HKP address LSB
   ljmp  BUILD_HKP_NEXT     ; loop to top
BUILD_HKP_09_END:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
BUILD_HKP_46:               ; OPCODE = 46 (hex)
   cjne  A, #46h, BUILD_HKP_46_END ; jump if opcode <> 46h
   ; opcode=46 (hex) --> TDC 'A' X-Analog reading (w/ 2 MUX setups)

   ; Temporarily prevent Timer0 ISR calls since we need the A/D converter
   ; here ...
   clr   ET0                ; disable Timer0 interrupt

   ; Setup TDC MUX by writing out 3-wire interface ...
   ; Put TDC MUX select value in R0.
   mov   R0, HKP_PARAMS+1
   lcall TDC_A              ; write R0:R1 to TDC 'A' (R1 irrelevant)

   ; Setup MUX for A/D reading ...
   mov   A, HKP_PARAMS+0    ; device MUX address for this sensor
   mov   DPTR, #MUX         ; I/O port for writing MUX address
   movx  @DPTR, A           ; write MUX address to port

   ; opcode 6 has all the A/D setup and delay code, so simply jump to that ...
   ljmp  BUILD_HKP_06_ENTRY_POINT
BUILD_HKP_46_END:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
BUILD_HKP_82:               ; OPCODE = 82 (hex)
   cjne  A, #82h, BUILD_HKP_82_END ; jump if opcode <> 82h
   ; opcode=82 (hex) --> TDC 'B' X-Analog reading (w/ 2 MUX setups)

   ; Temporarily prevent Timer0 ISR calls since we need the A/D converter
   ; here ...
   clr   ET0                ; disable Timer0 interrupt

   ; Setup TDC MUX by writing out 3-wire interface ...
   ; Put TDC MUX select value in R0.
   mov   R0, HKP_PARAMS+1
   lcall TDC_B              ; write R0:R1 to TDC 'B' (R1 irrelevant)

   ; Setup MUX for A/D reading ...
   mov   A, HKP_PARAMS+0    ; device MUX address for this sensor
   mov   DPTR, #MUX         ; I/O port for writing MUX address
   movx  @DPTR, A           ; write MUX address to port

   ; opcode 6 has all the A/D setup and delay code, so simply jump to that ...
   ljmp  BUILD_HKP_06_ENTRY_POINT
BUILD_HKP_82_END:
   ;--------------------------------------------------------------------------

BUILD_HKP_BAD_OPCODE:
   ; If we have an improper script opcode, we will reach this point
   ; and treat it as an end-of-script condition (opcode=0) ...
   clr   HKP_BUILD_FLAG     ; housekeeping construction is done
   setb  SEND_HKP_FLAG      ; time to send housekeeping buffer
   ljmp  BUILD_HKP_99       ; exit

$if (STAY_IN_BUILD_HKP=0)
BUILD_HKP_NEXT:
$endif

BUILD_HKP_99:
   ret                      ; BUILD_HKP
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- SEND_DUMP_PKT, SEND_HKP_PKT
; Purpose -- Entry points to routines that send 512 32-bit words of memory
;            dump data or housekeeping data.  Only 512 16-bit words of data
;            are pointed to.  The other 16 bits of each long word consist of
;            the address of the word being sent.
; Inputs -- none
; Outputs -- none
; Regisers used -- R0, R1, DPTR
; Timing -- 22,642 machine cycles (34.0 msec at 8 MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
SEND_DUMP_PKT:
   mov   DPTR, #TS01        ; point to the dump packet block of data
   sjmp  SEND_DUMP_PKT_10   ; jump to sending loop

SEND_HKP_PKT:
   mov   DPTR, #TS17        ; point to the housekeeping packet block of data

   ; Send 512 words (1024 bytes).  Loop 512 times ...
SEND_DUMP_PKT_10:
   mov   R0, #HIGH(512)     ; loop counter MSB
   mov   R1, #LOW(512)      ; loop counter LSB
SEND_DUMP_PKT_30:
   lcall SEND_PKT           ; DPTR points to value
   inc   DPTR               ; SEND_PKT sent two bytes of data so we advance
   inc   DPTR               ;    pointer by two.
   djnz  R1, SEND_DUMP_PKT_30   ; decrement counter LSB
   djnz  R0, SEND_DUMP_PKT_30   ; decrement counter MSB

SEND_DUMP_PKT_99:
   ret                      ; SEND_DUMP_PKT & SEND_HKP_PKT
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- SEND_PKT
; Purpose -- Send one 4-byte packet from COS DCE to CS computer
; Called by -- SEND_DUMP_PKT & SEND_HKP_PKT
; Inputs -- DPTR points to 16-bit value to send
; Outputs -- none
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
SEND_PKT:
   mov   HKPBUFF+2, DPL     ; Address LSB stored while changing DPTR
   mov   HKPBUFF+3, DPH     ; Address MSB stored while changing DPTR
   movx  A, @DPTR           ; Fetch the data LSB
   mov   HKPBUFF+0, A       ; save LSB
   inc   DPTR               ; point to upper data byte
   movx  A, @DPTR           ; Fetch the data MSB
   mov   HKPBUFF+1, A       ; save MSB

   ; The transmitter is four byte registers.  When a value is written to
   ; address 5300h, the transmitter will start shifting out the 32 bit word.
   ; Bytes may be written to the registers in any order, so long as the one
   ; at address 5300h is written last.  The housekeeping address is written
   ; to 5300:5200, the data written to 5100:5000.
   
   mov   A, HKPBUFF+0       ; fetch data LSB
   mov   DPTR, #TXD_HARDWARE  ; base transmit register (5000h)
   movx  @DPTR, A           ; write data LSB
   mov   A, HKPBUFF+1       ; fetch data MSB
   inc   DPH                ; 5100h is an xmit register
   movx  @DPTR, A           ; write data MSB
   mov   A, HKPBUFF+2       ; fetch address LSB
   inc   DPH                ; 5200h is an xmit register
   movx  @DPTR, A           ; write address LSB
   mov   A, HKPBUFF+3       ; fetch address MSB
;   anl   A, #1Fh            ; map to 0000h..1FFFh
   inc   DPH                ; 5300h is MSB xmit register
   movx  @DPTR, A           ; write address MSB and cause transfer to begin

   ; Restore DPTR which points to word just transmitted
   mov   DPL, HKPBUFF+2     ; restore pointer
   mov   DPH, HKPBUFF+3     ; restore pointer

SEND_PKT_99:
   ret                      ; SEND_PKT
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- SAFE_CONTROLS
; Purpose -- Places the hardware in a known, safe state.
; Inputs -- none
; Outputs -- none
; Timing -- 482 machine cycles (0.723 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
SAFE_CONTROLS:

   ; 'RAM_BANK' is the most-significant address bit of DCE-B RAM chip U8.
   ; Set it to '1' and don't change throughout the code ...
   setb  RAM_BANK

   clr   PHA_DIS            ; enable ACTEL to gather and store PH data

   ; Reset the ACTEL devices on the TDCs and DCE-A by giving them a low pulse
   ; of at least 10us ...
   clr   ACTEL_RESET        ; active low reset
   mov   R0, #3             ; Loop for 10 us
   djnz  R0, $
   setb  ACTEL_RESET        ; enable the ACTEL devices on the TDCs and DCE-A

   ; Turn off (reset to zero) all bits of CONTROLS1 and CONTROLS2 ...
   clr   C2_LOCK            ; make sure CONTROLS2 is unlocked
   mov   A, #0
   mov   DPTR, #CONTROLS1
   movx  @DPTR, A           ; write 00h to CONTROLS1 register
   mov   DPTR, #CONTROLS2
   movx  @DPTR, A           ; write 00h to CONTROLS2 register
   setb  C2_LOCK            ; make sure CONTROLS2 is locked
   
   ; Initialize the 3-wire interface to a safe state ...
   setb  TDC_CLOCK          ; P1.0, TDC 'A' & 'B' clock
   setb  TDC_DATA           ; P1.1, TDC 'A' & 'B' data out
   clr   TDC_A_ENABLE       ; P1.2, TDC 'A' Active high enable
   clr   TDC_B_ENABLE       ; P1.3, TDC 'B' Active high enable

   ; shut off STIMS ...
   mov   R0, #70h           ; Stim DAC address on TDC
   mov   R1, #0             ; shut off
   lcall TDC_A              ; Segment A Stims
   mov   R0, #70h           ; Stim DAC address on TDC
   mov   R1, #0             ; shut off
   lcall TDC_B              ; Segment B Stims

   ; Reset VMAX and VSET for both Segments ...
   mov   A, #0              ; zero out the settings
   mov   DPTR, #VSETA
   movx  @DPTR, A
   mov   DPTR, #VSETB
   movx  @DPTR, A
   mov   DPTR, #VMAXA
   movx  @DPTR, A
   mov   DPTR, #VMAXB
   movx  @DPTR, A

   ; Set the software voltage parameters to zero as well ...
   mov   A, #0              ; zero out the settings
   mov   DPTR, #hv_segA+3   ; LFHVTGTA
   movx  @DPTR, A
   inc   DPTR               ; LFHVDACA
   movx  @DPTR, A
   mov   DPTR, #hv_segB+3   ; LFHVTGTB
   movx  @DPTR, A
   inc   DPTR               ; LFHVDACB
   movx  @DPTR, A

   ; Disable voltage commands ...
   clr   LFHVENA

   ; Set the voltage state for housekeeping ...
   mov   DPTR, #hv_segAB+0
   mov   A, #0              ; LFHSTATE = 0
   movx  @DPTR, A

SAFE_CONTROLS_99:
   ret                      ; SAFE_CONTROLS
;-----------------------------------------------------------------------------

$if 0
;*****************************************************************************
; Subroutine -- INIT_HARDWARE
; Purpose -- Places the hardware in a known, initial state.
; Inputs -- none
; Outputs -- none
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
INIT_HARDWARE:
   ; Issue the following digitizer commands ...

   ; LFGLQT  0A, 0 ...
   mov   R0, #21h           ; TDC MUX address
   mov   R1, #0Ah           ; setting
   lcall TDC_A
   ; LFGLQT  0A, 1 ...
   mov   R0, #21h           ; TDC MUX address
   mov   R1, #0Ah           ; setting
   lcall TDC_B
   ; LFGUQT  FF, 0 ...
   mov   R0, #22h           ; TDC MUX address
   mov   R1, #0FFh          ; setting
   lcall TDC_A
   ; LFGUQT  FF, 1 ...
   mov   R0, #22h           ; TDC MUX address
   mov   R1, #0FFh          ; setting
   lcall TDC_B
   ; LFGBWK  64, 0, 0 ...
   mov   R0, #12h           ; TDC MUX address
   mov   R1, #64h           ; setting
   lcall TDC_A
   ; LFGBWK  64, 0, 1 ...
   mov   R0, #32h           ; TDC MUX address
   mov   R1, #64h           ; setting
   lcall TDC_A
   ; LFGBWK  64, 1, 0 ...
   mov   R0, #12h           ; TDC MUX address
   mov   R1, #64h           ; setting
   lcall TDC_B
   ; LFGBWK  64, 1, 1 ...
   mov   R0, #32h           ; TDC MUX address
   mov   R1, #64h           ; setting
   lcall TDC_B
   ; LFGEWK  64, 0, 0 ...
   mov   R0, #13h           ; TDC MUX address
   mov   R1, #64h           ; setting
   lcall TDC_A
   ; LFGEWK  64, 0, 1 ...
   mov   R0, #33h           ; TDC MUX address
   mov   R1, #64h           ; setting
   lcall TDC_A
   ; LFGEWK  64, 1, 0 ...
   mov   R0, #13h           ; TDC MUX address
   mov   R1, #64h           ; setting
   lcall TDC_B
   ; LFGEWK  64, 1, 1 ...
   mov   R0, #33h           ; TDC MUX address
   mov   R1, #64h           ; setting
   lcall TDC_B
   ; LFGSHFT  80, 0, 0 ...
   mov   R0, #10h           ; TDC MUX address
   mov   R1, #80h           ; setting
   lcall TDC_A
   ; LFGSHFT  80, 0, 1 ...
   mov   R0, #30h           ; TDC MUX address
   mov   R1, #80h           ; setting
   lcall TDC_A
   ; LFGSHFT  80, 1, 0 ...
   mov   R0, #10h           ; TDC MUX address
   mov   R1, #80h           ; setting
   lcall TDC_B
   ; LFGSHFT  80, 1, 1 ...
   mov   R0, #30h           ; TDC MUX address
   mov   R1, #80h           ; setting
   lcall TDC_B
   ; LFGSTR  80, 0, 0 ...
   mov   R0, #11h           ; TDC MUX address
   mov   R1, #80h           ; setting
   lcall TDC_A
   ; LFGSTR  80, 0, 1 ...
   mov   R0, #31h           ; TDC MUX address
   mov   R1, #80h           ; setting
   lcall TDC_A
   ; LFGSTR  80, 1, 0 ...
   mov   R0, #11h           ; TDC MUX address
   mov   R1, #80h           ; setting
   lcall TDC_B
   ; LFGSTR  80, 1, 1 ...
   mov   R0, #31h           ; TDC MUX address
   mov   R1, #80h           ; setting
   lcall TDC_B
   ; LFGTT  DC, 0, 0 ...
   mov   R0, #23h           ; TDC MUX address
   mov   R1, #80h           ; setting
   lcall TDC_A
   ; LFGTT  DC, 0, 1 ...
   mov   R0, #20h           ; TDC MUX address
   mov   R1, #80h           ; setting
   lcall TDC_A
   ; LFGTT  DC, 1, 0 ...
   mov   R0, #23h           ; TDC MUX address
   mov   R1, #80h           ; setting
   lcall TDC_B
   ; LFGTT  DC, 1, 1 ...
   mov   R0, #20h           ; TDC MUX address
   mov   R1, #80h           ; setting
   lcall TDC_B

INIT_HARDWARE_99:
   ret                      ; INIT_HARDWARE
;-----------------------------------------------------------------------------
$endif

;*****************************************************************************
; Subroutine -- INIT_RAM
; Purpose -- Clears internal & external RAM.
; Inputs -- none
; Outputs -- Various internal and external RAM cleared.
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
INIT_RAM:
   ; Clear internal RAM (DSEG and BSEG) up to bottom of the STACK ...
   clr   C
   mov   A, #_STACK+1       ; first address used by STACK segment
   subb  A, #20h            ; first address used by general purpose RAM
   mov   R7, A              ; number of bytes to clear
   mov   R0, #20h
INIT_RAM_00:
   mov   @R0, #0
   inc   R0
   djnz  R7, INIT_RAM_00

   ; Clear the upload data area (1024 bytes starting at address RS01)
   ; using 'CLEAR_RAM' ...
   mov   DPTR, #RS01
   mov   R6, #HIGH(1024)
   mov   R7, #LOW(1024)
   lcall CLEAR_RAM

   ; Clear the command packet area (28 bytes starting at address RS17)
   ; using 'CLEAR_RAM' ...
   mov   DPTR, #RS17
   mov   R6, #HIGH(28)
   mov   R7, #LOW(28)
   lcall CLEAR_RAM

   ; Clear the download data area (1024 bytes starting at address TS01)
   ; using 'CLEAR_RAM' ...
   mov   DPTR, #TS01
   mov   R6, #HIGH(1024)
   mov   R7, #LOW(1024)
   lcall CLEAR_RAM

   ; Clear the housekeeping packet area (1024 bytes starting at address TS17)
   ; using 'CLEAR_RAM' ...
   mov   DPTR, #TS17
   mov   R6, #HIGH(1024)
   mov   R7, #LOW(1024)
   lcall CLEAR_RAM

   ; Clear the scratch variable space in XDATA using 'CLEAR_RAM' ...
   mov   DPTR, #start_scratch_seg
   mov   R6, #HIGH(length_scratch_seg)
   mov   R7, #LOW(length_scratch_seg)
   lcall CLEAR_RAM

   ; Clear the history buffer space in XDATA using 'CLEAR_RAM' ...
   mov   DPTR, #start_history_seg
   mov   R6, #HIGH(length_history_seg)
   mov   R7, #LOW(length_history_seg)
   lcall CLEAR_RAM

INIT_RAM_99:
   ret                      ; INIT_RAM   
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- CLEAR_RAM
; Purpose -- Clears a specified range of External RAM (XDATA)
; Inputs -- source address in DPTR (first address cleared)
;           number of bytes = R6:R7
; Outputs -- Specified XDATA memory set to zero.
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CLEAR_RAM:
   ; fixup number of bytes if LSB is non-zero (to make DJNZ commands work)
   cjne  R7, #0, CLEAR_RAM_10
   sjmp  CLEAR_RAM_30
CLEAR_RAM_10:
   inc   R6                 ; if LSB<>0 we must increment MSB for looping

CLEAR_RAM_30:
   clr   A                  ; value to write to memory is zero

CLEAR_RAM_40:
   movx  @DPTR, A           ; write byte to XDATA
   inc   DPTR               ; advance destination pointer and re-save it
   djnz  R7, CLEAR_RAM_40   ; loop by decrementing counter LSB
   djnz  R6, CLEAR_RAM_40   ; loop by decrementing counter MSB

CLEAR_RAM_99:
   ret                      ; CLEAR_RAM
;-----------------------------------------------------------------------------
   
;*****************************************************************************
; Subroutine -- REPORT_DIAGNOSTIC
; Purpose -- Generates diagnostic and possibly HST error.
; Inputs -- Diagnostic code in ACC
; Outputs -- Diagnostic on 32-level push-down stack.  HST error and parameter
;            (if applicable) on 8-level error stack and parameter stack.
; Registers -- R0, R1, ACC, B, PSW, DPTR
; Timing -- 876 machine cycles for diagnostic alone (1.31 msec @ 8-MHz)
;           968 machine cycles for diag. + HST error (1.45 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
REPORT_DIAGNOSTIC:

   ; Only report diagnostics as HST errors if the table 'HSTERR_TABLE' has
   ; a non-null pointer at the 2-byte location corresponding to the
   ; diagnostic code specified (i.e., at the WORD beginning at address
   ; 'HSTERR_TABLE + 2*ACC')
   push  ACC                   ; save diagnostic code for later use
   mov   DPTR, #HSTERR_TABLE   ; table of HST error codes
   lcall XDATA_TLU_WORD        ; 2-byte look-up value returned in B:A
   ; if either A<>0 or B<>0, this is a non-null pointer and we save this
   ; diagnostic as an HST error ...
   jnz   REPORT_DIAGNOSTIC_10  ; non-null --> save the HST error
   mov   A, B
   jnz   REPORT_DIAGNOSTIC_10  ; non-null --> save the HST error
   sjmp  REPORT_DIAGNOSTIC_50  ; null pointer --> save only as a diagnostic

REPORT_DIAGNOSTIC_10:
   ; Note:  The diagnostic code resides on the 8051 stack.

   ; Save the diagnostic as an HST error.

   ; Look for the first unfilled location of the HST error buffer.  An
   ; unfilled location is a zero value.  If the buffer is filled, the
   ; last (8th) HST error is overwritten.
   mov   R0, #8             ; max number of hst errors allowed per second
   mov   R1, #-1            ; count-up value
   mov   DPTR, #mLFDERR-1   ; point 1 byte prior to HST error array

REPORT_DIAGNOSTIC_20:
   inc   DPTR               ; increment pointe into HST error array
   inc   R1
   movx  A, @DPTR           ; retrieve value held in HST error array
   jz    REPORT_DIAGNOSTIC_30 ; space available when zero
   djnz  R0, REPORT_DIAGNOSTIC_20 ; keep looking until we hit the end of list
   
   ; If more than 8 HST errors occurred during this 1-second period,
   ; we throw away the additional ones ...
   sjmp  REPORT_DIAGNOSTIC_50  ; save only as a diagnostic
   
REPORT_DIAGNOSTIC_30:
   pop   ACC                ; retrieve diagnostic code from 8051 stack
   movx  @DPTR, A           ; write diagnostic code to HST error array

REPORT_DIAGNOSTIC_40:
   ; Note:  The diagnostic code DOES NOT reside on the 8051 stack.  Only
   ;        in ACC.

   ; Perform a table look-up to find the HST error code associated
   ; with this diagnostic ...
   push  ACC                ; save diagnostic code for later use

   mov   DPTR, #HSTERR_TABLE ; table of HST error codes
   lcall XDATA_TLU_WORD      ; 2-byte look-up value returned in B:A
   mov   DPH, B
   mov   DPL, A
   clr   A                  ; zero index
   lcall XDATA_TLU_WORD     ; 2-byte look-up value returned in B:A
   push  ACC                ; push parameter MSB onto 8051 stack
   push  B                  ; push parameter LSB onto 8051 stack
   mov   DPTR, #mLFDERRP    ; base address of error parameter buf. 'mLFDERRP'
   mov   A, R1              ; offset into 'mLFDERRP' (from above)
   rl    A                  ; double it for word offsets (vice byte offsets)
   add   A, DPL             ; add offset to DPTR
   mov   DPL, A
   mov   A, DPH
   addc  A, #0              ; propagate the carry bit
   mov   DPH, A
   pop   ACC                ; pop parameter LSB off 8051 stack
   movx  @DPTR, A           ; Store LSB in location 'mLFDERRP'+2*R1
   inc   DPTR
   pop   ACC                ; pop parameter MSB off 8051 stack
   movx  @DPTR, A           ; Store MSB in location 'mLFDERRP'+2*R1 + 1

REPORT_DIAGNOSTIC_50:
   ; Note:  The diagnostic code resides on the 8051 stack.
   
   ; Put this new diagnostic on the 32-element push-down stack.
   
   ; Using (i) as subscript notation, we will put the most-recent
   ; diagnostic -- "diagnostic(i)" -- and the serial number corresponding
   ; to it -- "serial_number(i)" -- on the 8051 stack.  Older elements on
   ; the push-down stack are subscripted (i-1), (i-2), etc.

   ; "diagnostic(i)" still resides on the 8051 stack.
   pop   ACC                ; retrieve "diagnostic(i)" from 8051 stack
   mov   B, A               ; save "diagnostic(i)" temporarily

   ; First, update parameter 'mDIAG_NUM', the highest diagnostic serial number
   mov   DPTR, #mDIAG_NUM
   movx  A, @DPTR           ; retrieve "serial_number(i-1)"
   inc   A                  ; increment serial number --> "serial_number(i)"
   movx  @DPTR, A           ; store it
   
   ; Prepare for looping operation below ...
   push  ACC                ; push "serial_number(i)" on 8051 stack
   push  B                  ; push "diagnostic(i)" on 8051 stack

   ; Begin the push-down process and repeat 32 times ...
   mov   DPTR, #mLFDDIAGS   ; point to top of push-down stack
   mov   R0, #32            ; number of push-down operations
REPORT_DIAGNOSTIC_60:   
   movx  A, @DPTR           ; read "diagnostic(i-1)" from push-down
   mov   B, A               ; save it temporarily
   pop   ACC                ; retrieve "diagnostic(i)" from 8051 stack
   movx  @DPTR, A           ; put "diagnostic(i)" on push-down stack
   inc   DPTR               ; point to "serial_number(i-1)" on push-doan
   movx  A, @DPTR           ; read "serial_number(i-1)" from push-down
   mov   R1, A              ; save it temporarily
   pop   ACC                ; retrieve "serial_number(i)" from 8051 stack
   movx  @DPTR, A           ; put "serial_number(i)" on push-down stack
   mov   A, R1              ; mov "serial_number(i-1)" to ACC
   push  ACC                ; push "serial_number(i-1)" on 8051 stack
   push  B                  ; push "diagnostic(i-1)" on 8051 stack
   inc   DPTR               ; prepare for next loop
   djnz  R0, REPORT_DIAGNOSTIC_60  ; decrement loop counter and repeat
   
   ; Pop the last elements off the 8051 stack and throw them away.
   pop   ACC
   pop   ACC

   ; Increment the number of diagnostics stored this 1-second interval ...
   mov   DPTR, #mCURR_NUM
   movx  A, @DPTR
   inc   A
   movx  @DPTR, A

REPORT_DIAGNOSTIC_99:
   ret                      ; REPORT_DIAGNOSTIC
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DOCRC
; Purpose -- Compute the CRC for the next byte.
; Inputs -- Current CRC in CRCH:CRCL; Byte to process in ACC.
; Outputs -- Updated value to CRC in CRCH:CRCL.
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DOCRC:
   mov   TEMP, A            ; save the received byte
   mov   A, CRCL            ; swap high and low
   mov   DXH, A
   mov   A, CRCH
   xrl   A, TEMP            ; XOR with the data
   mov   DXL, A
   swap  A
   anl   A, #0FH            ; shift right 4
   xrl   A, DXL
   mov   DXL, A
   swap  A
   anl   A, #0F0H           ; shift left 4
   xrl   A, DXH
   mov   DXH, A
   mov   A, DXL
   swap  A
   mov   TEMP1, A
   anl   A, #0F0H
   clr   C
   rlc   A
   xch   A, TEMP1
   anl   A, #0FH
   rlc   A                  ; shift left 5
   xrl   A, DXH
   mov   CRCH, A
   mov   A, TEMP1
   xrl   A, DXL
   mov   CRCL, A
   mov   A, TEMP            ; restore the data byte
   ret                      ; DOCRC
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- BG_CRC (Background CRC computations)
; Purpose -- Computes CRC over a range of memory (XDATA or CODE).
; Input -- 'MEMPTR' points to next memory address to process (big-endian).
;          'CRC_LEN' is number of bytes remaining to compute (0 = 65536)
;          'REGION_FLAG' is memory specifier (0=XDATA, 1=CODE)
; Output -- Computed CRC in 'mLFMCRC'
; Timing -- Test case:  N=1024 bytes
;              w/o interruptions:  52,336 machine cycles (78.5 msec @ 8-MHz)
;                  Worst Case (N=65,536) --> 5.02 seconds @ 8-MHz
;              w/  interruptions:  Add 138 machine cycles per (timer)
;                  interrupt (0.207 msec per interruption @ 8-MHz).
;*****************************************************************************
$SET (STAY_IN_BG_CRC=1)

FILL_TO_BOUNDARY   16, 0FFh
BG_CRC:

   ; Retrieve the background copy of CRC ...
   mov   CRCH, CRCHa
   mov   CRCL, CRCLa
   
   mov   DPH, MEMPTR+0      ; address of next byte to read (big-endian)
   mov   DPL, MEMPTR+1
BG_CRC_10:

$if STAY_IN_BG_CRC
   ; Reasons to leave this background task prior to completion ...
   jb    RXD_FLAG,    BG_CRC_98  ; opcode received
   ; No reasons found to leave this background task, so continue ...
$endif
   
   ; read from XDATA or CODE?
   jnb   REGION_FLAG, BG_CRC_20  ; jump if XDATA space (REGION_FLAG=0)
   clr   A
   movc  A, @A+DPTR
   sjmp  BG_CRC_30
BG_CRC_20:
   movx  A, @DPTR

BG_CRC_30:
   lcall DOCRC              ; update CRC with this data byte
   inc   DPTR               ; advance pointer to next byte
   mov   MEMPTR+0, DPH      ; store address, in case we must leave BG task
   mov   MEMPTR+1, DPL      ; ditto

$if STAY_IN_BG_CRC
   djnz  CRC_LEN+1, BG_CRC_10 ; decrement counter LSB
   djnz  CRC_LEN+0, BG_CRC_10 ; decrement counter MSB
$else
   djnz  CRC_LEN+1, BG_CRC_98 ; decrement counter LSB
   djnz  CRC_LEN+0, BG_CRC_98 ; decrement counter MSB
$endif

   ; we're done, so clean up and exit ...
   mov   CRCHa, CRCH
   mov   CRCLa, CRCL
   lcall STORE_CRC          ; store the answer (little-endian)
                            ; return value:  ACC = 'next region'
   mov   REGION_CRC, A      ; store the 'next region'
   lcall INIT_CRC           ; initialize variables for the new region
   sjmp  BG_CRC_99          ; exit

BG_CRC_98:
   ; Store CRCH:CRCL in the registers CRCHa:CRCLa so that foreground calls to
   ; "DOCRC" will not destroy them.
   mov   CRCHa, CRCH
   mov   CRCLa, CRCL

BG_CRC_99:
   ret                      ; BG_CRC
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- INIT_CRC
; Purpose -- Initialize CRC calculations for a new region of memory.
; Inputs -- next region number in byte-parameter 'REGION_CRC'
; Outputs -- variables 'MEMPTR', 'CRC_LEN', 'CRCHa', and 'CRCLa' initialized
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
INIT_CRC:
   clr   CRC_FLAG           ; disable background CRC calculations

   clr   C                  ; prepare for SUBB operation
   mov   A, REGION_CRC      ; ensure the region requested is valid
   subb  A, #num_regions    ; number of regions allocated, 'num_regions'
   jnc   INIT_CRC_99        ; invalid region specifier -- exit

INIT_CRC_10:
   mov   DPTR, #CRC_REGIONS ; base address of CRC regions
   mov   A, REGION_CRC      ; region number
   mov   B, #sz_regions     ; size of each region specifier, 'sz_regions'
   mul   AB                 ; offset into 'CRC_REGIONS' table in B:A
   add   A, DPL             ; add offset to DPTR
   mov   DPL, A
   mov   A, B
   addc  A, DPH
   mov   DPH, A
   ; Read data from 'CRC_REGIONS' table and put in local variables ...
   movx  A, @DPTR
   mov   MEMPTR+0, A        ; MSB of start address
   inc   DPTR
   movx  A, @DPTR
   mov   MEMPTR+1, A        ; LSB of start address
   inc   DPTR
   movx  A, @DPTR
   mov   CRC_LEN+0, A       ; MSB of length
   inc   DPTR
   movx  A, @DPTR
   mov   CRC_LEN+1, A       ; LSB of length
   inc   DPTR
   movx  A, @DPTR           ; memory specifier (0=XDATA, 1=CODE)
   clr   REGION_FLAG        ; assume XDATA region
   jz    INIT_CRC_20        ; parameter OK
   setb  REGION_FLAG        ; assume CODE region
   dec   A
   jnz   INIT_CRC_99        ; parameter error -- exit

INIT_CRC_20:
   ; initialize CRC for the background task ...
   mov   CRCHa, #0FFh
   mov   CRCLa, #0FFh
   
   ; Fixup "number of bytes" if LSB is non-zero (to make DJNZ commands work)
   mov   A, CRC_LEN+1       ; length LSB
   jz    INIT_CRC_30        ; MSB is OK since LSB=0
   inc   CRC_LEN+0          ; if LSB<>0 we must increment MSB for looping
INIT_CRC_30:

   setb  CRC_FLAG           ; enable background CRC calculations

INIT_CRC_99:
   ret                      ; INIT_CRC
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- STORE_CRC
; Purpose -- Saves the calculated CRC in the designated location for the
;            current CRC region.
; Inputs -- CRC in CRCHa:CRCLa, current region specifier in 'REGION_CRC'
; Outputs -- CRCHa:CRCLa written to designated location in CRC table.
;            Return value is the 'next region' in ACC.
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
STORE_CRC:
   mov   DPTR, #CRC_REGIONS+5 ; 'next region' specifier of first region
   mov   A, REGION_CRC        ; region number
   mov   B, #sz_regions       ; size of each region specifier, 'sz_regions'
   mul   AB                   ; offset into 'CRC_REGIONS' table in B:A
   add   A, DPL               ; add offset to DPTR
   mov   DPL, A
   mov   A, B
   addc  A, DPH
   mov   DPH, A
   ; Read 'next region' from table ...
   movx  A, @DPTR
   push  ACC                ; temporary storage
   inc   DPTR               ; point to LSB of CRC
   mov   A, CRCLa
   movx  @DPTR, A           ; store LSB of CRC
   inc   DPTR               ; point to MSB of CRC
   mov   A, CRCHa
   movx  @DPTR, A           ; store MSB of CRC
   inc   DPTR
   ; Compare new CRC with previous value (if non-zero) and send a diagnostic
   ; if requested ...
   ; Read the previous CRC, push on stack, and store new CRC as 'previous'
   ; (in preparation for the next time we work on this region) ...
   mov   B, #0              ; use B as a temporary flag register
   movx  A, @DPTR           ; retrieve previous CRC LSB
   push  ACC
   mov   A, CRCLa
   movx  @DPTR, A           ; store new CRC LSB
   inc   DPTR
   pop   ACC                ; need to check previous CRC for zero
   push  ACC                ; save it again
   jz    STORE_CRC_10
   mov   B, #1              ; B non-zero means we must compare CRCs
STORE_CRC_10:
   movx  A, @DPTR           ; retrieve previous CRC MSB
   push  ACC
   mov   A, CRCHa
   movx  @DPTR, A           ; store new CRC MSB
   inc   DPTR
   pop   ACC                ; need to check previous CRC for zero
   push  ACC                ; save it again
   jz    STORE_CRC_20
   mov   B, #1              ; B non-zero means we must compare CRCs
STORE_CRC_20:
   mov   A, B               ; move our flag register (B) into ACC to check
   jnz   STORE_CRC_30       ; if B<>0, we must compare CRCs
   pop   ACC                ; remove the 'previous CRC' from stack
   pop   ACC
   sjmp  STORE_CRC_98       ; exit

   ; we must compare the CRCs
STORE_CRC_30:
   pop   ACC                ; retrieve previous CRC MSB
   cjne  A, CRCHa, STORE_CRC_40  ; jump on miscompare
   pop   ACC                ; retrieve previous CRC LSB
   cjne  A, CRCLa, STORE_CRC_50  ; jump on miscompare
   sjmp  STORE_CRC_98       ; CRCs are equal -- exit

   ; miscompare on first byte -- pop other byte off stack
STORE_CRC_40:
   pop   ACC                ; remove second part of CRC from stack
   
STORE_CRC_50:
   ; CRCs are different.  If the MSB of the diagnostic parameter in this
   ; CRC table entry is 5Ah, then we send the LSB as the diagnostic messge.
   movx  A, @DPTR           ; retrieve MSB of diagnostic message
   cjne  A, #5Ah, STORE_CRC_98  ; no diagnostic to send -- exit
   ; must send the diagnostic ...
   ; First, store diagnostic number on stack and write the bad CRC region
   ; number to the parameter 'mBADREGION' -- the HST error parameter for this
   ; situation ...
   inc   DPTR
   movx  A, @DPTR           ; diagnostic number is LSB
   push  ACC                ; store temporarily
   ; write to 'mBADREGION' ...
   mov   DPTR, #mBADREGION
   mov   A, REGION_CRC
   movx  @DPTR, A
   inc   DPTR
   mov   A, #0
   movx  @DPTR, A
   pop   ACC                ; recover diagnostic LSB
   lcall report_diagnostic  ; Record this diagnostic event

STORE_CRC_98:
   pop   ACC                ; recover the 'next region'

STORE_CRC_99:
   ret                      ; STORE_CRC
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- CLEAR_OLD_CRCS
; Purpose -- Zeros out all 'previous CRC' entries in the CRC_REGIONS table,
;            thereby preventing CRC comparison errors on the first pass.
; Inputs -- none
; Outputs -- All 'previous CRC' entries in the CRC_REGIONS table are set to
;            zero.
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CLEAR_OLD_CRCS:
   mov   A, #num_regions      ; number of regions allocated, 'num_regions'
   jz    CLEAR_OLD_CRCS_99    ; no regions to clear
   
   mov   R0, A                ; initialize loop counter
   mov   DPTR, #CRC_REGIONS+8 ; 'previous CRC' of first region
CLEAR_OLD_CRCS_10:
   clr   A
   movx  @DPTR, A             ; clear LSB of 'previous CRC'
   inc   DPTR
   movx  @DPTR, A             ; clear MSB of 'previous CRC'
   ; advance pointer to next region (next element in the table)
   mov   A, #sz_regions-1     ; 'sz_regions-1' because DPTR already incr.
   add   A, DPL
   mov   DPL, A
   mov   A, DPH
   addc  A, #0                ; propagate the carry
   mov   DPH, A
   djnz  R0, CLEAR_OLD_CRCS_10  ; loop until done

CLEAR_OLD_CRCS_99:
   ret                      ; CLEAR_OLD_CRCS
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- TDC_A  &  TDC_B
; Purpose -- Write 16 bits to either TDC 'A' or 'B'.
; Inputs -- Data to send in R0:R1.  Routine sends MSb first.
; Outputs -- 16 bits sent out appropriate 8051 port pin.
; Timing -- 205 machine cycles (0.308 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
TDC_A:
   clr   TDC_CLOCK          ; clock low (data latched on low-to-high)
   clr   TDC_B_ENABLE       ; ensure TDC 'B' is disabled
   setb  TDC_A_ENABLE       ; enable TDC 'A'
   sjmp  TDC_WRITE
TDC_B:
   clr   TDC_CLOCK          ; clock low (data latched on low-to-high)
   clr   TDC_A_ENABLE       ; ensure TDC 'A' is disabled
   setb  TDC_B_ENABLE       ; enable TDC 'B'
TDC_WRITE:
   mov   A, R0              ; MSB into ACC for shifting out
   mov   R0, #16            ; 16 bits to shift out
   ; we will rotate the 16-bit value held in ACC:R1 left 16 times
TDC_WRITE_10:
   xch   A, R1              ; put LSB in ACC (and hold MSB in R1)
   rlc   A                  ; put MS-bit of LSB into carry
   xch   A, R1              ; swap bytes back again
   rlc   A                  ; put carry into LS-bit of MSB and put next output
                            ; bit (MSB's MS-bit) into carry.
   mov   TDC_DATA, C        ; write Carry (the data bit) out port pin
   setb  TDC_CLOCK          ; Latch bit into ACTEL device (the TDC)
   nop                      ; remain high for 1.5 us
   nop
   clr   TDC_CLOCK          ; clock low in preparation for next cycle
   djnz  R0, TDC_WRITE_10   ; loop through 16 bits

TDC_WRITE_99:
   clr   TDC_A_ENABLE       ; disable TDC 'A'
   clr   TDC_B_ENABLE       ; disable TDC 'B'
   ret                      ; TDC_A  &  TDC_B
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- MEMORY_MONITORS
; Purpose -- Read the 8 memory monitor locations and store the 8 byte values.
;            (These will later be copied into the housekeeping packet.)
; Inputs -- 8 memory monitor address in parameter 'mLFDMADD' (little-endian).
;           Bitfield parameter 'mLFDMDATA' determines whether a particular
;           memory read is DATA/XDATA ('0') or CODE ('1').
;           Bitfield parameter 'mLFDMEXTR' further differentiates the data
;           source as EXTERNAL ('0') or INTERNAL ('1').
;           Bit 0 of the bitfield parameters corresponds to the first address
;           found in 'mLFDMADD'.
; Outputs -- 8 byte values are written to 'mLFDMONS' (which will later be
;            written to the housekeeping packet).
; Timing -- all 8 in CODE space --  496 machine cycles (0.744 msec @ 8-MHz)
;           all 8 in DATA space --  536 machine cycles (0.804 msec @ 8-MHz)
;           all 8 in XDATA space -- 560 machine cycles (0.840 msec @ 8-MHz)
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
MEMORY_MONITORS:

   mov   R7, #8             ; loop counter -- number of addresses to monitor
   mov   DPTR, #mLFDMADD    ; starting location of table of addresses
   mov   R6, DPH            ; save current memory address in R6:R5
   mov   R5, DPL
   mov   DPTR, #mLFDMONS    ; starting location of data values to store
   mov   R4, DPH            ; save current memory address in R4:R3
   mov   R3, DPL
   mov   R2, #1             ; mask bit (begins with bit 0)

MEMORY_MONITORS_00:
   mov   DPH, R6            ; pointer to address
   mov   DPL, R5
   movx  A, @DPTR
   mov   tempDPTR+1, A      ; LSB of address to read
   inc   DPTR
   movx  A, @DPTR
   mov   tempDPTR+0, A      ; MSB of address to read
   mov   DPTR, #mLFDMDATA   ; bit fields:  1=CODE, 0=RAM (DATA or XDATA)
   movx  A, @DPTR
   anl   A, R2              ; mask off the word
   jnz   MEMORY_MONITORS_50 ; a CODE space address
   
MEMORY_MONITORS_10:
   ; Memory monitor points to RAM space (either DATA or XDATA) ...
   mov   DPTR, #mLFDMEXTR   ; bit fields:  0=XDATA, 1=DATA
   movx  A, @DPTR
   anl   A, R2              ; mask off the word
   jz    MEMORY_MONITORS_30 ; an XDATA address
   
MEMORY_MONITORS_20:
   ; Memory monitor points to DATA space (i.e. internal) ...
   mov   R1, tempDPTR+1     ; LSB is whole address of internal memory
   mov   A, @R1             ; read internal memory
   mov   DPH, R4            ; recall memory write address
   mov   DPL, R3
   movx  @DPTR, A           ; store the value
   sjmp  MEMORY_MONITORS_70 ; increment pointers, etc.
   
MEMORY_MONITORS_30:
   ; Memory monitor points to XDATA space ...
   mov   DPH, tempDPTR+0    ; recall memory read address
   mov   DPL, tempDPTR+1
   movx  A, @DPTR           ; read memory
   mov   DPH, R4            ; recall memory write address
   mov   DPL, R3
   movx  @DPTR, A           ; store the value
   sjmp  MEMORY_MONITORS_70 ; increment pointers, etc.

MEMORY_MONITORS_50:
   ; Memory monitor points to CODE space ...
   mov   DPH, tempDPTR+0    ; recall memory read address
   mov   DPL, tempDPTR+1
   clr   A
   movc  A, @A+DPTR         ; read memory
   mov   DPH, R4            ; recall memory write address
   mov   DPL, R3
   movx  @DPTR, A           ; store the value

MEMORY_MONITORS_70:
   mov   DPH, R4            ; advance storage pointer by 1 byte
   mov   DPL, R3
   inc   DPTR
   mov   R4, DPH
   mov   R3, DPL
   mov   DPH, R6            ; advance memory address pointer by 2 bytes
   mov   DPL, R5
   inc   DPTR
   inc   DPTR
   mov   R6, DPH
   mov   R5, DPL
   mov   A, R2              ; rotate mask byte by 1
   rl    A
   mov   R2, A              ; store new mask byte
   djnz  R7, MEMORY_MONITORS_00

MEMORY_MONITORS_99:
   ret                      ; MEMORY_MONITORS
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- XDATA_TLU_WORD
; Purpose -- Performs a table look-up for word-sized (16-bit) values from
;            XDATA
; Inputs -- DPTR points to base of structure, ACC is index into the table.
;           Accumulator must be in the range:  0 <= ACC <= 7Fh.  Table values
;           are assumed to be BIG-ENDIAN.  That is, the first value read is
;           the MSB and the next value (after incrementing DPTR) is the LSB.
; Outputs -- B:A holds the 2-byte value (B is MSB).
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
XDATA_TLU_WORD:
   push  DPH                ; store DPTR
   push  DPL
   anl   A, #7Fh            ; mask of msb just in case it was '1'
   rl    A                  ; multiply by 2 (this is a WORD offset)
   add   A, DPL
   mov   DPL, A
   mov   A, DPH
   addc  A, #0
   mov   DPH, A
   movx  A, @DPTR
   mov   B, A               ; B = MSB of table look-up value
   inc   DPTR
   movx  A, @DPTR           ; A = LSB
   pop   DPL                ; restore DPTR
   pop   DPH
XDATA_TLU_WORD_99:
   ret                      ; XDATA_TLU_WORD
;-----------------------------------------------------------------------------   

;*****************************************************************************
; Subroutine -- CODE_TLU_WORD
; Purpose -- Performs a table look-up for word-sized (16-bit) values from
;            CODE
; Inputs -- DPTR points to base of structure, ACC is index into the table.
;           Accumulator must be in the range:  0 <= ACC <= 7Fh.  Table values
;           are assumed to be BIG-ENDIAN.  That is, the first value read is
;           the MSB and the next value (after incrementing DPTR) is the LSB.
; Outputs -- B:A holds the 2-byte value (B is MSB).
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CODE_TLU_WORD:
   push  DPH                ; store DPTR
   push  DPL
   anl   A, #7Fh            ; mask of msb just in case it was '1'
   rl    A                  ; multiply by 2 (this is a WORD offset)
   add   A, DPL
   mov   DPL, A
   mov   A, DPH
   addc  A, #0
   mov   DPH, A
   clr   A                  ; prepare for movc command
   movc  A, @A+DPTR
   mov   B, A               ; B = MSB of table look-up value
   inc   DPTR
   clr   A                  ; prepare for movc command
   movc  A, @A+DPTR         ; A = LSB
   pop   DPL                ; restore DPTR
   pop   DPH
CODE_TLU_WORD_99:
   ret                      ; CODE_TLU_WORD
;-----------------------------------------------------------------------------   

;*****************************************************************************
; 'COS_OPER_TBL' is a table of function addresses.  When a command opcode is
; received, the address of the corresponding routine is looked up in this
; table and then a jump is performed.
; NOTE:  Routine 'CMD_ILLEGAL' handles commands that are illegal for Operate
;    mode.
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
COS_OPER_TBL:
   DW   DCE_LFDNOOP   ; 80H  No Operation
   DW   DCE_LFDMADDR  ; 81H  Sets MEB Monitor Address
   DW   DCE_LFDCRC    ; 82H  Calculate CRC on DCE memory range
   DW   DCE_LFDCOPY   ; 83H  Memory-to-memory copy
   DW   CMD_ILLEGAL   ; 84H
   DW   DCE_LFGBWK    ; 85H  TDC Begin Walk Adjust
   DW   CMD_ILLEGAL   ; 86H
   DW   CMD_ILLEGAL   ; 87H
   DW   CMD_ILLEGAL   ; 88H
   DW   DCE_LFGEWK    ; 89H  TDC End Walk Adjust
   DW   CMD_ILLEGAL   ; 8AH
   DW   CMD_ILLEGAL   ; 8BH
   DW   CMD_ILLEGAL   ; 8CH
   DW   DCE_LFHVMAX   ; 8DH  Maximum Level for HV
   DW   CMD_ILLEGAL   ; 8EH
   DW   CMD_ILLEGAL   ; 8FH
   DW   CMD_ILLEGAL   ; 90H
   DW   CMD_ILLEGAL   ; 91H
   DW   CMD_ILLEGAL   ; 92H
   DW   CMD_ILLEGAL   ; 93H
   DW   CMD_ILLEGAL   ; 94H
   DW   DCE_LFGLQT    ; 95H  TDC Lower Charge Threshold
   DW   CMD_ILLEGAL   ; 96H
   DW   DCE_LFGTT     ; 97H  TDC Timing Threshold
   DW   CMD_ILLEGAL   ; 98H
   DW   CMD_ILLEGAL   ; 99H
   DW   CMD_ILLEGAL   ; 9AH
   DW   CMD_ILLEGAL   ; 9BH
   DW   DCE_LFRLSOVD  ; 9Ch  Door Limit Switch Override
   DW   CMD_ILLEGAL   ; 9DH
   DW   DCE_LFGSTIM   ; 9Eh  Stim Control, On/Off, Rate
   DW   CMD_ILLEGAL   ; 9FH
   DW   CMD_ILLEGAL   ; A0H
   DW   CMD_ILLEGAL   ; A1H
   DW   DCE_LFGUQT    ; A2H  TDC Upper Charge Threshold
   DW   CMD_ILLEGAL   ; A3H
   DW   DCE_LFGSHFT   ; A4H  Segment A Image Shift, X Axis
   DW   CMD_ILLEGAL   ; A5H
   DW   DCE_LFGSTR    ; A6H  Segment Image Stretch
   DW   CMD_ILLEGAL   ; A7H
   DW   CMD_ILLEGAL   ; A8H
   DW   CMD_ILLEGAL   ; A9H
   DW   CMD_ILLEGAL   ; AAH
   DW   CMD_ILLEGAL   ; ABH
   DW   CMD_ILLEGAL   ; ACH
   DW   DCE_LFDUPLOD  ; ADH  Upload Memory Block
   DW   DCE_LFDDNLOD  ; AEH  Download Memory Block
   DW   CMD_ILLEGAL   ; AFH
   DW   CMD_ILLEGAL   ; B0H
   DW   CMD_ILLEGAL   ; B1H
   DW   CMD_ILLEGAL   ; B2H
   DW   CMD_ILLEGAL   ; B3H
   DW   CMD_ILLEGAL   ; B4H
   DW   CMD_ILLEGAL   ; B5H
   DW   DCE_LFHVLOW   ; B6H  HV Level for HVLOW State
   DW   CMD_ILLEGAL   ; B7H
   DW   CMD_ILLEGAL   ; B8H
   DW   CMD_ILLEGAL   ; B9H
   DW   DCE_LFPCRP    ; BAH  CRP Algorithm Parameters
   DW   CMD_ILLEGAL   ; BBH
   DW   CMD_ILLEGAL   ; BCH
   DW   CMD_ILLEGAL   ; BDH
   DW   CMD_ILLEGAL   ; BEH
   DW   CMD_ILLEGAL   ; BFH
   DW   DCE_LFHVPWR   ; C0H  HV Power Supply On/Off
   DW   DCE_LFRACT1   ; C1H  Actuator 1 On/Off
   DW   DCE_LFRACT2   ; C2H  Actuator 2 On/Off
   DW   DCE_LFRACTEN  ; C3H  Actuator Control Enable/Disable
   DW   DCE_LFRAXPWR  ; C4H  Aux Power Supply On/Off
   DW   DCE_LFRACTRS  ; C5H  Actuator Reset
   DW   DCE_LFHVNOM   ; C6H  HV Level for HVNOM State
   DW   CMD_ILLEGAL   ; C7H
   DW   CMD_ILLEGAL   ; C8H
   DW   CMD_ILLEGAL   ; C9H
   DW   CMD_ILLEGAL   ; CAH
   DW   DCE_LFHVSET   ; CBH  MCP HV Level Setting
   DW   CMD_ILLEGAL   ; CCH
   DW   CMD_ILLEGAL   ; CDH
   DW   DCE_LFHRAMPT  ; CEH  HV Ramp Time Constant
   DW   DCE_LFHQPWR   ; CFH  QDE Grid HV On/Off
   DW   DCE_LFHVILIM  ; D0H  HV Current Limit Setting
   DW   DCE_LFHVENA   ; D1H  HV Power Supply Enable/Disable
   DW   CMD_ILLEGAL   ; D2H
   DW   CMD_ILLEGAL   ; D3H
   DW   CMD_ILLEGAL   ; D4H
   DW   CMD_ILLEGAL   ; D5H
   DW   CMD_ILLEGAL   ; D6H
   DW   CMD_ILLEGAL   ; D7H
   DW   CMD_ILLEGAL   ; D8H
   DW   DCE_LFRMDIR   ; D9H  Select the Door Direction
   DW   DCE_LFRMPWR   ; DAH  Door Motor Power On/Off
   DW   DCE_LFRMENA   ; DBH  Door Motor Enable/Disable
   DW   DCE_LFRILIM   ; DCH  Door Motor/Actuator/HVPS Current Limit Setting
   DW   DCE_LFSAFE    ; DDH  Door_Stop and Safe_Controls
   DW   CMD_ILLEGAL   ; DEH
   DW   CMD_ILLEGAL   ; DFH
   DW   CMD_ILLEGAL   ; E0H
   DW   CMD_ILLEGAL   ; E1H
   DW   CMD_ILLEGAL   ; E2H
   DW   CMD_ILLEGAL   ; E3H
   DW   CMD_ILLEGAL   ; E4H
   DW   CMD_ILLEGAL   ; E5H
   DW   DCE_LFHSTATE  ; E6H  Set HV State
   DW   CMD_ILLEGAL   ; E7H
   DW   CMD_ILLEGAL   ; E8H
   DW   CMD_ILLEGAL   ; E9H
   DW   DCE_LFDGOTO   ; EAH  Jump to a Specified Address and Execute
   DW   CMD_ILLEGAL   ; EBH
   DW   CMD_ILLEGAL   ; ECH
   DW   CMD_ILLEGAL   ; EDH
   DW   CMD_ILLEGAL   ; EEH
   DW   CMD_ILLEGAL   ; EFH
   DW   DCE_LFDRSTP   ; F0H  Power-On Reset (POR)
   DW   DCE_LFDWDOG   ; F1H  Watchdog Enable/Disable
   DW   CMD_ILLEGAL   ; F2H
   DW   CMD_ILLEGAL   ; F3H  Boot Only -- Jump to Operate
   DW   DCE_LFDDIAGC  ; F4H  Clear Diagnostic Code Stack
   DW   DCE_LFDRSTW   ; F5H  Watchdog Reset (WDR)
   DW   CMD_ILLEGAL   ; F6H
   DW   CMD_ILLEGAL   ; F7H
   DW   CMD_ILLEGAL   ; F8H
   DW   CMD_ILLEGAL   ; F9H
   DW   DCE_LFDRSTA   ; FAH  Reset the ACTEL counter
   DW   CMD_ILLEGAL   ; FBH
   DW   CMD_ILLEGAL   ; FCH
   DW   CMD_ILLEGAL   ; FDH
   DW   CMD_ILLEGAL   ; FEH
   DW   DCE_LFDHKREQ  ; FFH  Send Full Housekeeping Packet
;-----------------------------------------------------------------------------

FILL_TO_ADDRESS_CRC    0AE00h, 0FFh, 0BC50h
reg9len   equ   $-reg9




$if INCLUDE_TEST_CODE
;*****************************************************************************
;*****************************************************************************
;*****************************************************************************
; Test code is placed in its own segment well beyond any boot code ...
   CSEG  at  TESTCODE_SEG
;*****************************************************************************
;*****************************************************************************
;*****************************************************************************

;-----------------------------------------------------------------------------
; Write byte 'value' to XDATA memory location 'dest' ...
WRITE_BYTE MACRO dest, value
$SAVE
$NOLIST
   mov   DPTR, #dest
   mov   A, #LOW(value)
   movx  @DPTR, A
$RESTORE
ENDM
;-----------------------------------------------------------------------------
; Write word 'value' to XDATA memory location 'dest' with MSB at lower
; address (big-endian) ...
WRITE_BIG_END MACRO dest, value
$SAVE
$NOLIST
   mov   DPTR, #dest
   mov   A, #HIGH(value)
   movx  @DPTR, A
   inc   DPTR
   mov   A, #LOW(value)
   movx  @DPTR, A
$RESTORE
ENDM
;-----------------------------------------------------------------------------
; Write word 'value' to XDATA memory location 'dest' with LSB at lower
; address (little-endian) ...
WRITE_LITTLE_END MACRO dest, value
$SAVE
$NOLIST
   mov   DPTR, #dest
   mov   A, #LOW(value)
   movx  @DPTR, A
   inc   DPTR
   mov   A, #HIGH(value)
   movx  @DPTR, A
$RESTORE
ENDM
;-----------------------------------------------------------------------------
; This macro puts command parameters directly into the command receipt buffer,
; thereby avoiding the overhead of the ISR call.  It also sets the 'RXD_FLAG',
; to tell the main routine that a command has been received.
TEST_MAKE_COMMAND MACRO  cmd, seq, p1, p2, p3, p4, p5
$SAVE
$NOLIST
cmd1   equ   ((LOW(cmd) SHL 8) OR LOW(cmd))
addr   equ   RS17

   mov   DPTR, #addr
   IRP  PARAM, <cmd1,seq,p1,p2,p3,p4,p5>
      mov   A, #LOW(PARAM)
      movx  @DPTR, A
      inc   DPTR
      mov   A, #HIGH(PARAM)
      movx  @DPTR, A
      inc   DPTR
      mov   A, #LOW(PARAM) XOR 0FFh
      movx  @DPTR, A
      inc   DPTR
      mov   A, #HIGH(PARAM) XOR 0FFh
      movx  @DPTR, A
      inc   DPTR
   ENDM      
   setb  RXD_FLAG           ; signal main routine that command received

$RESTORE
ENDM
;-----------------------------------------------------------------------------
; This macro puts command parameters in the receive registers and calls
; the command receive ISR to process them in normal fashion ...
TEST_MAKE_COMMAND1 MACRO  cmd, seq, p1, p2, p3, p4, p5
$SAVE
$NOLIST
cmd_buffer   equ   0440h
addr         equ   cmd_buffer+4

   IRP  PARAM, <seq,p1,p2,p3,p4,p5>
      mov   DPTR, #RXD_HARDWARE+300h   ; put command MSB here
      mov   A, #HIGH(addr)
      movx  @DPTR, A
      dec   DPH
      mov   A, #LOW(addr)
      movx  @DPTR, A
      dec   DPH
      mov   A, #HIGH(PARAM)
      movx  @DPTR, A
      dec   DPH
      mov   A, #LOW(PARAM)
      movx  @DPTR, A
      lcall ISR_Pri_Cmd_Received
addr   equ   addr+2
      mov   DPTR, #RXD_HARDWARE+300h   ; put command MSB here
      mov   A, #HIGH(addr)
      movx  @DPTR, A
      dec   DPH
      mov   A, #LOW(addr)
      movx  @DPTR, A
      dec   DPH
      mov   A, #HIGH(PARAM) XOR 0FFh
      movx  @DPTR, A
      dec   DPH
      mov   A, #LOW(PARAM) XOR 0FFh
      movx  @DPTR, A
      lcall ISR_Pri_Cmd_Received
addr   equ   addr+2
   ENDM

   ; Finally insert the command commplement and then the command ...
addr   equ   cmd_buffer+2
   mov   DPTR, #RXD_HARDWARE+300h   ; put command MSB here
   mov   A, #HIGH(addr)
   movx  @DPTR, A
   dec   DPH
   mov   A, #LOW(addr)
   movx  @DPTR, A
   dec   DPH
   mov   A, #LOW(cmd) XOR 0FFh
   movx  @DPTR, A
   dec   DPH
   mov   A, #LOW(cmd) XOR 0FFh
   movx  @DPTR, A
   lcall ISR_Pri_Cmd_Received
addr   equ   cmd_buffer
   mov   DPTR, #RXD_HARDWARE+300h   ; put command MSB here
   mov   A, #HIGH(addr)
   movx  @DPTR, A
   dec   DPH
   mov   A, #LOW(addr)
   movx  @DPTR, A
   dec   DPH
   mov   A, #LOW(cmd)
   movx  @DPTR, A
   dec   DPH
   mov   A, #LOW(cmd)
   movx  @DPTR, A
   lcall ISR_Pri_Cmd_Received

$RESTORE
ENDM
;-----------------------------------------------------------------------------
; This macro creates a command sequence in CODE space for later copying
; to the command receipt buffer.
TEST_MAKE_COMMAND2 MACRO  cmd, seq, p1, p2, p3, p4, p5
cmd1     equ   ((LOW(cmd) SHL 8) OR LOW(cmd))
   IRP  PARAM, <cmd1,seq_num,p1,p2,p3,p4,p5>
      DB   LOW(PARAM), HIGH(PARAM), LOW(NOT PARAM), HIGH(NOT PARAM)
   ENDM      
seq_num   equ   seq_num+1
ENDM
;-----------------------------------------------------------------------------
; This macro creates a command in which the opcode is not in both upper and
; lower halves of the command word
TEST_MAKE_BAD_CMD MACRO
   DB   0, 80h, 0FFh, 07Fh
   DB   LOW(seq_num), HIGH(seq_num), LOW(NOT seq_num), HIGH(NOT seq_num)
   DB   0,   0, 0FFh, 0FFh
   DB   0,   0, 0FFh, 0FFh
   DB   0,   0, 0FFh, 0FFh
   DB   0,   0, 0FFh, 0FFh
   DB   0,   0, 0FFh, 0FFh
seq_num   equ   seq_num+1
ENDM
;-----------------------------------------------------------------------------
; This macro creates a command in which a parameter does not have a proper
; complement
TEST_MAKE_BAD_PARAM MACRO
   DB   80h, 80h, 07Fh, 07Fh
   DB   LOW(seq_num), HIGH(seq_num), LOW(NOT seq_num), HIGH(NOT seq_num)
   DB     0,   0, 0FFh, 0FFh
   DB     0,   0, 0FFh, 0FFh
   DB     0,   0,    0, 0FFh
   DB     0,   0, 0FFh, 0FFh
   DB     0,   0, 0FFh, 0FFh
seq_num   equ   seq_num+1
ENDM
;-----------------------------------------------------------------------------

;*****************************************************************************
; 'TEST_MEMORY_MONITORS' inserts data into the memory monitor structure
; and verifies that 
;*****************************************************************************
; Test MEMORY_MONITORS
FILL_TO_BOUNDARY   16, 0FFh
TEST_MEMORY_MONITORS:
   WRITE_BYTE  mLFDMDATA, 10101010b
   WRITE_BYTE  mLFDMEXTR, 00001111b
   WRITE_LITTLE_END  mLFDMADD+0, 6000h+128+0
   WRITE_LITTLE_END  mLFDMADD+2, 6000h+128+1
   WRITE_LITTLE_END  mLFDMADD+4, 6000h+128+2
   WRITE_LITTLE_END  mLFDMADD+6, 6000h+128+3
   WRITE_LITTLE_END  mLFDMADD+8, 6000h+128+4
   WRITE_LITTLE_END  mLFDMADD+10, 6000h+128+5
   WRITE_LITTLE_END  mLFDMADD+12, 6000h+128+6
   WRITE_LITTLE_END  mLFDMADD+14, 6000h+128+7
   WRITE_BYTE  6000h+128+0, 5
   WRITE_BYTE  6000h+128+1, 10
   WRITE_BYTE  6000h+128+2, 15
   WRITE_BYTE  6000h+128+3, 20
   WRITE_BYTE  6000h+128+4, 25
   WRITE_BYTE  6000h+128+5, 30
   WRITE_BYTE  6000h+128+6, 35
   WRITE_BYTE  6000h+128+7, 40
   mov   R0, #128+0
   mov   @R0, #5
   mov   R0, #128+2
   mov   @R0, #15
;   lcall MEMORY_MONITORS
   ret
;-----------------------------------------------------------------------------

;*****************************************************************************
; Test command jump table ...
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
TEST_COMMAND_VECTOR:
   mov   HKPBUFF, #0        ; Temporary use
OPERATE_MAIN_00:
   mov   A, HKPBUFF
   lcall COMMAND_VECTOR
   inc   HKPBUFF
   mov   A, HKPBUFF
   jnz   OPERATE_MAIN_00
   ret
;-----------------------------------------------------------------------------

;*****************************************************************************
; Create command packet 80000000h to test 'jump to power-on reset' ...
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
TEST_CMD80000000:
   mov   A, #80h
   mov   DPTR, #RXD_HARDWARE + 300h   ; Primary command input channel
   movx  @DPTR, A
   mov   A, #0
   dec   DPH
   movx  @DPTR, A
   dec   DPH
   movx  @DPTR, A
   dec   DPH
   movx  @DPTR, A
   lcall ISR_Pri_Cmd_Received
   ret
;-----------------------------------------------------------------------------

;*****************************************************************************
; 'test_timing' calls all the specified commands so timing can be computed.
; One should reset the timer (in debug mode) prior to each call.
;*****************************************************************************
; Call this subroutine on the occurrence of a timer TOCK (1-second timeout).
FILL_TO_BOUNDARY   16, 0FFh
test_timing:
   lcall ISR_Pri_Cmd_Received  ; Boot mode ISR 03
   lcall ISR_Timer0            ; Boot mode ISR 0B
   lcall ISR_Sec_Cmd_Received  ; Boot mode ISR 13
   lcall ISR_Timer1            ; Boot mode ISR 1B

   lcall CLEAR_HKP_PKT

   ret                      ; test_timing
;-----------------------------------------------------------------------------

;*****************************************************************************
; 'test_commands' uses the 1-second timer to select the next function from
; a pre-defined list of functions.  This function is called on the 1-second
; 'tock' in the main executive loop.
;*****************************************************************************
; Call this subroutine on the occurrence of a timer TOCK (1-second timeout).
FILL_TO_BOUNDARY   16, 0FFh
test_commands:

   ; Prepare registers for MEM_COPY routine.
   ; Put source address in R3:R2 ...
   mov   DPTR, #mLFCTIME
   movx  A, @DPTR           ; 1-second timer's LSB determines next command
   mov   B, #(cmd_sequence_1 - cmd_sequence)  ; spacing between commands
   mul   AB
   add   A, #LOW(cmd_sequence)
   mov   R2, A
   mov   A, B
   addc  A, #HIGH(cmd_sequence)
   mov   R3, A
   ; Put destination address in R5:R4 (command-received buffer) ...
   mov   R5, #HIGH(RS17)
   mov   R4, #LOW(RS17)
   ; Number of bytes to copy in R0:R7 (28 bytes for command packets) ...
   mov   R0, #HIGH(28)
   mov   R7, #LOW(28)
   ; Source of move is CODE space ...
   setb  SRC_FLAG           ; 1 = copy from CODE space
   lcall MEM_COPY           ; copy the command into the buffer
   setb  RXD_FLAG           ; flag a command received

test_commands_99:
   ret
;-----------------------------------------------------------------------------


; Create various command packets for testing.  These are copied to the
; command-received buffer once per second and RXD_FLAG is set true so
; the main executive loop will execute the command.
$SET (cmd_seq_num=1)

seq_num   equ   0   ; Macros increment this giving a rolling seq. number

cmd_sequence:
;00 -- not executed because first timer rollover is to '1'
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
cmd_sequence_1:   ; label used to find spacing
$if (cmd_seq_num=1)
;01

   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP

$elseif (cmd_seq_num=2)
$elseif (cmd_seq_num=3)
$elseif (cmd_seq_num=4)
$elseif (cmd_seq_num=5)
$elseif (cmd_seq_num=6)
$elseif (cmd_seq_num=7)
$elseif (cmd_seq_num=8)
$elseif (cmd_seq_num=9)
$else

   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    081h, 0,     0, 2460h,      0, 0, 0   ; LFDMADDR
   TEST_MAKE_COMMAND2    082h, 0, 6000h, 0400h,      0, 0, 0   ; LFDCRC
   TEST_MAKE_COMMAND2    083h, 0,     0, 4000h,  0400h, 0, 0   ; LFDCOPY
   TEST_MAKE_COMMAND2    0ADh, 0, 6000h, 0400h,      0, 0, 0   ; LFDUPLOD
   TEST_MAKE_COMMAND2    0AEh, 0, 6000h, 0400h,      0, 0, 0   ; LFDDNLOD
   TEST_MAKE_COMMAND2    0EAh, 0,     0,     0,      0, 0, 0   ; LFDGOTO
   TEST_MAKE_COMMAND2    0F0h, 0,     0,     0,      0, 0, 0   ; LFDRSTP
   TEST_MAKE_COMMAND2    0F1h, 0,     1,     0,      0, 0, 0   ; LFDWDOG ON
   TEST_MAKE_COMMAND2    0F1h, 0,     0,     0,      0, 0, 0   ; LFDWDOG OFF
   TEST_MAKE_COMMAND2    0F3h, 0,     0,     0,      0, 0, 0   ; LFDJMPCS
   TEST_MAKE_COMMAND2    0F4h, 0,     0,     0,      0, 0, 0   ; LFDDIAGC
   TEST_MAKE_COMMAND2    0F5h, 0,     0,     0,      0, 0, 0   ; LFDRSTW
   TEST_MAKE_COMMAND2    0FFh, 0,     0,     0,      0, 0, 0   ; LFDHKREQ

   TEST_MAKE_COMMAND2    085h, 0,  0ABh,     0,      1, 0, 0   ; LFGBWK
   TEST_MAKE_COMMAND2    089h, 0,  056h,     1,      1, 0, 0   ; LFGEWK
   TEST_MAKE_COMMAND2    08Dh, 0,  0C0h,     1,      0, 0, 0   ; LFHVMAX
   TEST_MAKE_COMMAND2    095h, 0,  030h,     1,      0, 0, 0   ; LFGLQT
   TEST_MAKE_COMMAND2    097h, 0,  055h,     0,      1, 0, 0   ; LFGTT
   TEST_MAKE_COMMAND2    09Ch, 0,     0,     0,      0, 0, 0   ; LFRLSOVD
   TEST_MAKE_COMMAND2    09Ch, 0,     1,     0,      0, 0, 0   ; LFRLSOVD
   TEST_MAKE_COMMAND2    09Eh, 0,     1,     0,      0, 0, 0   ; LFGSTIM
   TEST_MAKE_COMMAND2    0A2h, 0,  0B0h,     1,      0, 0, 0   ; LFGUQT
   TEST_MAKE_COMMAND2    0A4h, 0,  098h,     1,      1, 0, 0   ; LFGSHFT
   TEST_MAKE_COMMAND2    0A6h, 0,  0DCh,     0,      1, 0, 0   ; LFGSTR
   TEST_MAKE_COMMAND2    0B6h, 0,  055h,     1,      0, 0, 0   ; LFHVLOW
   TEST_MAKE_COMMAND2    0BAh, 0,     7,     0,  6789h, 0, 0   ; LFPCRP
   TEST_MAKE_COMMAND2    0C0h, 0,     0,     0,      0, 0, 0   ; LFHVPWR
   TEST_MAKE_COMMAND2    0C1h, 0,     1,     0,      0, 0, 0   ; LFRACT1
   TEST_MAKE_COMMAND2    0C2h, 0,     1,     0,      0, 0, 0   ; LFRACT2
   TEST_MAKE_COMMAND2    0C3h, 0,     1,     0,      0, 0, 0   ; LFRACTEN
   TEST_MAKE_COMMAND2    0C4h, 0,     1,     0,      0, 0, 0   ; LFRAXPWR
   TEST_MAKE_COMMAND2    0C5h, 0,     1,     0,      0, 0, 0   ; LFRACTRS
   TEST_MAKE_COMMAND2    0C6h, 0,  095h,     1,      0, 0, 0   ; LFHVNOM
   TEST_MAKE_COMMAND2    0CBh, 0,  0A3h,     1,      0, 0, 0   ; LFHVSET
   TEST_MAKE_COMMAND2    0CEh, 0,   500,     0,      0, 0, 0   ; LFHRAMPT
   TEST_MAKE_COMMAND2    0CFh, 0,     1,     0,      0, 0, 0   ; LFHQPWR
   TEST_MAKE_COMMAND2    0D0h, 0,  0A0h,     0,      0, 0, 0   ; LFHVILIM
   TEST_MAKE_COMMAND2    0D1h, 0,     1,     0,      0, 0, 0   ; LFHVENA
   TEST_MAKE_COMMAND2    0D9h, 0,     2,     0,      0, 0, 0   ; LFRMDIR
   TEST_MAKE_COMMAND2    0DAh, 0,     1,     0,      0, 0, 0   ; LFRMPWR
   TEST_MAKE_COMMAND2    0DBh, 0,     1,     0,      0, 0, 0   ; LFRMENA
   TEST_MAKE_COMMAND2    0DCh, 0,  067h,     0,      0, 0, 0   ; LFRILIM
   TEST_MAKE_COMMAND2    0DDh, 0,     0,     0,      0, 0, 0   ; LFSAFE
   TEST_MAKE_COMMAND2    0E6h, 0,     3,     0,      0, 0, 0   ; LFHSTATE

   TEST_MAKE_COMMAND2    0BAh, 0,     5,     0,  40000, 0, 0   ; LFPCRP
   TEST_MAKE_COMMAND2    0BAh, 0,     8,     1,  55000, 0, 0   ; LFPCRP

   TEST_MAKE_COMMAND2    095h, 0,     0,     0,      0, 0, 0   ; Illegal
   TEST_MAKE_COMMAND2    035h, 0,     0,     0,      0, 0, 0   ; error
   TEST_MAKE_COMMAND2    085h, 0,     0,     0,      0, 0, 0   ; Operate Only

$endif      ; (cmd_seq_num=??)
;-----------------------------------------------------------------------------

$endif      ; INCLUDE_TEST_CODE

END                         ; bootmain.a51

